@@ -58,6 +58,8 @@ static struct sc_atr_table cardos_atrs[] = {
 	{ "3b:d2:18:02:c1:0a:31:fe:58:c8:0d:51", NULL, NULL, SC_CARD_TYPE_CARDOS_M4_4, 0, NULL},
 	/* CardOS v5.0 */
 	{ "3b:d2:18:00:81:31:fe:58:c9:01:14", NULL, NULL, SC_CARD_TYPE_CARDOS_V5_0, 0, NULL},
+	/* CardOS v5.3 */
+	{ "3b:d2:18:00:81:31:fe:58:c9:03:16", NULL, NULL, SC_CARD_TYPE_CARDOS_V5_0, 0, NULL},
 	{ NULL, NULL, NULL, 0, 0, NULL }
 };
 
@@ -165,16 +167,21 @@ static int cardos_have_2048bit_package(sc_card_t *card)
 static int cardos_init(sc_card_t *card)
 {
 	unsigned long	flags, rsa_2048 = 0;
+	size_t data_field_length;
+	sc_apdu_t apdu;
+	u8 rbuf[2];
 
 	card->name = "CardOS M4";
 	card->cla = 0x00;
 
 	/* Set up algorithm info. */
-	flags = SC_ALGORITHM_NEED_USAGE
-		| SC_ALGORITHM_RSA_RAW
+	flags = SC_ALGORITHM_RSA_RAW
 		| SC_ALGORITHM_RSA_HASH_NONE
 		| SC_ALGORITHM_ONBOARD_KEY_GEN
 		;
+	if (card->type != SC_CARD_TYPE_CARDOS_V5_0)
+		flags |= SC_ALGORITHM_NEED_USAGE;
+
 	_sc_card_add_rsa_alg(card,  512, flags, 0);
 	_sc_card_add_rsa_alg(card,  768, flags, 0);
 	_sc_card_add_rsa_alg(card, 1024, flags, 0);
@@ -195,6 +202,29 @@ static int cardos_init(sc_card_t *card)
 		card->caps |= SC_CARD_CAP_APDU_EXT;
 	}
 
+	/* probe DATA FIELD LENGTH with GET DATA */
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xca, 0x01, 0x8D);
+	apdu.le = sizeof rbuf;
+	apdu.resp = rbuf;
+	apdu.resplen = sizeof(rbuf);
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL,
+			sc_transmit_apdu(card, &apdu),
+			"APDU transmit failed");
+	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL,
+			sc_check_sw(card, apdu.sw1, apdu.sw2),
+			"GET DATA command returned error");
+	if (apdu.resplen != 2)
+		return SC_ERROR_WRONG_LENGTH;
+	data_field_length = ((rbuf[0] << 8) | rbuf[1]);
+
+	/* strip the length of possible Lc and Le bytes */
+	if (card->caps & SC_CARD_CAP_APDU_EXT)
+		card->max_send_size = data_field_length - 6;
+	else
+		card->max_send_size = data_field_length - 3;
+	/* strip the length of SW bytes */
+	card->max_recv_size = data_field_length - 2;
+
 	if (rsa_2048 == 1) {
 		_sc_card_add_rsa_alg(card, 1280, flags, 0);
 		_sc_card_add_rsa_alg(card, 1536, flags, 0);
@@ -202,6 +232,11 @@ static int cardos_init(sc_card_t *card)
 		_sc_card_add_rsa_alg(card, 2048, flags, 0);
 	}
 
+	if (card->type == SC_CARD_TYPE_CARDOS_V5_0) {
+		/* Starting with CardOS 5, the card supports PIN query commands */
+		card->caps |= SC_CARD_CAP_ISO7816_PIN_INFO;
+	}
+
 	return 0;
 }
 
@@ -216,7 +251,7 @@ static const struct sc_card_error cardos_errors[] = {
 { 0x6f82, SC_ERROR_CARD_CMD_FAILED,	"not enough memory in xram"}, 
 { 0x6f84, SC_ERROR_CARD_CMD_FAILED,	"general protection fault"}, 
 
-/* the card doesn't now thic combination of ins+cla+p1+p2 */
+/* the card doesn't know this combination of ins+cla+p1+p2 */
 /* i.e. command will never work */
 { 0x6881, SC_ERROR_NO_CARD_SUPPORT,	"logical channel not supported"}, 
 { 0x6a86, SC_ERROR_INCORRECT_PARAMETERS,"p1/p2 invalid"}, 
@@ -508,6 +543,7 @@ static int cardos_set_file_attributes(sc_card_t *card, sc_file_t *file)
 				 * generation. */
 			case SC_FILE_EF_LINEAR_VARIABLE_TLV:
 				type[1] = 0xff;
+				/* fall through */
 			default:
 				type[0] |= file->ef_structure & 7;
 				break;
@@ -748,7 +784,7 @@ cardos_set_security_env(sc_card_t *card,
 		cardos_restore_security_env(card, 0x30);
 		apdu.p1 = 0xF1;
 	} else {
-		apdu.p1 = 0x01;
+		apdu.p1 = 0x41;
 	}
 	switch (env->operation) {
 	case SC_SEC_OPERATION_DECIPHER:
@@ -887,8 +923,9 @@ cardos_compute_signature(sc_card_t *card, const u8 *data, size_t datalen,
 	}
 
 	/* check if any operation was selected */
-	if(do_rsa_sig == 0 && do_rsa_pure_sig == 0)  {
-		/* no operation selected. we just have to try both, for the lack of any better reasoning */
+	if (do_rsa_sig == 0 && do_rsa_pure_sig == 0) {
+		/* no operation selected. we just have to try both,
+		 * for the lack of any better reasoning */
 		sc_log(ctx, "I was unable to determine, whether this key can be used with RSA_SIG or RSA_PURE_SIG. I will just try both.");
 		do_rsa_sig = 1;
 		do_rsa_pure_sig = 1;
@@ -1177,8 +1214,12 @@ cardos_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data,
 	data->pin_reference |= 0x80;
 
 	sc_log(ctx, "PIN_CMD(cmd:%i, ref:%i)", data->cmd, data->pin_reference);
-	sc_log(ctx, "PIN1(max:%i, min:%i)", data->pin1.max_length, data->pin1.min_length);
-	sc_log(ctx, "PIN2(max:%i, min:%i)", data->pin2.max_length, data->pin2.min_length);
+	sc_log(ctx,
+	       "PIN1(max:%"SC_FORMAT_LEN_SIZE_T"u, min:%"SC_FORMAT_LEN_SIZE_T"u)",
+	       data->pin1.max_length, data->pin1.min_length);
+	sc_log(ctx,
+	       "PIN2(max:%"SC_FORMAT_LEN_SIZE_T"u, min:%"SC_FORMAT_LEN_SIZE_T"u)",
+	       data->pin2.max_length, data->pin2.min_length);
 
 	/* FIXME: the following values depend on what pin length was
 	 * used when creating the BS objects */
@@ -1195,8 +1236,13 @@ cardos_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data,
 static int
 cardos_logout(sc_card_t *card)
 {
-	if (card->type == SC_CARD_TYPE_CARDOS_M4_01 ||
-	    card->type == SC_CARD_TYPE_CARDOS_M4_2) {
+	if (card->type == SC_CARD_TYPE_CARDOS_M4_01
+		   	|| card->type == SC_CARD_TYPE_CARDOS_M4_2
+		   	|| card->type == SC_CARD_TYPE_CARDOS_M4_2B
+		   	|| card->type == SC_CARD_TYPE_CARDOS_M4_2C
+		   	|| card->type == SC_CARD_TYPE_CARDOS_M4_3
+		   	|| card->type == SC_CARD_TYPE_CARDOS_M4_4
+			|| card->type == SC_CARD_TYPE_CARDOS_V5_0) {
 		sc_apdu_t apdu;
 		int       r;
 		sc_path_t path;
@@ -1217,35 +1263,6 @@ cardos_logout(sc_card_t *card)
 		return SC_ERROR_NOT_SUPPORTED;
 }
 
-static int cardos_get_data(struct sc_card *card, unsigned int tag,  u8 *buf, size_t len)
-{
-	int                             r;
-	struct sc_apdu                  apdu;
-
-	SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_VERBOSE);
-
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xCA,
-			(tag >> 8) & 0xff, tag & 0xff);
-	apdu.lc = 0;
-	apdu.datalen = 0;
-	apdu.le = len;
-	apdu.resp = buf;
-	apdu.resplen = len;
-	r = sc_transmit_apdu(card, &apdu);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-
-	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "GET_DATA returned error");
-
-	if (apdu.resplen > len)
-		r = SC_ERROR_WRONG_LENGTH;
-	else
-		r = apdu.resplen;
-
-	SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, r);
-}
-
-
 /* eToken R2 supports WRITE_BINARY, PRO Tokens support UPDATE_BINARY */
 
 static struct sc_card_driver * sc_get_driver(void)
@@ -1266,7 +1283,6 @@ static struct sc_card_driver * sc_get_driver(void)
 	cardos_ops.card_ctl = cardos_card_ctl;
 	cardos_ops.pin_cmd = cardos_pin_cmd;
 	cardos_ops.logout  = cardos_logout;
-	cardos_ops.get_data = cardos_get_data;
 
 	return &cardos_drv;
 }
