@@ -28,7 +28,7 @@
 #include "config.h"
 #endif
 
-#ifdef ENABLE_OPENSSL		/* empty file without openssl */
+#if defined(ENABLE_OPENSSL) && defined(ENABLE_SM)	/* empty file without openssl or sm */
 
 #include <stdlib.h>
 #include <string.h>
@@ -41,9 +41,24 @@
 #include "compression.h"
 #include "cwa14890.h"
 #include "cwa-dnie.h"
-#include "user-interface.h"
 
-extern cwa_provider_t *dnie_get_cwa_provider(sc_card_t * card);
+#ifdef _WIN32
+
+#ifndef UNICODE
+#define UNICODE
+#endif
+
+#include <windows.h>
+#endif
+#ifdef __APPLE__
+#include <Carbon/Carbon.h>
+#endif
+
+#define MAX_RESP_BUFFER_SIZE 2048
+
+/* default titles */
+#define USER_CONSENT_TITLE "Confirm"
+
 extern int dnie_read_file(
 	sc_card_t * card, 
 	const sc_path_t * path, 
@@ -142,6 +157,218 @@ const char *user_consent_message="Está a punto de realizar una firma electróni
 const char *user_consent_message="Esta a punto de realizar una firma digital\ncon su clave de FIRMA del DNI electronico.\nDesea permitir esta operacion?";
 #endif
 
+#ifdef ENABLE_DNIE_UI
+/**
+ * Messages used on pinentry protocol
+ */
+char *user_consent_msgs[] = { "SETTITLE", "SETDESC", "CONFIRM", "BYE" };
+
+/**
+ * Do fgets() without interruptions.
+ *
+ * Retry the operation if it is interrupted, such as with receiving an alarm.
+ *
+ * @param s Buffer receiving the data
+ * @param size Size of the buffer
+ * @param stream Stream to read
+ * @return s on success, NULL on error
+ */
+static char *nointr_fgets(char *s, int size, FILE *stream)
+{
+	while (fgets(s, size, stream) == NULL) {
+		if (feof(stream) || errno != EINTR)
+			return NULL;
+	}
+	return s;
+}
+
+/**
+ * Ask for user consent.
+ *
+ * Check for user consent configuration,
+ * Invoke proper gui app and check result
+ *
+ * @param card pointer to sc_card structure
+ * @param title Text to appear in the window header
+ * @param text Message to show to the user
+ * @return SC_SUCCESS on user consent OK , else error code
+ */
+int dnie_ask_user_consent(struct sc_card * card, const char *title, const char *message)
+{
+#ifdef __APPLE__
+	CFOptionFlags result;  /* result code from the message box */
+	/* convert the strings from char* to CFStringRef */
+	CFStringRef header_ref; /* to store title */
+	CFStringRef message_ref; /* to store message */
+#endif
+#ifdef linux
+	pid_t pid;
+	FILE *fin=NULL;
+	FILE *fout=NULL;	/* to handle pipes as streams */
+	struct stat st_file;	/* to verify that executable exists */
+	int srv_send[2];	/* to send data from server to client */
+	int srv_recv[2];	/* to receive data from client to server */
+	char outbuf[1024];	/* to compose and send messages */
+	char buf[1024];		/* to store client responses */
+	int n = 0;		/* to iterate on to-be-sent messages */
+#endif
+	int res = SC_ERROR_INTERNAL;	/* by default error :-( */
+	char *msg = NULL;	/* to makr errors */
+
+	if ((card == NULL) || (card->ctx == NULL))
+		return SC_ERROR_INVALID_ARGUMENTS;
+	LOG_FUNC_CALLED(card->ctx);
+
+	if ((title==NULL) || (message==NULL))
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+
+	if (GET_DNIE_UI_CTX(card).user_consent_enabled == 0
+			|| card->ctx->flags & SC_CTX_FLAG_DISABLE_POPUPS) {
+		sc_log(card->ctx,
+		       "User Consent or popups are disabled in configuration file");
+		LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+	}
+#ifdef _WIN32
+	/* in Windows, do not use pinentry, but MessageBox system call */
+	res = MessageBox (
+		NULL,
+		TEXT(message),
+		TEXT(title),
+		MB_ICONWARNING | MB_OKCANCEL | MB_DEFBUTTON2 | MB_APPLMODAL
+		);
+	if ( res == IDOK )
+		LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+	LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_ALLOWED);
+#elif __APPLE__
+	/* Also in Mac OSX use native functions */
+
+	/* convert the strings from char* to CFStringRef */
+	header_ref = CFStringCreateWithCString( NULL, title, strlen(title) );
+	message_ref = CFStringCreateWithCString( NULL,message, strlen(message) );
+
+	/* Displlay user notification alert */
+	CFUserNotificationDisplayAlert(
+		0, /* no timeout */
+		kCFUserNotificationNoteAlertLevel,  /* Alert level */
+		NULL,	/* IconURL, use default, you can change */
+			/* it depending message_type flags */
+		NULL,	/* SoundURL (not used) */
+		NULL,	/* localization of strings */
+		header_ref,	/* header. Cannot be null */
+		message_ref,	/* message text */
+		CFSTR("Cancel"), /* default ( "OK" if null) button text */
+		CFSTR("OK"), /* second button title */
+                NULL, /* third button title, null--> no other button */
+		&result /* response flags */
+	);
+
+	/* Clean up the strings */
+	CFRelease( header_ref );
+        CFRelease( message_ref );
+	/* Return 0 only if "OK" is selected */
+	if( result == kCFUserNotificationAlternateResponse )
+		LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+	LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_ALLOWED);
+#elif linux
+	/* check that user_consent_app exists. TODO: check if executable */
+	res = stat(GET_DNIE_UI_CTX(card).user_consent_app, &st_file);
+	if (res != 0) {
+		sc_log(card->ctx, "Invalid pinentry application: %s\n",
+		       GET_DNIE_UI_CTX(card).user_consent_app);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+	}
+
+	/* just a simple bidirectional pipe+fork+exec implementation */
+	/* In a pipe, xx[0] is for reading, xx[1] is for writing */
+	if (pipe(srv_send) < 0) {
+		msg = "pipe(srv_send)";
+		goto do_error;
+	}
+	if (pipe(srv_recv) < 0) {
+		msg = "pipe(srv_recv)";
+		goto do_error;
+	}
+	pid = fork();
+	switch (pid) {
+	case -1:		/* error  */
+		msg = "fork()";
+		goto do_error;
+	case 0:		/* child  */
+		/* make our pipes, our new stdin & stderr, closing older ones */
+		dup2(srv_send[0], STDIN_FILENO);	/* map srv send for input */
+		dup2(srv_recv[1], STDOUT_FILENO);	/* map srv_recv for output */
+		/* once dup2'd pipes are no longer needed on client; so close */
+		close(srv_send[0]);
+		close(srv_send[1]);
+		close(srv_recv[0]);
+		close(srv_recv[1]);
+		/* call exec() with proper user_consent_app from configuration */
+		/* if ok should never return */
+		execlp(GET_DNIE_UI_CTX(card).user_consent_app, GET_DNIE_UI_CTX(card).user_consent_app, (char *)NULL);
+		sc_log(card->ctx, "execlp() error");
+		abort();
+	default:		/* parent */
+		/* Close the pipe ends that the child uses to read from / write to
+		 * so when we close the others, an EOF will be transmitted properly.
+		 */
+		close(srv_send[0]);
+		close(srv_recv[1]);
+		/* use iostreams to take care on newlines and text based data */
+		fin = fdopen(srv_recv[0], "r");
+		if (fin == NULL) {
+			msg = "fdopen(in)";
+			goto do_error;
+		}
+		fout = fdopen(srv_send[1], "w");
+		if (fout == NULL) {
+			msg = "fdopen(out)";
+			goto do_error;
+		}
+		/* read and ignore first line */
+		if (nointr_fgets(buf, sizeof(buf), fin) == NULL) {
+			res = SC_ERROR_INTERNAL;
+			msg = "nointr_fgets() Unexpected IOError/EOF";
+			goto do_error;
+		}
+		for (n = 0; n<4; n++) {
+			char *pt;
+			if (n==0) snprintf(outbuf, sizeof outbuf,"%s %s\n",user_consent_msgs[0],title);
+			else if (n==1) snprintf(outbuf, sizeof outbuf,"%s %s\n",user_consent_msgs[1],message);
+			else snprintf(outbuf, sizeof outbuf,"%s\n",user_consent_msgs[n]);
+			/* send message */
+			fputs(outbuf, fout);
+			fflush(fout);
+			/* get response */
+			pt=nointr_fgets(buf, sizeof(buf), fin);
+			if (pt==NULL) {
+				res = SC_ERROR_INTERNAL;
+				msg = "nointr_fgets() Unexpected IOError/EOF";
+				goto do_error;
+			}
+			if (strstr(buf, "OK") == NULL) {
+				res = SC_ERROR_NOT_ALLOWED;
+				msg = "fail/cancel";
+				goto do_error;
+			}
+		}
+	}			/* switch */
+	/* arriving here means signature has been accepted by user */
+	res = SC_SUCCESS;
+	msg = NULL;
+do_error:
+	/* close out channel to force client receive EOF and also die */
+	if (fout != NULL) fclose(fout);
+	if (fin != NULL) fclose(fin);
+#else
+#error "Don't know how to handle user consent in this (rare) Operating System"
+#endif
+	if (msg != NULL)
+		sc_log(card->ctx, "%s", msg);
+	LOG_FUNC_RETURN(card->ctx, res);
+}
+
+#endif				/* ENABLE_DNIE_UI */
+
 /**
  * DNIe specific card driver operations
  */
@@ -195,9 +422,8 @@ static int dnie_get_environment(
 	/* look for sc block in opensc.conf */
 	ctx = card->ctx;
 	for (i = 0; ctx->conf_blocks[i]; i++) {
-		blocks =
-		    scconf_find_blocks(ctx->conf, ctx->conf_blocks[i],
-				       "card_driver", "dnie");
+		blocks = scconf_find_blocks(ctx->conf, ctx->conf_blocks[i],
+				"card_driver", "dnie");
 		if (!blocks)
 			continue;
 		blk = blocks[0];
@@ -293,7 +519,7 @@ data_found:
 static int dnie_get_info(sc_card_t * card, char *data[])
 {
 	sc_file_t *file = NULL;
-        sc_path_t *path = NULL;
+        sc_path_t path;
         u8 *buffer = NULL;
 	size_t bufferlen = 0;
 	char *msg = NULL;
@@ -309,14 +535,8 @@ static int dnie_get_info(sc_card_t * card, char *data[])
 	/* phase 1: get DNIe number, Name and GivenName */
 
 	/* read EF(CDF) at 3F0050156004 */
-	path = (sc_path_t *) calloc(1, sizeof(sc_path_t));
-	if (!path) {
-		msg = "Cannot allocate path data for EF(CDF) read";
-		res = SC_ERROR_OUT_OF_MEMORY;
-		goto get_info_end;
-	}
-	sc_format_path("3F0050156004", path);
-	res = dnie_read_file(card, path, &file, &buffer, &bufferlen);
+	sc_format_path("3F0050156004", &path);
+	res = dnie_read_file(card, &path, &file, &buffer, &bufferlen);
 	if (res != SC_SUCCESS) {
 		msg = "Cannot read EF(CDF)";
 		goto get_info_end;
@@ -334,17 +554,15 @@ static int dnie_get_info(sc_card_t * card, char *data[])
         }
 
 	/* phase 2: get IDESP */
-	sc_format_path("3F000006", path);
-	if (file) {
-		sc_file_free(file);
-		file = NULL;
-	}
+	sc_format_path("3F000006", &path);
+	sc_file_free(file);
+	file = NULL;
 	if (buffer) {
 		free(buffer); 
 		buffer=NULL; 
 		bufferlen=0;
 	}
-	res = dnie_read_file(card, path, &file, &buffer, &bufferlen);
+	res = dnie_read_file(card, &path, &file, &buffer, &bufferlen);
 	if (res != SC_SUCCESS) {
 		data[3]=NULL;
 		goto get_info_ph3;
@@ -359,11 +577,9 @@ static int dnie_get_info(sc_card_t * card, char *data[])
 
 get_info_ph3:
 	/* phase 3: get DNIe software version */
-	sc_format_path("3F002F03", path);
-	if (file) {
-		sc_file_free(file);
-		file = NULL;
-	}
+	sc_format_path("3F002F03", &path);
+	sc_file_free(file);
+	file = NULL;
 	if (buffer) {
 		free(buffer); 
 		buffer=NULL; 
@@ -373,7 +589,7 @@ get_info_ph3:
 	* Some old DNIe cards seems not to include SW version file,
  	* so let this code fail without notice
  	*/
-	res = dnie_read_file(card, path, &file, &buffer, &bufferlen);
+	res = dnie_read_file(card, &path, &file, &buffer, &bufferlen);
 	if (res != SC_SUCCESS) {
 		msg = "Cannot read DNIe Version EF";
 		data[4]=NULL;
@@ -393,15 +609,15 @@ get_info_ph3:
 	msg = NULL;
 
 get_info_end:
-	if (file) {
-		sc_file_free(file);
+	sc_file_free(file);
+	file = NULL;
+	if (buffer) {
 		free(buffer);
-		file = NULL;
-		buffer = NULL;
-		bufferlen = 0;
+		buffer=NULL;
+		bufferlen=0;
 	}
 	if (msg)
-		sc_log(card->ctx,msg);
+		sc_log(card->ctx, "%s", msg);
         LOG_FUNC_RETURN(card->ctx, res);
 }
 
@@ -419,7 +635,7 @@ static int dnie_get_serialnr(sc_card_t * card, sc_serial_number_t * serial)
 {
 	int result;
 	sc_apdu_t apdu;
-	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 rbuf[MAX_RESP_BUFFER_SIZE];
 	if ((card == NULL) || (card->ctx == NULL) || (serial == NULL))
 		return SC_ERROR_INVALID_ARGUMENTS;
 
@@ -434,17 +650,15 @@ static int dnie_get_serialnr(sc_card_t * card, sc_serial_number_t * serial)
 		LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 	}
 	/* not cached, retrieve it by mean of an APDU */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xb8, 0x00, 0x00);
+	/* official driver read 0x11 bytes, but only uses 7. Manual says just 7 (for le) */
+	dnie_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xb8, 0x00, 0x00, 0x07, 0,
+					rbuf, sizeof(rbuf), NULL, 0);
 	apdu.cla = 0x90;	/* propietary cmd */
-	apdu.resp = rbuf;
-	apdu.resplen = sizeof(rbuf);
-	/* official driver read 0x11 bytes, but only uses 7. Manual says just 7 */
-	apdu.le = 0x07;
-	apdu.lc = 0;
-	apdu.datalen = 0;
 	/* send apdu */
-	result = dnie_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(card->ctx, result, "APDU transmit failed");
+	result = sc_transmit_apdu(card, &apdu);
+	if (result != SC_SUCCESS) {
+		LOG_TEST_RET(card->ctx, result, "APDU transmit failed");
+	}
 	if (apdu.sw1 != 0x90 || apdu.sw2 != 0x00)
 		return SC_ERROR_INTERNAL;
 	/* cache serial number */
@@ -487,7 +701,7 @@ static void dnie_clear_cache(dnie_private_data_t * data)
  *
  * @param card pointer to card data
  */
-static inline void init_flags(struct sc_card *card)
+static void init_flags(struct sc_card *card)
 {
 	unsigned long algoflags;
 	/* set up flags according documentation */
@@ -529,6 +743,78 @@ int dnie_match_card(struct sc_card *card)
 	LOG_FUNC_RETURN(card->ctx, result);
 }
 
+static int dnie_sm_free_wrapped_apdu(struct sc_card *card,
+		struct sc_apdu *plain, struct sc_apdu **sm_apdu)
+{
+	struct sc_context *ctx = card->ctx;
+	cwa_provider_t *provider = NULL;
+	int rv = SC_SUCCESS;
+
+	LOG_FUNC_CALLED(ctx);
+	provider = GET_DNIE_PRIV_DATA(card)->cwa_provider;
+	if (!sm_apdu)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
+	if (!(*sm_apdu))
+		LOG_FUNC_RETURN(ctx, SC_SUCCESS);
+
+	if ((*sm_apdu) != plain) {
+		rv = cwa_decode_response(card, provider, *sm_apdu);
+		if (plain) {
+			plain->resplen = (*sm_apdu)->resplen;
+			plain->sw1 = (*sm_apdu)->sw1;
+			plain->sw2 = (*sm_apdu)->sw2;
+			if (((*sm_apdu)->data) != plain->data)
+				free((unsigned char *) (*sm_apdu)->data);
+			if ((*sm_apdu)->resp != plain->resp)
+				free((*sm_apdu)->resp);
+		}
+		free(*sm_apdu);
+	}
+	*sm_apdu = NULL;
+
+	LOG_FUNC_RETURN(ctx, rv);
+}
+
+static int dnie_sm_get_wrapped_apdu(struct sc_card *card,
+		struct sc_apdu *plain, struct sc_apdu **sm_apdu)
+{
+	struct sc_context *ctx = card->ctx;
+	struct sc_apdu *apdu = NULL;
+	cwa_provider_t *provider = NULL;
+	int rv = SC_SUCCESS;
+
+	LOG_FUNC_CALLED(ctx);
+	if (!plain || !sm_apdu)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
+
+	provider = GET_DNIE_PRIV_DATA(card)->cwa_provider;
+
+	if (((plain->cla & 0x0C) == 0) && (plain->ins != 0xC0)) {
+		*sm_apdu = NULL;
+		//construct new SM apdu from original apdu
+		apdu = calloc(1, sizeof(struct sc_apdu));
+		if (!apdu)
+			return SC_ERROR_OUT_OF_MEMORY;
+
+		memcpy(apdu, plain, sizeof(sc_apdu_t));
+
+		rv = cwa_encode_apdu(card, provider, plain, apdu);
+
+		if (rv != SC_SUCCESS) {
+			dnie_sm_free_wrapped_apdu(card, NULL, &apdu);
+			goto err;
+		}
+
+		*sm_apdu = apdu;
+	} else 
+		*sm_apdu = plain;
+
+	apdu = NULL;
+err:
+	free(apdu);
+	LOG_FUNC_RETURN(ctx, rv);
+}
+
 /**
  * OpenDNIe card structures initialization.
  *
@@ -556,19 +842,16 @@ static int dnie_init(struct sc_card *card)
 	if (!provider) 
 	    LOG_TEST_RET(card->ctx, SC_ERROR_INTERNAL, "Error initializing cwa-dnie provider");
 
-#ifdef ENABLE_SM
 	/** Secure messaging initialization section **/
 	memset(&(card->sm_ctx), 0, sizeof(sm_context_t));
-	card->sm_ctx.ops.get_sm_apdu = NULL;
-	card->sm_ctx.ops.free_sm_apdu = NULL;
-#endif
+	card->sm_ctx.ops.get_sm_apdu = dnie_sm_get_wrapped_apdu;
+	card->sm_ctx.ops.free_sm_apdu = dnie_sm_free_wrapped_apdu;
+	card->sm_ctx.sm_mode = SM_MODE_NONE;
 
 	init_flags(card);
 
-#ifdef ENABLE_SM
 	res=cwa_create_secure_channel(card,provider,CWA_SM_OFF);
 	LOG_TEST_RET(card->ctx, res, "Failure creating CWA secure channel.");
-#endif
 
 	/* initialize private data */
 	card->drv_data = calloc(1, sizeof(dnie_private_data_t));
@@ -603,12 +886,10 @@ static int dnie_finish(struct sc_card *card)
 	int result = SC_SUCCESS;
 	LOG_FUNC_CALLED(card->ctx);
 	dnie_clear_cache(GET_DNIE_PRIV_DATA(card));
-#ifdef ENABLE_SM
 	/* disable sm channel if established */
 	result = cwa_create_secure_channel(card, GET_DNIE_PRIV_DATA(card)->cwa_provider, CWA_SM_OFF);
-#endif
-	if (card->drv_data != NULL)
-		free(card->drv_data);
+	free(GET_DNIE_PRIV_DATA(card)->cwa_provider);
+	free(card->drv_data);
 	LOG_FUNC_RETURN(card->ctx, result);
 }
 
@@ -641,12 +922,12 @@ static unsigned long le2ulong(u8 * pt)
  */
 static u8 *dnie_uncompress(sc_card_t * card, u8 * from, size_t *len)
 {
-	int res = SC_SUCCESS;
 	u8 *upt = from;
+#ifdef ENABLE_ZLIB
+	int res = SC_SUCCESS;
 	size_t uncompressed = 0L;
 	size_t compressed = 0L;
 
-#ifdef ENABLE_ZLIB
 	if (!card || !card->ctx || !from || !len)
 		return NULL;
 	LOG_FUNC_CALLED(card->ctx);
@@ -683,7 +964,8 @@ static u8 *dnie_uncompress(sc_card_t * card, u8 * from, size_t *len)
 	*len = uncompressed;
 	sc_log(card->ctx, "Compressed data:\n%s\n",
 	       sc_dump_hex(from + 8, compressed));
-	sc_log(card->ctx, "Uncompress() done. Before:'%lu' After: '%lu'",
+	sc_log(card->ctx,
+	       "Uncompress() done. Before:'%"SC_FORMAT_LEN_SIZE_T"u' After: '%"SC_FORMAT_LEN_SIZE_T"u'",
 	       compressed, uncompressed);
 	sc_log(card->ctx, "Uncompressed data:\n%s\n",
 	       sc_dump_hex(upt, uncompressed));
@@ -716,7 +998,7 @@ static u8 *dnie_uncompress(sc_card_t * card, u8 * from, size_t *len)
  */
 static int dnie_fill_cache(sc_card_t * card)
 {
-	u8 tmp[SC_MAX_APDU_BUFFER_SIZE];
+	u8 tmp[MAX_RESP_BUFFER_SIZE];
 	sc_apdu_t apdu;
 	size_t count = 0;
 	size_t len = 0;
@@ -744,13 +1026,14 @@ static int dnie_fill_cache(sc_card_t * card)
 		apdu.p1 = 0xff & (len >> 8);
 		apdu.p2 = 0xff & len;
 		apdu.le = count;
-		apdu.resplen = count;
+		apdu.resplen = MAX_RESP_BUFFER_SIZE;
 		apdu.resp = tmp;
 		/* transmit apdu */
-		r = dnie_transmit_apdu(card, &apdu);
+		r = sc_transmit_apdu(card, &apdu);
 		if (r != SC_SUCCESS) {
-			if (buffer)
-				free(buffer);
+			free(buffer);
+			if (apdu.resp != tmp)
+				free(apdu.resp);
 			sc_log(ctx, "read_binary() APDU transmit failed");
 			LOG_FUNC_RETURN(ctx, r);
 		}
@@ -760,20 +1043,33 @@ static int dnie_fill_cache(sc_card_t * card)
 			r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 			if (r == SC_ERROR_WRONG_LENGTH) {
 				count = 0xff & apdu.sw2;
-				if (count != 0)
+				if (count != 0) {
+					if (apdu.resp != tmp)
+						free(apdu.resp);
 					continue;	/* read again with correct size */
+				}
 				goto read_done;	/* no more data to read */
 			}
 			if (r == SC_ERROR_INCORRECT_PARAMETERS)
 				goto read_done;
+			if (apdu.resp != tmp)
+				free(apdu.resp);
 			LOG_FUNC_RETURN(ctx, r);	/* arriving here means response error */
 		}
 		/* copy received data into buffer. realloc() if not enought space */
 		count = apdu.resplen;
 		buffer = realloc(buffer, len + count);
-		if (!buffer)
+		if (!buffer) {
+			free((void *)apdu.data);
+			if (apdu.resp != tmp)
+				free(apdu.resp);
 			LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+		}
 		memcpy(buffer + len, apdu.resp, count);
+		if (apdu.resp != tmp) {
+			free(apdu.resp);
+			apdu.resp = tmp;
+		}
 		len += count;
 		if (count != card->max_recv_size)
 			goto read_done;
@@ -782,20 +1078,23 @@ static int dnie_fill_cache(sc_card_t * card)
  read_done:
 	/* no more data to read: check if data is compressed */
 	pt = dnie_uncompress(card, buffer, &len);
+	free((void *)apdu.data);
+	if (apdu.resp != tmp)
+		free(apdu.resp);
 	if (pt == NULL) {
 		sc_log(ctx, "Uncompress proccess failed");
-		if (buffer)
-			free(buffer);
+		free(buffer);
 		LOG_FUNC_RETURN(ctx, SC_ERROR_INTERNAL);
 	}
 	if (pt != buffer)
-		if (buffer)
-			free(buffer);
+		free(buffer);
 
 	/* ok: as final step, set correct cache data into dnie_priv structures */
 	GET_DNIE_PRIV_DATA(card)->cache = pt;
 	GET_DNIE_PRIV_DATA(card)->cachelen = len;
-	sc_log(ctx, "fill_cache() done. length '%d' bytes", len);
+	sc_log(ctx,
+	       "fill_cache() done. length '%"SC_FORMAT_LEN_SIZE_T"u' bytes",
+	       len);
 	LOG_FUNC_RETURN(ctx,len);
 }
 
@@ -855,7 +1154,7 @@ static int dnie_compose_and_send_apdu(sc_card_t *card, const u8 *path, size_t pa
 {
 	int res = 0;
 	sc_apdu_t apdu;
-	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 rbuf[MAX_RESP_BUFFER_SIZE];
 	sc_file_t *file = NULL;
 
 	sc_context_t *ctx = NULL;
@@ -866,25 +1165,18 @@ static int dnie_compose_and_send_apdu(sc_card_t *card, const u8 *path, size_t pa
 
 	LOG_FUNC_CALLED(ctx);
 
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "called, p1=%u, path=%s\n", p1, sc_dump_hex(path, pathlen));
-
-	/* Arriving here means need to compose and send apdu */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0xA4, p1, 0);
-	apdu.resp = rbuf;
-	apdu.resplen = sizeof(rbuf);
-	apdu.lc = pathlen;
-	apdu.data = path;
-	apdu.datalen = pathlen;
-	apdu.le = card->max_recv_size > 0 ? card->max_recv_size : 256;
+	dnie_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0xA4, p1, 0, 
+					sc_get_max_recv_size(card), pathlen,
+					rbuf, sizeof(rbuf), path, pathlen);
 	if (p1 == 3)
 		apdu.cse= SC_APDU_CASE_1;
 
-	if (file_out == NULL) {
-		apdu.cse = (apdu.lc == 0) ? SC_APDU_CASE_1 : SC_APDU_CASE_3_SHORT;
-		apdu.le = 0;
-	}
-	res = dnie_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(ctx, res, "SelectFile() APDU transmit failed");
+	if (file_out == NULL)
+		apdu.cse = SC_APDU_CASE_4_SHORT;
+
+	res = sc_transmit_apdu(card, &apdu);
+	if ((res != SC_SUCCESS) || (file_out == NULL))
+		LOG_TEST_RET(ctx, res, "SelectFile() APDU transmit failed");
 	if (file_out == NULL) {
 		if (apdu.sw1 == 0x61)
 			SC_FUNC_RETURN(ctx, SC_LOG_DEBUG_VERBOSE, 0);
@@ -894,17 +1186,20 @@ static int dnie_compose_and_send_apdu(sc_card_t *card, const u8 *path, size_t pa
 
 	/* analyze response. if FCI, try to parse */
 	res = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	LOG_TEST_RET(ctx, res, "SelectFile() check_sw failed");
-	if (apdu.resplen < 2)
-		LOG_FUNC_RETURN(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
-	if (apdu.resp[0] == 0x00)	/* proprietary coding */
+	if (res != SC_SUCCESS) {
+		LOG_TEST_RET(ctx, res, "SelectFile() check_sw failed");
+	}
+	if ((apdu.resplen < 2) || (apdu.resp[0] == 0x00)) {
 		LOG_FUNC_RETURN(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
+	}
 
 	/* finally process FCI response */
 	file = sc_file_new();
-	if (file == NULL)
+	if (file == NULL) {
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+	}
 	res = card->ops->process_fci(card, file, apdu.resp + 2, apdu.resp[1]);
+	sc_file_free(*file_out);
 	*file_out = file;
 	LOG_FUNC_RETURN(ctx, res);
 }
@@ -977,7 +1272,6 @@ static int dnie_select_file(struct sc_card *card,
 
 		sc_log(ctx, "select_file(PATH): requested:%s ", sc_dump_hex(in_path->value, in_path->len));
 
-
 		/* convert to SC_PATH_TYPE_FILE_ID */
 		res = sc_lock(card); /* lock to ensure path traversal */
 		LOG_TEST_RET(ctx, res, "sc_lock() failed");
@@ -1050,10 +1344,12 @@ static int dnie_select_file(struct sc_card *card,
  * @param len requested challenge length
  * @return SC_SUCCESS if OK; else error code
  */
+#define BUFFER_SIZE 8
+
 static int dnie_get_challenge(struct sc_card *card, u8 * rnd, size_t len)
 {
 	sc_apdu_t apdu;
-	u8 buf[10];
+	u8 buf[MAX_RESP_BUFFER_SIZE];
 	int result = SC_SUCCESS;
 	if ((card == NULL) || (card->ctx == NULL))
 		return SC_ERROR_INVALID_ARGUMENTS;
@@ -1065,20 +1361,24 @@ static int dnie_get_challenge(struct sc_card *card, u8 * rnd, size_t len)
 		result = SC_ERROR_INVALID_ARGUMENTS;
 		goto dnie_get_challenge_error;
 	}
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x84, 0x00, 0x00);
-	apdu.le = 8;
-	apdu.resp = buf;
-	apdu.resplen = 8;	/* include SW's */
+	dnie_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x84, 0x00, 0x00, BUFFER_SIZE, 0,
+					buf, MAX_RESP_BUFFER_SIZE, NULL, 0);
 
 	/* 
 	* As DNIe cannot handle other data length than 0x08 and 0x14, 
 	* perform consecutive reads of 8 bytes until retrieve requested length
 	*/
 	while (len > 0) {
-		size_t n = len > 8 ? 8 : len;
-		result = dnie_transmit_apdu(card, &apdu);
-		LOG_TEST_RET(card->ctx, result, "APDU transmit failed");
-		if (apdu.resplen != 8) {
+		size_t n = len > BUFFER_SIZE ? BUFFER_SIZE : len;
+		sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x84, 0x00, 0x00);
+		apdu.le = BUFFER_SIZE;
+		apdu.resp = buf;
+		apdu.resplen = MAX_RESP_BUFFER_SIZE;	/* include SW's */
+		result = sc_transmit_apdu(card, &apdu);
+		if (result != SC_SUCCESS) {
+			LOG_TEST_RET(card->ctx, result, "APDU transmit failed");
+		}
+		if (apdu.resplen != BUFFER_SIZE) {
 			result = sc_check_sw(card, apdu.sw1, apdu.sw2);
 			goto dnie_get_challenge_error;
 		}
@@ -1106,16 +1406,26 @@ static int dnie_get_challenge(struct sc_card *card, u8 * rnd, size_t len)
 static int dnie_logout(struct sc_card *card)
 {
 	int result = SC_SUCCESS;
+	sc_file_t *file = NULL;
 
 	if ((card == NULL) || (card->ctx == NULL))
 		return SC_ERROR_INVALID_ARGUMENTS;
+
 	LOG_FUNC_CALLED(card->ctx);
-#ifdef ENABLE_SM
-	/* disable and free any sm channel related data */
-	result =
-	    cwa_create_secure_channel(card, GET_DNIE_PRIV_DATA(card)->cwa_provider, CWA_SM_OFF);
-#endif
-	/* TODO: _logout() see comments.txt on what to do here */
+	if (card->sm_ctx.sm_mode != SM_MODE_NONE) {
+		/* mark the channel as closed */
+		result = cwa_create_secure_channel(card, 
+			GET_DNIE_PRIV_DATA(card)->cwa_provider, CWA_SM_OFF);
+		LOG_TEST_RET(card->ctx, result, "Cannot close the secure channel");
+		/* request the Master File to provoke an SM error and close the channel */
+		result = dnie_compose_and_send_apdu(card, (const u8 *) DNIE_MF_NAME, 
+			sizeof(DNIE_MF_NAME) - 1, 4, &file);
+		if (result == SC_ERROR_SM)
+			result = SC_SUCCESS;
+	}
+
+	if (file != NULL)
+		sc_file_free(file);
 	LOG_FUNC_RETURN(card->ctx, result);
 }
 
@@ -1141,6 +1451,7 @@ static int dnie_set_security_env(struct sc_card *card,
 {
 	sc_apdu_t apdu;
 	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];	/* buffer to compose apdu data */
+	u8 rbuf[MAX_RESP_BUFFER_SIZE];
 	u8 *p = sbuf;
 	int result = SC_SUCCESS;
 	if ((card == NULL) || (card->ctx == NULL) || (env == NULL))
@@ -1212,7 +1523,8 @@ static int dnie_set_security_env(struct sc_card *card,
 	}
 
 	/* create and format apdu */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x22, 0x00, 0x00);
+	dnie_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x22, 0x00, 0x00, 255, p - sbuf,
+					rbuf, MAX_RESP_BUFFER_SIZE, sbuf, p - sbuf);
 
 	/* check and perform operation */
 	switch (env->operation) {
@@ -1234,18 +1546,12 @@ static int dnie_set_security_env(struct sc_card *card,
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 	}
 
-	/* complete apdu contents with buffer data */
-	apdu.data = sbuf;
-	apdu.datalen = p - sbuf;
-	apdu.lc = p - sbuf;
-	apdu.resplen = 0;
-
 	/* Notice that Manual states that DNIE only allows handle of 
 	 * current security environment, so se_num is ignored, and
 	 * store sec env apdu (00 22 F2 se_num) command will not be issued */
 
 	/* send composed apdu and parse result */
-	result = dnie_transmit_apdu(card, &apdu);
+	result = sc_transmit_apdu(card, &apdu);
 	LOG_TEST_RET(card->ctx, result, "Set Security Environment failed");
 	result = sc_check_sw(card, apdu.sw1, apdu.sw2);
 
@@ -1279,7 +1585,7 @@ static int dnie_decipher(struct sc_card *card,
 {
 	struct sc_apdu apdu;
 	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
-	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 sbuf[MAX_RESP_BUFFER_SIZE];
 	size_t len;
 	int result = SC_SUCCESS;
 	if ((card == NULL) || (card->ctx == NULL))
@@ -1299,21 +1605,17 @@ static int dnie_decipher(struct sc_card *card,
 	 * (90 74 40 keyID). This code uses standard 00 2A 80 8x one)
 	 * as shown in card-atrust-acos.c and card-jcop.c
 	 */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x2A,	/* INS: 0x2A  perform security operation */
-		       0x80,	/* P1: Response is plain value */
-		       0x86	/* P2: 8x: Padding indicator byte followed by cryptogram */
+	dnie_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 
+			0x2A,	/* INS: 0x2A  perform security operation */
+			0x80,	/* P1: Response is plain value */
+			0x86,	/* P2: 8x: Padding indicator byte followed by cryptogram */
+			256, crgram_len + 1, rbuf, sizeof(rbuf), sbuf, crgram_len + 1
 	    );
-	apdu.resp = rbuf;
-	apdu.resplen = sizeof(rbuf);
 
 	sbuf[0] = 0;		/* padding indicator byte, 0x00 = No further indication */
 	memcpy(sbuf + 1, crgram, crgram_len);
-	apdu.data = sbuf;
-	apdu.lc = crgram_len + 1;
-	apdu.datalen = crgram_len + 1;
-	apdu.le = 256;
 	/* send apdu */
-	result = dnie_transmit_apdu(card, &apdu);
+	result = sc_transmit_apdu(card, &apdu);
 	LOG_TEST_RET(card->ctx, result, "APDU transmit failed");
 	/* check response */
 	result = sc_check_sw(card, apdu.sw1, apdu.sw2);
@@ -1355,8 +1657,9 @@ static int dnie_compute_signature(struct sc_card *card,
 				  u8 * out, size_t outlen)
 {
 	int result = SC_SUCCESS;
+	int result_resplen = 0;
 	struct sc_apdu apdu;
-	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];	/* to receive sign response */
+	u8 rbuf[MAX_RESP_BUFFER_SIZE];	/* to receive sign response */
 
 	/* some preliminar checks */
 	if ((card == NULL) || (card->ctx == NULL))
@@ -1374,7 +1677,7 @@ static int dnie_compute_signature(struct sc_card *card,
 #ifdef ENABLE_DNIE_UI
 	/* (Requested by DGP): on signature operation, ask user consent */
 	if (GET_DNIE_PRIV_DATA(card)->rsa_key_ref == 0x02) {	/* TODO: revise key ID handling */
-		result = sc_ask_user_consent(card,user_consent_title,user_consent_message);
+		result = dnie_ask_user_consent(card,user_consent_title,user_consent_message);
 		LOG_TEST_RET(card->ctx, result, "User consent denied");
 	}
 #endif    
@@ -1387,30 +1690,30 @@ static int dnie_compute_signature(struct sc_card *card,
 	   data and feed them into sign() command
 	 */
 	sc_log(card->ctx,
-	       "Compute signature len: '%d' bytes:\n%s\n============================================================",
+	       "Compute signature len: '%"SC_FORMAT_LEN_SIZE_T"u' bytes:\n%s\n============================================================",
 	       datalen, sc_dump_hex(data, datalen));
 
 	/*INS: 0x2A  PERFORM SECURITY OPERATION
 	 * P1:  0x9E  Resp: Digital Signature
 	 * P2:  0x9A  Cmd: Input for Digital Signature */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x2A, 0x9E, 0x9A);
-	apdu.resp = rbuf;
-	apdu.resplen = sizeof(rbuf);
-	apdu.le = 256;		/* signature response size */
-	apdu.data = data;
-	apdu.lc = datalen;	/*  Caller determines the type of hash and its size */
-	apdu.datalen = datalen;
+	dnie_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x2A, 0x9E, 0x9A, 256, datalen,
+					rbuf, sizeof(rbuf), data, datalen);
 	/* tell card to compute signature */
-	result = dnie_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(card->ctx, result, "compute_signature() failed");
+	result = sc_transmit_apdu(card, &apdu);
+	if (result != SC_SUCCESS) {
+		LOG_TEST_RET(card->ctx, result, "compute_signature() failed");
+	}
 	/* check response */
 	result = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	LOG_TEST_RET(card->ctx, result, "compute_signature() response error");
+	if (result != SC_SUCCESS) {
+		LOG_TEST_RET(card->ctx, result, "compute_signature() response error");
+	}
 
 	/* ok: copy result from buffer */
-	memcpy(out, apdu.resp, apdu.resplen);
+	result_resplen = apdu.resplen;
+	memcpy(out, apdu.resp, result_resplen);
 	/* and return response length */
-	LOG_FUNC_RETURN(card->ctx, apdu.resplen);
+	LOG_FUNC_RETURN(card->ctx, result_resplen);
 }
 
 /*
@@ -1440,7 +1743,6 @@ static int dnie_list_files(sc_card_t * card, u8 * buf, size_t buflen)
 	size_t count = 0;
 	u8 data[2];
 	sc_apdu_t apdu;
-	sc_apdu_t back;
 	if (!card || !card->ctx)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
@@ -1449,21 +1751,8 @@ static int dnie_list_files(sc_card_t * card, u8 * buf, size_t buflen)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 
 	/* compose select_file(ID) command */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xA4, 0x00, 0x00);
-	apdu.le = 0;
-	apdu.lc = 2;
-	apdu.data = data;
-	apdu.resp = NULL;
-	apdu.datalen = 2;
-	apdu.resplen = 0;
-	/* compose select_file(PARENT) command */
-	sc_format_apdu(card, &back, SC_APDU_CASE_1, 0xA4, 0x03, 0x00);
-	back.le = 0;
-	back.lc = 0;
-	back.data = NULL;
-	back.resp = NULL;
-	back.datalen = 0;
-	back.resplen = 0;
+	dnie_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0xA4, 0x00, 0x00, 0, 2,
+					NULL, 0, data, 2);
 	/* iterate on every possible ids */
 	for (id1 = 0; id1 < 256; id1++) {
 		for (id2 = 0; id2 < 256; id2++) {
@@ -1482,7 +1771,7 @@ static int dnie_list_files(sc_card_t * card, u8 * buf, size_t buflen)
 			/* compose and transmit select_file() cmd */
 			data[0] = (u8) (0xff & id1);
 			data[1] = (u8) (0xff & id2);
-			res = dnie_transmit_apdu(card, &apdu);
+			res = sc_transmit_apdu(card, &apdu);
 			if (res != SC_SUCCESS) {
 				sc_log(card->ctx, "List file '%02X%02X' failed",
 				       id1, id2);
@@ -1619,7 +1908,7 @@ static int dnie_read_header(struct sc_card *card)
 {
 	sc_apdu_t apdu;
 	int r;
-	u8 buf[SC_MAX_APDU_BUFFER_SIZE];
+	u8 buf[MAX_RESP_BUFFER_SIZE];
 	unsigned long uncompressed = 0L;
 	unsigned long compressed = 0L;
 	sc_context_t *ctx = NULL;
@@ -1630,14 +1919,10 @@ static int dnie_read_header(struct sc_card *card)
 	LOG_FUNC_CALLED(ctx);
 
 	/* initialize apdu */
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xB0, 0x00, 0x00);
-	apdu.p1 = 0x00;
-	apdu.p2 = 0x00;
-	apdu.le = 8;		/* read 8 bytes at begining of file */
-	apdu.resplen = SC_MAX_APDU_BUFFER_SIZE;
-	apdu.resp = buf;
+	dnie_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xB0, 0x00, 0x00, 8, 0,
+					buf, MAX_RESP_BUFFER_SIZE, NULL, 0);
 	/* transmit apdu */
-	r = dnie_transmit_apdu(card, &apdu);
+	r = sc_transmit_apdu(card, &apdu);
 	if (r != SC_SUCCESS) {
 		sc_log(ctx, "read_header() APDU transmit failed");
 		LOG_FUNC_RETURN(ctx, r);
@@ -1860,11 +2145,9 @@ static int dnie_pin_change(struct sc_card *card, struct sc_pin_cmd_data * data)
 {
 	int res=SC_SUCCESS;
 	LOG_FUNC_CALLED(card->ctx);
-#ifdef ENABLE_SM
-    /* Ensure that secure channel is established from reset */
-    res = cwa_create_secure_channel(card, GET_DNIE_PRIV_DATA(card)->cwa_provider, CWA_SM_COLD);
-    LOG_TEST_RET(card->ctx, res, "Establish SM failed");
-#endif
+	/* Ensure that secure channel is established from reset */
+	res = cwa_create_secure_channel(card, GET_DNIE_PRIV_DATA(card)->cwa_provider, CWA_SM_ON);
+	LOG_TEST_RET(card->ctx, res, "Establish SM failed");
 	LOG_FUNC_RETURN(card->ctx,SC_ERROR_NOT_SUPPORTED);
 }
 
@@ -1881,7 +2164,6 @@ static int dnie_pin_change(struct sc_card *card, struct sc_pin_cmd_data * data)
 static int dnie_pin_verify(struct sc_card *card,
                         struct sc_pin_cmd_data *data, int *tries_left)
 {
-#ifdef ENABLE_SM
 	int res=SC_SUCCESS;
 	sc_apdu_t apdu;
 
@@ -1891,7 +2173,12 @@ static int dnie_pin_verify(struct sc_card *card,
 
 	LOG_FUNC_CALLED(card->ctx);
 	/* ensure that secure channel is established from reset */
-	res = cwa_create_secure_channel(card, GET_DNIE_PRIV_DATA(card)->cwa_provider, CWA_SM_COLD);
+	if (card->atr.value[15] >= DNIE_30_VERSION) {
+		/* the provider should be prepared for using PIN information */
+		sc_log(card->ctx, "DNIe 3.0 detected doing PIN initialization");
+		dnie_change_cwa_provider_to_pin(card);
+	}
+	res = cwa_create_secure_channel(card, GET_DNIE_PRIV_DATA(card)->cwa_provider, CWA_SM_ON);
 	LOG_TEST_RET(card->ctx, res, "Establish SM failed");
 
 	data->apdu = &apdu;	/* prepare apdu struct */
@@ -1905,21 +2192,14 @@ static int dnie_pin_verify(struct sc_card *card,
 	pinlen = res;
 
 	/* compose apdu */
-	memset(&apdu, 0, sizeof(apdu));	/* clear buffer */
-	apdu.cla = 0x00;
-	apdu.cse = SC_APDU_CASE_3_SHORT;
-	apdu.ins = (u8) 0x20;	/* Verify cmd */
-	apdu.p1 = (u8) 0x00;
-	apdu.p2 = (u8) 0x00;
-	apdu.lc = pinlen;
-	apdu.datalen = pinlen;
-	apdu.data = pinbuffer;
-	apdu.resplen = 0;
-	apdu.le = 0;
+	dnie_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x20, 0x00, 0x00, 0, pinlen,
+					NULL, 0, pinbuffer, pinlen);
 
 	/* and send to card throught virtual channel */
-	res = dnie_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(card->ctx, res, "VERIFY APDU Transmit fail");
+	res = sc_transmit_apdu(card, &apdu);
+	if (res != SC_SUCCESS) {
+		LOG_TEST_RET(card->ctx, res, "VERIFY APDU Transmit fail");
+	}
 
 	/* check response and if requested setup tries_left */
 	if (tries_left != NULL) {	/* returning tries_left count is requested */
@@ -1930,14 +2210,14 @@ static int dnie_pin_verify(struct sc_card *card,
 	}
 	res = dnie_check_sw(card, apdu.sw1, apdu.sw2);	/* not a pinerr: parse result */
 
-	/* the end: a bit of Mister Proper and return */
-	memset(&apdu, 0, sizeof(apdu));	/* clear buffer */
-	data->apdu = NULL;
+	/* ensure that secure channel is established after a PIN channel in 3.0 */
+	if (card->atr.value[15] >= DNIE_30_VERSION) {
+		sc_log(card->ctx, "DNIe 3.0 detected => re-establish secure channel");
+		dnie_change_cwa_provider_to_secure(card);
+		res = cwa_create_secure_channel(card, GET_DNIE_PRIV_DATA(card)->cwa_provider, CWA_SM_ON);
+	}
+
 	LOG_FUNC_RETURN(card->ctx, res);
-#else
-    LOG_TEST_RET(card->ctx, SC_ERROR_NOT_SUPPORTED, "built without support of SM and External Authentication");
-    return SC_ERROR_NOT_SUPPORTED;
-#endif
 }
 
 /* pin_cmd: verify/change/unblock command; optionally using the
