@@ -28,11 +28,16 @@
 #include "asn1.h"
 #include "cardctl.h"
 
+#define GEMSAFEV1_ALG_REF_FREEFORM	0x12
+#define GEMSAFEV3_ALG_REF_FREEFORM	0x02
+#define GEMSAFEV3_ALG_REF_SHA1		0x12
+#define GEMSAFEV3_ALG_REF_SHA256	0x42
+
 static struct sc_card_operations gemsafe_ops;
 static struct sc_card_operations *iso_ops = NULL;
 
 static struct sc_card_driver gemsafe_drv = {
-	"driver for the Gemplus GemSAFE V1 applet",
+	"Gemalto GemSafe V1 applet",
 	"gemsafeV1",
 	&gemsafe_ops,
 	NULL, 0, NULL
@@ -58,6 +63,8 @@ static struct sc_atr_table gemsafe_atrs[] = {
     {"3B:7D:95:00:00:80:31:80:65:B0:83:11:00:C8:83:00:90:00", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_PTEID, 0, NULL},
     /* Swedish eID card */
     {"3B:7D:96:00:00:80:31:80:65:B0:83:11:00:C8:83:00:90:00", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_SEEID, 0, NULL},
+    /* European Patent Office epoline card*/
+    {"3b:7d:96:00:00:80:31:80:65:b0:83:02:01:f3:83:00:90:00", NULL, NULL, SC_CARD_TYPE_GEMSAFEV1_SEEID, 0, NULL},
     {NULL, NULL, NULL, 0, 0, NULL}
 };
 
@@ -92,7 +99,7 @@ static int get_conf_aid(sc_card_t *card, u8 *aid, size_t *len)
 	for (i = 0; ctx->conf_blocks[i] != NULL; i++) {
 		blocks = scconf_find_blocks(ctx->conf, ctx->conf_blocks[i],
 						"card", "gemsafeV1");
-		if (blocks[0] != NULL)
+		if (blocks != NULL && blocks[0] != NULL)
 			conf_block = blocks[0];
 		free(blocks);
 	}
@@ -199,12 +206,30 @@ static int gemsafe_init(struct sc_card *card)
 		flags |= SC_ALGORITHM_ONBOARD_KEY_GEN;
 		flags |= SC_ALGORITHM_RSA_HASH_NONE;
 
+		/* GemSAFE V3 cards support SHA256 */
+		if (card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
+		    card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID)
+			flags |= SC_ALGORITHM_RSA_HASH_SHA256;
+
 		_sc_card_add_rsa_alg(card,  512, flags, 0);
 		_sc_card_add_rsa_alg(card,  768, flags, 0);
 		_sc_card_add_rsa_alg(card, 1024, flags, 0);
 		_sc_card_add_rsa_alg(card, 2048, flags, 0);
+
+		/* fake algorithm to persuade register_mechanisms()
+		 * to register these hashes */
+		if (card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
+		    card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) {
+			flags  = SC_ALGORITHM_RSA_HASH_SHA1;
+			flags |= SC_ALGORITHM_RSA_HASH_MD5;
+			flags |= SC_ALGORITHM_RSA_HASH_MD5_SHA1;
+			flags |= SC_ALGORITHM_RSA_HASH_RIPEMD160;
+
+			_sc_card_add_rsa_alg(card,  512, flags, 0);
+		}
 	}
 
+	card->caps |= SC_CARD_CAP_ISO7816_PIN_INFO;
 	card->drv_data = exdata;
 
 	return 0;
@@ -363,15 +388,21 @@ static u8 gemsafe_flags2algref(struct sc_card *card, const struct sc_security_en
 	u8 ret = 0;
 
 	if (env->operation == SC_SEC_OPERATION_SIGN) {
-		if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PKCS1)
+		if (env->algorithm_flags & SC_ALGORITHM_RSA_HASH_SHA256)
+			ret = GEMSAFEV3_ALG_REF_SHA256;
+		else if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PKCS1)
 			ret = (card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
-			       card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) ? 0x02 : 0x12;
+			       card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) ?
+			      GEMSAFEV3_ALG_REF_FREEFORM :
+			      GEMSAFEV1_ALG_REF_FREEFORM;
 		else if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_ISO9796)
 			ret = 0x11;
 	} else if (env->operation == SC_SEC_OPERATION_DECIPHER) {
 		if (env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PKCS1)
 			ret = (card->type == SC_CARD_TYPE_GEMSAFEV1_PTEID ||
-			       card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) ? 0x02 : 0x12;
+			       card->type == SC_CARD_TYPE_GEMSAFEV1_SEEID) ?
+			      GEMSAFEV3_ALG_REF_FREEFORM :
+			      GEMSAFEV1_ALG_REF_FREEFORM;
 	}
 
 	return ret;
@@ -429,8 +460,11 @@ static int gemsafe_compute_signature(struct sc_card *card, const u8 * data,
 
 	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_VERBOSE);
 
+	/* the card can sign 36 bytes of free form data */
 	if (data_len > 36) {
-		sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "error: input data too long: %lu bytes\n", data_len);
+		sc_debug(ctx, SC_LOG_DEBUG_NORMAL,
+			 "error: input data too long: %"SC_FORMAT_LEN_SIZE_T"u bytes\n",
+			 data_len);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 
@@ -532,146 +566,6 @@ static int gemsafe_get_challenge(sc_card_t *card, u8 *rnd, size_t len)
 	return r;
 }
 
-static int gemsafe_build_pin_apdu(struct sc_card *card,
-		struct sc_apdu *apdu,
-		struct sc_pin_cmd_data *data)
-{
-	static u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
-	int r, len = 0, pad = 0, use_pin_pad = 0, ins, p1 = 0;
-
-	switch (data->pin_type) {
-	case SC_AC_CHV:
-		break;
-	default:
-		return SC_ERROR_INVALID_ARGUMENTS;
-	}
-
-	if (data->flags & SC_PIN_CMD_NEED_PADDING)
-		pad = 1;
-	if (data->flags & SC_PIN_CMD_USE_PINPAD)
-		use_pin_pad = 1;
-
-	data->pin1.offset = 5;
-
-	switch (data->cmd) {
-	case SC_PIN_CMD_VERIFY:
-		ins = 0x20;
-		if ((r = sc_build_pin(sbuf, sizeof(sbuf), &data->pin1, pad)) < 0)
-			return r;
-		len = r;
-		break;
-	case SC_PIN_CMD_CHANGE:
-		ins = 0x24;
-		if (data->pin1.len != 0 || use_pin_pad) {
-			if ((r = sc_build_pin(sbuf, sizeof(sbuf), &data->pin1, pad)) < 0)
-				return r;
-			len += r;
-		} else {
-			/* implicit test */
-			p1 = 1;
-		}
-
-		data->pin2.offset = data->pin1.offset + len;
-		if ((r = sc_build_pin(sbuf+len, sizeof(sbuf)-len, &data->pin2, pad)) < 0)
-			return r;
-		len += r;
-		break;
-	case SC_PIN_CMD_UNBLOCK:
-		ins = 0x2C;
-		if (data->pin1.len != 0 || use_pin_pad) {
-			if ((r = sc_build_pin(sbuf, sizeof(sbuf), &data->pin1, pad)) < 0)
-				return r;
-			len += r;
-		} else {
-			p1 |= 0x02;
-		}
-
-		if (data->pin2.len != 0 || use_pin_pad) {
-			data->pin2.offset = data->pin1.offset + len;
-			if ((r = sc_build_pin(sbuf+len, sizeof(sbuf)-len, &data->pin2, pad)) < 0)
-				return r;
-			len += r;
-		} else {
-			p1 |= 0x01;
-		}
-		break;
-	default:
-		return SC_ERROR_NOT_SUPPORTED;
-	}
-
-	sc_format_apdu(card, apdu, SC_APDU_CASE_3_SHORT,
-				ins, p1, data->pin_reference);
-
-	apdu->lc = len;
-	apdu->datalen = len;
-	apdu->data = sbuf;
-	apdu->resplen = 0;
-
-	return 0;
-}
-
-static int gemsafe_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data,
-			   int *tries_left)
-{
-	struct sc_apdu local_apdu, *apdu;
-	int r;
-
-	if (tries_left)
-		*tries_left = -1;
-
-	/* See if we've been called from another card driver, which is
-	 * passing an APDU to us (this allows to write card drivers
-	 * whose PIN functions behave "mostly like ISO" except in some
-	 * special circumstances.
-	 */
-	if (data->apdu == NULL) {
-		r = gemsafe_build_pin_apdu(card, &local_apdu, data);
-		if (r < 0)
-			return r;
-		data->apdu = &local_apdu;
-	}
-	apdu = data->apdu;
-
-	if (!(data->flags & SC_PIN_CMD_USE_PINPAD)) {
-		/* Transmit the APDU to the card */
-		r = sc_transmit_apdu(card, apdu);
-
-		/* Clear the buffer - it may contain pins */
-		memset((void *) apdu->data, 0, apdu->datalen);
-	} else {
-		/* Call the reader driver to collect
-		 * the PIN and pass on the APDU to the card */
-		if (data->pin1.offset == 0) {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-				"Card driver didn't set PIN offset");
-			return SC_ERROR_INVALID_ARGUMENTS;
-		}
-		if (card->reader
-		 && card->reader->ops
-		 && card->reader->ops->perform_verify) {
-			r = card->reader->ops->perform_verify(card->reader, data);
-			/* sw1/sw2 filled in by reader driver */
-		} else {
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
-				"Card reader driver does not support "
-				"PIN entry through reader key pad");
-			r = SC_ERROR_NOT_SUPPORTED;
-		}
-	}
-
-	/* Don't pass references to local variables up to the caller. */
-	if (data->apdu == &local_apdu)
-		data->apdu = NULL;
-
-	SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
-	if (apdu->sw1 == 0x63) {
-		if ((apdu->sw2 & 0xF0) == 0xC0 && tries_left != NULL)
-			*tries_left = apdu->sw2 & 0x0F;
-		return SC_ERROR_PIN_CODE_INCORRECT;
-	}
-	return sc_check_sw(card, apdu->sw1, apdu->sw2);
-}
-
 static struct sc_card_driver *sc_get_driver(void)
 {
 	struct sc_card_driver *iso_drv = sc_get_iso7816_driver();
@@ -690,7 +584,7 @@ static struct sc_card_driver *sc_get_driver(void)
 	gemsafe_ops.compute_signature    = gemsafe_compute_signature;
 	gemsafe_ops.get_challenge 		 = gemsafe_get_challenge;
 	gemsafe_ops.process_fci	= gemsafe_process_fci;
-	gemsafe_ops.pin_cmd		 = gemsafe_pin_cmd;
+	gemsafe_ops.pin_cmd		 = iso_ops->pin_cmd;
 
 	return &gemsafe_drv;
 }
