@@ -43,10 +43,24 @@
 #define LOAD_KEY_EC_PRIVATE	0x97
 #define LOAD_KEY_EC_PUBLIC	0x96
 
+#define LOAD_KEY_SYMMETRIC	0xa0
+
 #define MYEID_STATE_CREATION	0x01
 #define MYEID_STATE_ACTIVATED	0x07
 
-#define MYEID_ECC_SUPPORT
+#define MYEID_CARD_NAME_MAX_LEN		100
+
+/* The following flags define the features supported by the card currently in use.
+  They are used in 'card_supported_features' field in myeid_card_caps struct */
+#define MYEID_CARD_CAP_RSA		0x01
+#define MYEID_CARD_CAP_3DES		0x02
+#define MYEID_CARD_CAP_AES		0x04
+#define MYEID_CARD_CAP_ECC		0x08
+#define MYEID_CARD_CAP_GRIDPIN		0x10
+#define MYEID_CARD_CAP_PIV_EMU		0x20
+
+static const char *myeid_card_name = "MyEID";
+static char card_name_buf[MYEID_CARD_NAME_MAX_LEN];
 
 static struct sc_card_operations myeid_ops;
 static struct sc_card_driver myeid_drv = {
@@ -63,6 +77,8 @@ static const char *myeid_atrs[] = {
 	"3B:F5:18:00:00:81:31:FE:45:4D:79:45:49:44:9A",
 	"3B:85:80:01:4D:79:45:49:44:78",
 	"3B:89:80:01:09:38:33:B1:4D:79:45:49:44:4C",
+	"3B:F5:96:00:00:80:31:FE:45:4D:79:45:49:44:15", /* Infineon's chip */
+	"3B:F5:96:00:00:81:31:FE:45:4D:79:45:49:44:14",
 	NULL
 };
 
@@ -77,20 +93,30 @@ typedef struct myeid_private_data {
 	const struct sc_security_env* sec_env;
 } myeid_private_data_t;
 
+typedef struct myeid_card_caps {
+	unsigned char card_caps_ver;
+	unsigned short card_supported_features;
+	unsigned short max_rsa_key_length;
+	unsigned short max_des_key_length;
+	unsigned short max_aes_key_length;
+	unsigned short max_ecc_key_length;
+} myeid_card_caps_t;
+
 static struct myeid_supported_ec_curves {
-        char *curve_name;
-        struct sc_object_id curve_oid;
-        size_t size;
+	char *curve_name;
+	struct sc_object_id curve_oid;
+	size_t size;
 } ec_curves[] = {
 	{"secp192r1", {{1, 2, 840, 10045, 3, 1, 1, -1}},192},
 	/* {"secp224r1", {{1, 3, 132, 0, 33, -1}},		224}, */
 	{"secp256r1", {{1, 2, 840, 10045, 3, 1, 7, -1}},256},
-	/* {"secp384r1", {{1, 3, 132, 0, 34, -1}},		384}, */
-	/* {"secp521r1", {{1, 3, 132, 0, 35, -1}},		521}, */
+	{"secp384r1", {{1, 3, 132, 0, 34, -1}},		384},
+	{"secp521r1", {{1, 3, 132, 0, 35, -1}},		521},
 	{NULL, {{-1}}, 0},
 };
 
 static int myeid_get_info(struct sc_card *card, u8 *rbuf, size_t buflen);
+static int myeid_get_card_caps(struct sc_card *card, myeid_card_caps_t* card_caps);
 
 static int myeid_match_card(struct sc_card *card)
 {
@@ -119,15 +145,19 @@ static int myeid_match_card(struct sc_card *card)
 
 static int myeid_init(struct sc_card *card)
 {
-	unsigned long flags = 0,
-	          ext_flags = 0;
+	unsigned long flags = 0, ext_flags = 0;
 	myeid_private_data_t *priv;
 	u8 appletInfo[20];
 	size_t appletInfoLen;
 	int r;
+	myeid_card_caps_t card_caps;
 
 	LOG_FUNC_CALLED(card->ctx);
+
+	card->name = myeid_card_name;
+
 	priv = calloc(1, sizeof(myeid_private_data_t));
+
 	if (!priv)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 
@@ -152,25 +182,54 @@ static int myeid_init(struct sc_card *card)
 	_sc_card_add_rsa_alg(card, 1024, flags, 0);
 	_sc_card_add_rsa_alg(card, 1536, flags, 0);
 	_sc_card_add_rsa_alg(card, 2048, flags, 0);
+	
+	memset(&card_caps, 0, sizeof(myeid_card_caps_t));
+	card_caps.max_ecc_key_length = 256;
+	card_caps.max_rsa_key_length = 2048;
 
-#ifdef MYEID_ECC_SUPPORT
+	if (card->version.fw_major >= 40) {
+	    /* Since 4.0, we can query available algorithms and key sizes.
+	     * Since 3.5.0 RSA up to 2048 and ECC up to 256 are always supported, so we check only max ECC key length. */
+	    r = myeid_get_card_caps(card, &card_caps);
+
+	    if (r != SC_SUCCESS) {
+		sc_log(card->ctx, "Failed to get card capabilities. Using default max ECC key length 256.");
+	    }
+	}
 
 	/* show ECC algorithms if the applet version of the inserted card supports them */
-	if ((card->version.fw_major == 3 && card->version.fw_minor > 5) ||
-			card->version.fw_major >= 4)   {
+	if (card->version.fw_major >= 35) {
 		int i;
 
 		flags = SC_ALGORITHM_ECDSA_RAW | SC_ALGORITHM_ECDH_CDH_RAW | SC_ALGORITHM_ONBOARD_KEY_GEN;
 		flags |= SC_ALGORITHM_ECDSA_HASH_NONE | SC_ALGORITHM_ECDSA_HASH_SHA1;
 		ext_flags = SC_ALGORITHM_EXT_EC_NAMEDCURVE | SC_ALGORITHM_EXT_EC_UNCOMPRESES;
 
-		for (i=0; ec_curves[i].curve_name != NULL; i++)
-			_sc_card_add_ec_alg(card,  ec_curves[i].size, flags, ext_flags, &ec_curves[i].curve_oid);
+		for (i=0; ec_curves[i].curve_name != NULL; i++) {
+			if (card_caps.max_ecc_key_length >= ec_curves[i].size)
+				_sc_card_add_ec_alg(card, ec_curves[i].size, flags, ext_flags, &ec_curves[i].curve_oid);
+		}
+	}
+
+	/* show supported symmetric algorithms */
+	flags = 0;
+	if (card_caps.card_supported_features & MYEID_CARD_CAP_3DES) {
+		if (card_caps.max_des_key_length >= 56)
+			_sc_card_add_symmetric_alg(card, SC_ALGORITHM_DES, 56, flags);
+		if (card_caps.max_des_key_length >= 128)
+			_sc_card_add_symmetric_alg(card, SC_ALGORITHM_3DES, 128, flags);
+		if (card_caps.max_des_key_length >= 192)
+			_sc_card_add_symmetric_alg(card, SC_ALGORITHM_3DES, 192, flags);
+	}
+	if (card_caps.card_supported_features & MYEID_CARD_CAP_AES) {
+		if (card_caps.max_aes_key_length >= 128)
+			_sc_card_add_symmetric_alg(card, SC_ALGORITHM_AES, 128, flags);
+		if (card_caps.max_aes_key_length >= 256)
+			_sc_card_add_symmetric_alg(card, SC_ALGORITHM_AES, 256, flags);
 	}
-#endif
 
 	/* State that we have an RNG */
-	card->caps |= SC_CARD_CAP_RNG;
+	card->caps |= SC_CARD_CAP_RNG | SC_CARD_CAP_ISO7816_PIN_INFO;
 
 	card->max_recv_size = 255;
 	card->max_send_size = 255;
@@ -263,16 +322,10 @@ static void parse_sec_attr(struct sc_file *file, const u8 *buf, size_t len)
 static int myeid_select_file(struct sc_card *card, const struct sc_path *in_path,
 		struct sc_file **file)
 {
-	struct sc_file *dummy_file = NULL;
 	int r;
 
 	LOG_FUNC_CALLED(card->ctx);
-	r = iso_ops->select_file(card, in_path, &dummy_file);
-
-	if (file)
-		*file = dummy_file;
-	else  if (dummy_file)
-		sc_file_free(dummy_file);
+	r = iso_ops->select_file(card, in_path, file);
 
 	if (r == 0 && file != NULL && *file != NULL)
 		parse_sec_attr(*file, (*file)->sec_attr, (*file)->sec_attr_len);
@@ -302,7 +355,7 @@ static int myeid_list_files(struct sc_card *card, u8 *buf, size_t buflen)
 }
 
 static int myeid_process_fci(struct sc_card *card, struct sc_file *file,
-	         const u8 *buf, size_t buflen)
+		const u8 *buf, size_t buflen)
 {
 	myeid_private_data_t *priv = (myeid_private_data_t *) card->drv_data;
 	size_t taglen = 0;
@@ -324,8 +377,8 @@ static int myeid_process_fci(struct sc_card *card, struct sc_file *file,
 	}
 	if(file->sec_attr_len >= 3)
 	{
-	    sc_log(card->ctx, "id (%X) sec_attr (%X %X %X)", file->id,
-	             file->sec_attr[0],file->sec_attr[1],file->sec_attr[2]);
+		sc_log(card->ctx, "id (%X) sec_attr (%X %X %X)", file->id,
+			file->sec_attr[0],file->sec_attr[1],file->sec_attr[2]);
 	}
 	tag = sc_asn1_find_tag(NULL, buf, buflen, 0x8A, &taglen);
 	if (tag != NULL && taglen > 0)
@@ -394,11 +447,13 @@ static int encode_file_structure(sc_card_t *card, const sc_file_t *file,
 		buf[17] = file->sec_attr[2];
 
 		sc_log(card->ctx, "id (%X), sec_attr %X %X %X", file->id,
-	                     file->sec_attr[0],file->sec_attr[1],file->sec_attr[2]);
+				file->sec_attr[0],file->sec_attr[1],file->sec_attr[2]);
 	}
 	else   {
 		delete = sc_file_get_acl_entry(file, SC_AC_OP_DELETE);
 
+		sc_log(card->ctx, "id (%X), type (%X)", file->id, file->type);
+
 		switch (file->type) {
 		case SC_FILE_TYPE_WORKING_EF:
 
@@ -524,42 +579,6 @@ static int myeid_delete_file(struct sc_card *card, const struct sc_path *path)
 	LOG_FUNC_RETURN(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2));
 }
 
-static int myeid_pin_info(sc_card_t *card, struct sc_pin_cmd_data *data,
-			   int *tries_left)
-{
-	sc_apdu_t apdu;
-	int r;
-
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x20, 0x00, data->pin_reference);
-
-	r = sc_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
-
-	r =  sc_check_sw(card, apdu.sw1, apdu.sw2);
-
-	if (r == SC_ERROR_PIN_CODE_INCORRECT) {
-		data->pin1.tries_left = apdu.sw2 & 0xF;
-		r = SC_SUCCESS;
-	} else if (r == SC_ERROR_AUTH_METHOD_BLOCKED) {
-		data->pin1.tries_left = 0;
-		r = SC_SUCCESS;
-	}
-	LOG_TEST_RET(card->ctx, r, "Check SW error");
-
-	if (r == SC_SUCCESS)
-	{
-		data->pin1.pad_length = data->pin2.pad_length = 8;
-	        data->pin1.pad_char = data->pin2.pad_char = 0xFF;
-	}
-
-
-	if (tries_left != NULL) {
-		*tries_left = data->pin1.tries_left;
-	}
-
-	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
-}
-
 static int myeid_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 			 int *tries_left)
 {
@@ -567,13 +586,8 @@ static int myeid_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 
 	LOG_FUNC_CALLED(card->ctx);
 
-	if (data->cmd == SC_PIN_CMD_GET_INFO)
-	{
-	        return myeid_pin_info(card, data, tries_left);
-	}
-
 	sc_log(card->ctx, "ref (%d), pin1 len(%d), pin2 len (%d)\n",
-	              data->pin_reference, data->pin1.len, data->pin2.len);
+			data->pin_reference, data->pin1.len, data->pin2.len);
 
 	if(data->pin1.len > 8 || data->pin2.len > 8)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_PIN_LENGTH);
@@ -685,12 +699,12 @@ static int myeid_set_security_env_ec(sc_card_t *card, const sc_security_env_t *e
 
 	if (env->flags & SC_SEC_ENV_KEY_REF_ASYMMETRIC)
 	{
-		sc_log(card->ctx, "asymmetric keyref not supported.\n");
+		sc_log(card->ctx, "asymmetric keyref not supported.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 	if (se_num > 0)
 	{
-		sc_log(card->ctx, "restore security environment not supported.\n");
+		sc_log(card->ctx, "restore security environment not supported.");
 		return SC_ERROR_NOT_SUPPORTED;
 	}
 
@@ -698,13 +712,17 @@ static int myeid_set_security_env_ec(sc_card_t *card, const sc_security_env_t *e
 	switch (env->operation)
 	{
 	case SC_SEC_OPERATION_DECIPHER:
-	        sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Decipher operation is not supported with EC keys.\n");
+		sc_log(card->ctx, "Decipher operation is not supported with EC keys.");
 		return SC_ERROR_NOT_SUPPORTED;
 		break;
 	case SC_SEC_OPERATION_SIGN:
 		apdu.p1 = 0x41;
 		apdu.p2 = 0xB6;
 		break;
+	case SC_SEC_OPERATION_DERIVE:
+		apdu.p1 = 0x41;
+		apdu.p2 = 0xA4;
+		break;
 	default:
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -787,16 +805,11 @@ static int myeid_set_security_env(struct sc_card *card,
 
 			return myeid_set_security_env_rsa(card, &tmp, se_num);
 		}
-	else if (tmp.algorithm == SC_ALGORITHM_EC)
-	{
-#ifdef MYEID_ECC_SUPPORT
+		else if (tmp.algorithm == SC_ALGORITHM_EC)
+		{
 			tmp.algorithm_ref = 0x04;
 			tmp.algorithm_flags = 0;
 			return myeid_set_security_env_ec(card, &tmp, se_num);
-#else
-			sc_log(ctx, "Elliptic curves are not supported in this version.");
-			return SC_ERROR_NOT_SUPPORTED;
-#endif
 		}
 		else
 		{
@@ -815,26 +828,57 @@ myeid_convert_ec_signature(struct sc_context *ctx, size_t s_len, unsigned char *
 	unsigned char *buf;
 	size_t buflen;
 	int r;
+	size_t len_size = 1;
+	size_t sig_len = 0;
 
-	assert(data && datalen);
+	assert(data && datalen && datalen > 3);
 
-	if (*data != 0x30 || *(data + 1) != (datalen - 2) || *(data + 2) != 0x02)
+	/*
+	 *	When validating the signature data, we have to consider that length of the signature
+	 *	can be encoded in either one or two bytes depending on key size. With 521 bit keys
+	 *	length of the structure takes two bytes.
+	 */
+
+	if (*data != 0x30)
 		return SC_ERROR_INVALID_DATA;
 
-	buf = calloc(1, (s_len + 7)/8*2);
-	if (!buf)
-		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+	if ((*(data + 1) & 0x80) == 0x80)
+		len_size += *(data + 1) & 0x7F;
+
+	if (len_size == 1)
+	    sig_len = *(data + 1);
+	else if (len_size == 2)
+	    sig_len = *(data + 2);
+	else if (len_size == 3)
+	{
+	    sig_len = *(data + 2) | (*data + 3) << 8;
+	}
+	else
+	    return SC_ERROR_INVALID_DATA;
+
+	if (*(data + 1 + len_size) != 0x02)		/* Verify that it is an INTEGER */
+
+	if (sig_len != (datalen - len_size - 1))	/* validate size of the DER structure */
+	    return SC_ERROR_INVALID_DATA;
+
+	/* test&fail early */
 	buflen = (s_len + 7)/8*2;
-
-	r = sc_asn1_sig_value_sequence_to_rs(ctx, data, datalen, buf, buflen);
-	if (r < 0)
-		free(buf);
-        LOG_TEST_RET(ctx, r, "Failed to cenvert Sig-Value to the raw RS format");
-
 	if (buflen > datalen)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
 
+	buf = calloc(1, buflen);
+	if (!buf)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+
+	r = sc_asn1_sig_value_sequence_to_rs(ctx, data, datalen, buf, buflen);
+	if (r < 0) {
+		free(buf);
+		sc_log(ctx, "Failed to convert Sig-Value to the raw RS format");
+		return r;
+	}
+
 	memmove(data, buf, buflen);
+	free(buf);
 	return buflen;
 }
 
@@ -843,12 +887,15 @@ static int
 myeid_compute_signature(struct sc_card *card, const u8 * data, size_t datalen,
 		u8 * out, size_t outlen)
 {
-	struct sc_context *ctx = card->ctx;
+	struct sc_context *ctx;
 	struct sc_apdu apdu;
 	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
 	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
 	struct myeid_private_data* priv;
 	int r;
+	size_t field_length = 0;
+	size_t pad_chars = 0;
+
 
 	assert(card != NULL && data != NULL && out != NULL);
 	ctx = card->ctx;
@@ -857,7 +904,19 @@ myeid_compute_signature(struct sc_card *card, const u8 * data, size_t datalen,
 	priv = (myeid_private_data_t*) card->drv_data;
 	sc_log(ctx, "key type %i, key length %i", priv->sec_env->algorithm, priv->sec_env->algorithm_ref);
 
-	if (datalen > 256)
+	if (priv->sec_env->algorithm == SC_ALGORITHM_EC ) {
+
+	    field_length = priv->sec_env->algorithm_ref;
+
+	    /* pad with zeros if needed */
+		if (datalen < (field_length + 7) / 8 ) {
+			pad_chars = ((field_length + 7) / 8) - datalen;
+
+			memset(sbuf, 0, pad_chars);
+		}
+	}
+
+	if ((datalen + pad_chars) > 256)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
 
 	/* INS: 0x2A  PERFORM SECURITY OPERATION
@@ -874,16 +933,16 @@ myeid_compute_signature(struct sc_card *card, const u8 * data, size_t datalen,
 		apdu.datalen = datalen - 1;
 	}
 	else   {
-		memcpy(sbuf, data, datalen);
-		apdu.lc = datalen;
-		apdu.datalen = datalen;
+		memcpy(sbuf + pad_chars, data, datalen);
+		apdu.lc = datalen + pad_chars;
+		apdu.datalen = datalen + pad_chars;
 	}
 
 	apdu.data = sbuf;
 	r = sc_transmit_apdu(card, &apdu);
 	LOG_TEST_RET(ctx, r, "APDU transmit failed");
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-        LOG_TEST_RET(ctx, r, "compute_signature failed");
+	LOG_TEST_RET(ctx, r, "compute_signature failed");
 
 	if (priv->sec_env->algorithm == SC_ALGORITHM_EC)   {
 		r = myeid_convert_ec_signature(ctx, priv->sec_env->algorithm_ref, apdu.resp, apdu.resplen);
@@ -910,22 +969,43 @@ int myeid_ecdh_derive(struct sc_card *card, const u8* pubkey, size_t pubkey_len,
 	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
 
 	int r;
+	size_t ext_len_bytes;
 
-	sc_format_apdu(card, &apdu,
-	        SC_APDU_CASE_4_SHORT,
-	        0x86, 0x00, 0x00);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0x86, 0x00, 0x00);
 
 	apdu.resp = rbuf;
 	apdu.resplen = sizeof(rbuf);
 
-	/* Fill in "Data objects in dynamic authentication template (tag 0x7C) structure */
-	sbuf[0] = 0x7C;
-	sbuf[1] = pubkey_len + 4;
-	sbuf[2] = 0x85;
-	sbuf[3] = pubkey_len;
-	memcpy(&sbuf[4], pubkey, pubkey_len);
+	/* Fill in "Data objects in dynamic authentication template" (tag 0x7C) structure
+	*
+	* TODO: encode the structure using OpenSC's ASN1-functions.
+	*
+	*  Size of the structure depends on key length. With 521 bit keys two bytes are needed for defining length of a point.
+	*/
 
-	apdu.lc = pubkey_len + 4;
+	sbuf[0] = 0x7C;
+	ext_len_bytes = 0;
+
+	if (pubkey_len > 127)
+	{
+		sbuf[1] = 0x81;
+		sbuf[2] = (u8) (pubkey_len + 3);
+		sbuf[3] = 0x85;
+		sbuf[4] = 0x81;
+		sbuf[5] = (u8) (pubkey_len);
+		ext_len_bytes = 2;
+	}
+	else
+	{
+		sbuf[1] = pubkey_len + 2;
+		sbuf[2] = 0x85;
+		sbuf[3] = pubkey_len;
+	}
+
+	memcpy(&sbuf[4 + ext_len_bytes], pubkey, pubkey_len);
+
+	apdu.lc = pubkey_len + 4 + ext_len_bytes;
+	apdu.le = pubkey_len / 2;
 	apdu.datalen = apdu.lc;
 	apdu.data = sbuf;
 
@@ -938,13 +1018,13 @@ int myeid_ecdh_derive(struct sc_card *card, const u8* pubkey, size_t pubkey_len,
 
 	if (outlen < apdu.resplen)
 	{
-	        r = SC_ERROR_BUFFER_TOO_SMALL;
-	        LOG_TEST_RET(card->ctx, r, "Buffer too small to hold shared secret.");
+		r = SC_ERROR_BUFFER_TOO_SMALL;
+		LOG_TEST_RET(card->ctx, r, "Buffer too small to hold shared secret.");
 	}
 
 	memcpy(out, rbuf, apdu.resplen);
 
-	LOG_FUNC_RETURN(card->ctx, r);
+	LOG_FUNC_RETURN(card->ctx, apdu.resplen);
 }
 
 
@@ -965,23 +1045,23 @@ static int myeid_decipher(struct sc_card *card, const u8 * crgram,
 	priv = (myeid_private_data_t*) card->drv_data;
 
 	if (priv->sec_env && priv->sec_env->algorithm == SC_ALGORITHM_EC
-	        && priv->sec_env->operation == SC_SEC_OPERATION_DERIVE
-	        && priv->sec_env->algorithm_flags & SC_ALGORITHM_ECDSA_RAW)
+		&& priv->sec_env->operation == SC_SEC_OPERATION_DERIVE
+		&& priv->sec_env->algorithm_flags & SC_ALGORITHM_ECDSA_RAW)
 	{
-	        r = myeid_ecdh_derive(card, crgram, crgram_len, out, outlen);
-	        priv->sec_env = NULL; /* clear after operation */
-	        LOG_FUNC_RETURN(card->ctx, r);
+		r = myeid_ecdh_derive(card, crgram, crgram_len, out, outlen);
+		priv->sec_env = NULL; /* clear after operation */
+		LOG_FUNC_RETURN(card->ctx, r);
 	}
 
 	if (crgram_len > 256)
-	        LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 
 	/* INS: 0x2A  PERFORM SECURITY OPERATION
 	    * P1:  0x80  Resp: Plain value
 	    * P2:  0x86  Cmd: Padding indicator byte followed by cryptogram */
 	sc_format_apdu(card, &apdu,
-	        (crgram_len < 256) ? SC_APDU_CASE_4_SHORT : SC_APDU_CASE_3_SHORT,
-	        0x2A, 0x80, 0x86);
+		(crgram_len < 256) ? SC_APDU_CASE_4_SHORT : SC_APDU_CASE_3_SHORT,
+		0x2A, 0x80, 0x86);
 
 	apdu.resp = rbuf;
 	apdu.resplen = sizeof(rbuf);
@@ -989,17 +1069,17 @@ static int myeid_decipher(struct sc_card *card, const u8 * crgram,
 
 	if (crgram_len == 256)
 	{
-	        apdu.le = 0;
-	        /* padding indicator byte, 0x81 = first half of 2048 bit cryptogram */
-	        sbuf[0] = 0x81;
-	        memcpy(sbuf + 1, crgram, crgram_len / 2);
-	        apdu.lc = crgram_len / 2 + 1;
+		apdu.le = 0;
+		/* padding indicator byte, 0x81 = first half of 2048 bit cryptogram */
+		sbuf[0] = 0x81;
+		memcpy(sbuf + 1, crgram, crgram_len / 2);
+		apdu.lc = crgram_len / 2 + 1;
 	}
 	else
 	{
-	        sbuf[0] = 0; /* padding indicator byte, 0x00 = No further indication */
-	        memcpy(sbuf + 1, crgram, crgram_len);
-	        apdu.lc = crgram_len + 1;
+		sbuf[0] = 0; /* padding indicator byte, 0x00 = No further indication */
+		memcpy(sbuf + 1, crgram, crgram_len);
+		apdu.lc = crgram_len + 1;
 	}
 
 	apdu.datalen = apdu.lc;
@@ -1009,39 +1089,39 @@ static int myeid_decipher(struct sc_card *card, const u8 * crgram,
 	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
 	{
-	        if (crgram_len == 256)
-	        {
-	                sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT,
-	                        0x2A, 0x80, 0x86);
-	                apdu.resp = rbuf;
-	                apdu.resplen = sizeof(rbuf);
-	                apdu.le = crgram_len;
-	                /* padding indicator byte,
-	                    * 0x82 = Second half of 2048 bit cryptogram */
-	                sbuf[0] = 0x82;
-	                memcpy(sbuf + 1, crgram + crgram_len / 2, crgram_len / 2);
-	                apdu.lc = crgram_len / 2 + 1;
-	                apdu.datalen = apdu.lc;
-	                apdu.data = sbuf;
+		if (crgram_len == 256)
+		{
+			sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT,
+				0x2A, 0x80, 0x86);
+			apdu.resp = rbuf;
+			apdu.resplen = sizeof(rbuf);
+			apdu.le = crgram_len;
+			/* padding indicator byte,
+			    * 0x82 = Second half of 2048 bit cryptogram */
+			sbuf[0] = 0x82;
+			memcpy(sbuf + 1, crgram + crgram_len / 2, crgram_len / 2);
+			apdu.lc = crgram_len / 2 + 1;
+			apdu.datalen = apdu.lc;
+			apdu.data = sbuf;
 
-	                r = sc_transmit_apdu(card, &apdu);
+			r = sc_transmit_apdu(card, &apdu);
 
-	                LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+			LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 
-	                if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
-	                {
-	                        int len = apdu.resplen > outlen ? outlen : apdu.resplen;
-	                        memcpy(out, apdu.resp, len);
-	                        LOG_FUNC_RETURN(card->ctx, len);
-	                }
-	        }
-	        else
-	        {
-	                int len = apdu.resplen > outlen ? outlen : apdu.resplen;
+			if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00)
+			{
+				int len = apdu.resplen > outlen ? outlen : apdu.resplen;
+				memcpy(out, apdu.resp, len);
+				LOG_FUNC_RETURN(card->ctx, len);
+			}
+		}
+		else
+		{
+			int len = apdu.resplen > outlen ? outlen : apdu.resplen;
 
-	                memcpy(out, apdu.resp, len);
-	                LOG_FUNC_RETURN(card->ctx, len);
-	        }
+			memcpy(out, apdu.resp, len);
+			LOG_FUNC_RETURN(card->ctx, len);
+		}
 	}
 	LOG_FUNC_RETURN(card->ctx, sc_check_sw(card, apdu.sw1, apdu.sw2));
 }
@@ -1123,7 +1203,8 @@ static int myeid_loadkey(sc_card_t *card, int mode, u8* value, int value_len)
 
 	if(value    != NULL &&
 	   value[0] != 0x0 &&
-	   mode     != LOAD_KEY_PUBLIC_EXPONENT)
+	   mode     != LOAD_KEY_PUBLIC_EXPONENT &&
+	   mode     != LOAD_KEY_SYMMETRIC)
 		sbuf[len++] = 0x0;
 
 	if(mode == LOAD_KEY_MODULUS && value_len >= 256)
@@ -1263,6 +1344,12 @@ static int myeid_generate_store_key(struct sc_card *card,
 					data->ecpublic_point_len)) >= 0)
 			LOG_FUNC_RETURN(card->ctx, r);
 		}
+		else if(data->key_type == SC_CARDCTL_MYEID_KEY_AES ||
+			data->key_type == SC_CARDCTL_MYEID_KEY_DES) {
+			if((r = myeid_loadkey(card, LOAD_KEY_SYMMETRIC, data->d,
+					data->d_len)) >= 0)
+			LOG_FUNC_RETURN(card->ctx, r);
+		}
 	}
 
 	LOG_FUNC_RETURN(card->ctx, r);
@@ -1294,7 +1381,6 @@ static int myeid_get_info(struct sc_card *card, u8 *rbuf, size_t buflen)
 {
 	sc_apdu_t apdu;
 	int r;
-	char nameBuf[100];
 
 	LOG_FUNC_CALLED(card->ctx);
 
@@ -1311,7 +1397,7 @@ static int myeid_get_info(struct sc_card *card, u8 *rbuf, size_t buflen)
 
 	if (apdu.resplen != 20)
 	{
-		sc_log(card->ctx, "Unexpected response to GET DATA (applet info)\n");
+		sc_log(card->ctx, "Unexpected response to GET DATA (applet info)");
 		return SC_ERROR_INTERNAL;
 	}
 
@@ -1319,9 +1405,10 @@ static int myeid_get_info(struct sc_card *card, u8 *rbuf, size_t buflen)
 	card->version.fw_major = rbuf[5] * 10 + rbuf[6];
 	card->version.fw_minor = rbuf[7];
 	/* add version to name */
-	sprintf((char *) nameBuf, "%s %d.%d.%d", card->name, rbuf[5], rbuf[6], rbuf[7]);
-	card->name = nameBuf;
-	//card->driver->name
+	snprintf(card_name_buf, sizeof(card_name_buf),
+			"%s %d.%d.%d", card->name, rbuf[5], rbuf[6], rbuf[7]);
+	card->name = card_name_buf;
+
 	LOG_FUNC_RETURN(card->ctx, r);
 }
 
@@ -1352,6 +1439,45 @@ static int myeid_get_serialnr(sc_card_t *card, sc_serial_number_t *serial)
 	LOG_FUNC_RETURN(card->ctx, r);
 }
 
+/*
+ Get information of features that the card supports. MyEID 4.x cards are available on different
+ hardware and maximum key sizes cannot be determined simply from the version number anymore.
+ */
+static int myeid_get_card_caps(struct sc_card *card, myeid_card_caps_t* card_caps)
+{
+	sc_apdu_t apdu;
+	int r;
+	unsigned char rbuf[SC_MAX_APDU_BUFFER_SIZE];
+
+	LOG_FUNC_CALLED(card->ctx);
+
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0xca, 0x01, 0xAA);
+	apdu.resp    = rbuf;
+	apdu.resplen = sizeof(myeid_card_caps_t);
+	apdu.le      = sizeof(myeid_card_caps_t);
+
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, r,  "APDU transmit failed");
+
+	if (apdu.sw1 != 0x90 || apdu.sw2 != 0x00)
+		return SC_ERROR_INTERNAL;
+
+	if (apdu.resplen < 11) {
+		sc_log(card->ctx, "Unexpected response to GET DATA (MyEIC card capabilities)");
+		return SC_ERROR_INTERNAL;
+	}
+
+	card_caps->card_caps_ver = rbuf[0];
+	/* the card returns big endian values */
+	card_caps->card_supported_features = (unsigned short) rbuf[1] << 8 | rbuf[2];
+	card_caps->max_rsa_key_length = (unsigned short) rbuf[3] << 8 | rbuf[4];
+	card_caps->max_des_key_length = (unsigned short) rbuf[5] << 8 | rbuf[6];
+	card_caps->max_aes_key_length = (unsigned short) rbuf[7] << 8 | rbuf[8];
+	card_caps->max_ecc_key_length = (unsigned short) rbuf[9] << 8 | rbuf[10];
+
+	LOG_FUNC_RETURN(card->ctx, r);
+}
+
 static int myeid_card_ctl(struct sc_card *card, unsigned long cmd, void *ptr)
 {
 	int r = SC_ERROR_NOT_SUPPORTED;
