@@ -64,16 +64,155 @@ static struct sc_atr_table sc_hsm_atrs[] = {
 	/* standard version */
 	{"3B:FE:18:00:00:81:31:FE:45:80:31:81:54:48:53:4D:31:73:80:21:40:81:07:FA", NULL, NULL, SC_CARD_TYPE_SC_HSM, 0, NULL},
 	{"3B:8E:80:01:80:31:81:54:48:53:4D:31:73:80:21:40:81:07:18", NULL, NULL, SC_CARD_TYPE_SC_HSM, 0, NULL},
+	{"3B:80:80:01:01", NULL, NULL, SC_CARD_TYPE_SC_HSM_SOC, 0, NULL},	// SoC Sample Card
+	{
+		"3B:84:80:01:47:6f:49:44:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:85:80:01:47:6f:49:44:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:86:80:01:47:6f:49:44:00:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:87:80:01:47:6f:49:44:00:00:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:88:80:01:47:6f:49:44:00:00:00:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:89:80:01:47:6f:49:44:00:00:00:00:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:8a:80:01:47:6f:49:44:00:00:00:00:00:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:8b:80:01:47:6f:49:44:00:00:00:00:00:00:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:8c:80:01:47:6f:49:44:00:00:00:00:00:00:00:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:8d:80:01:47:6f:49:44:00:00:00:00:00:00:00:00:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00:00:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:8e:80:01:47:6f:49:44:00:00:00:00:00:00:00:00:00:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00:00:00:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
+	{
+		"3B:8f:80:01:47:6f:49:44:00:00:00:00:00:00:00:00:00:00:00:00",
+		"FF:FF:FF:FF:FF:FF:FF:FF:00:00:00:00:00:00:00:00:00:00:00:00",
+		"GoID", SC_CARD_TYPE_SC_HSM_SOC, 0, NULL
+	},
 	{NULL, NULL, NULL, 0, 0, NULL}
 };
 
-/* Known ATRs for JavaCards that qualify for SmartCard-HSMs */
-static struct sc_atr_table sc_hsm_jc_atrs[] = {
-	/* standard version */
-	{"3b:f8:13:00:00:81:31:fe:45:4a:43:4f:50:76:32:34:31:b7", NULL, NULL, SC_CARD_TYPE_SC_HSM, 0, NULL},	// JCOP 2.4.1 Default ATR contact based
-	{"3b:88:80:01:4a:43:4f:50:76:32:34:31:5e", NULL, NULL, SC_CARD_TYPE_SC_HSM, 0, NULL},	// JCOP 2.4.1 Default ATR contactless
-	{NULL, NULL, NULL, 0, 0, NULL}
-};
+
+
+static int sc_hsm_select_file_ex(sc_card_t *card,
+			       const sc_path_t *in_path, int forceselect,
+			       sc_file_t **file_out)
+{
+	int rv;
+	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
+	sc_file_t *file = NULL;
+	sc_path_t cpath;
+
+	if (file_out == NULL) {				// Versions before 0.16 of the SmartCard-HSM do not support P2='0C'
+		rv = sc_hsm_select_file_ex(card, in_path, forceselect, &file);
+		if (file != NULL) {
+			sc_file_free(file);
+		}
+		return rv;
+	}
+
+	if ((in_path->type == SC_PATH_TYPE_FILE_ID) && in_path->aid.len) {
+		// Split applet selection and file selection into two separate calls
+		cpath = *in_path;
+		cpath.len = 0;
+		cpath.type = SC_PATH_TYPE_DF_NAME;
+		rv = sc_hsm_select_file_ex(card, &cpath, forceselect, NULL);
+		LOG_TEST_RET(card->ctx, rv, "Could not select SmartCard-HSM application");
+
+		if (in_path->len) {
+			cpath = *in_path;
+			cpath.aid.len = 0;
+			rv = sc_hsm_select_file_ex(card, &cpath, forceselect, file_out);
+		}
+		return rv;
+	}
+
+	// Prevent selection of applet unless this is the first time, selection is forced or the device is not authenticated
+	if (in_path->type == SC_PATH_TYPE_DF_NAME
+			|| (in_path->type == SC_PATH_TYPE_PATH
+				&& in_path->len == sc_hsm_aid.len
+				&& !memcmp(in_path->value, sc_hsm_aid.value, sc_hsm_aid.len))
+			|| (in_path->type == SC_PATH_TYPE_PATH
+				&& in_path->len == 0
+				&& in_path->aid.len == sc_hsm_aid.len
+				&& !memcmp(in_path->aid.value, sc_hsm_aid.value, sc_hsm_aid.len))) {
+		if (!priv || (priv->dffcp == NULL) || forceselect) {
+			rv = (*iso_ops->select_file)(card, in_path, file_out);
+			LOG_TEST_RET(card->ctx, rv, "Could not select SmartCard-HSM application");
+
+			if (priv) {
+				if (priv->dffcp != NULL) {
+					sc_file_free(priv->dffcp);
+				}
+				// Cache the FCP returned when selecting the applet
+				sc_file_dup(&priv->dffcp, *file_out);
+			}
+		} else {
+			sc_file_dup(file_out, priv->dffcp);
+			rv = SC_SUCCESS;
+		}
+		return rv;
+	}
+
+	if ((in_path->value[0] == 0x3F) && (in_path->value[1] == 0x00)) {
+		// The SmartCard-HSM is an applet that is not default selected. Simulate selection of the MF
+		if (in_path->len == 2) {
+			file = sc_file_new();
+			if (file == NULL)
+				LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
+			file->path = *in_path;
+			file->id = 0x3F00;
+			file->type = SC_FILE_TYPE_DF;
+			file->magic = SC_FILE_MAGIC;
+
+			*file_out = file;
+			return SC_SUCCESS;
+		} else {
+			sc_path_t truncated;
+			memcpy(&truncated, in_path, sizeof truncated);
+			truncated.len = in_path->len - 2;
+			memcpy(truncated.value, in_path->value+2, truncated.len);
+			return (*iso_ops->select_file)(card, &truncated, file_out);
+		}
+	}
+	return (*iso_ops->select_file)(card, in_path, file_out);
+}
 
 
 
@@ -81,35 +220,7 @@ static int sc_hsm_select_file(sc_card_t *card,
 			       const sc_path_t *in_path,
 			       sc_file_t **file_out)
 {
-	int rv;
-	sc_file_t *file = NULL;
-
-	if (file_out == NULL) {				// Versions before 0.16 of the SmartCard-HSM do not support P2='0C'
-		if (!in_path->len && in_path->aid.len) {
-			sc_log(card->ctx, "Preventing reselection of applet which would clear the security state");
-			return SC_SUCCESS;
-		}
-		rv = sc_hsm_select_file(card, in_path, &file);
-		if (file != NULL) {
-			sc_file_free(file);
-		}
-		return rv;
-	}
-
-	if ((in_path->len == 2) && (in_path->value[0] == 0x3F) && (in_path->value[1] == 0x00)) {
-		// The SmartCard-HSM is an applet that is not default selected. Simulate selection of the MF
-		file = sc_file_new();
-		if (file == NULL)
-			LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
-		file->path = *in_path;
-		file->id = 0x3F00;
-		file->type = SC_FILE_TYPE_DF;
-		file->magic = SC_FILE_MAGIC;
-
-		*file_out = file;
-		return SC_SUCCESS;
-	}
-	return (*iso_ops->select_file)(card, in_path, file_out);
+	return sc_hsm_select_file_ex(card, in_path, 0, file_out);
 }
 
 
@@ -123,51 +234,17 @@ static int sc_hsm_match_card(struct sc_card *card)
 	if (i >= 0)
 		return 1;
 
-	i = _sc_match_atr(card, sc_hsm_jc_atrs, &card->type);
-	if (i < 0)
-		return 0;
-
 	sc_path_set(&path, SC_PATH_TYPE_DF_NAME, sc_hsm_aid.value, sc_hsm_aid.len, 0, 0);
 	r = sc_hsm_select_file(card, &path, NULL);
 	LOG_TEST_RET(card->ctx, r, "Could not select SmartCard-HSM application");
 
-	// Select Applet to be sure
+	card->type = SC_CARD_TYPE_SC_HSM;
+
 	return 1;
 }
 
 
 
-static int sc_hsm_pin_info(sc_card_t *card, struct sc_pin_cmd_data *data,
-			   int *tries_left)
-{
-	sc_apdu_t apdu;
-	int r;
-
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x20, 0x00, data->pin_reference);
-
-	r = sc_transmit_apdu(card, &apdu);
-	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
-
-	r =  sc_check_sw(card, apdu.sw1, apdu.sw2);
-
-	if (r == SC_ERROR_PIN_CODE_INCORRECT) {
-		data->pin1.tries_left = apdu.sw2 & 0xF;
-		r = SC_SUCCESS;
-	} else if (r == SC_ERROR_AUTH_METHOD_BLOCKED) {
-		data->pin1.tries_left = 0;
-		r = SC_SUCCESS;
-	}
-	LOG_TEST_RET(card->ctx, r, "Check SW error");
-
-	if (tries_left != NULL) {
-		*tries_left = data->pin1.tries_left;
-	}
-
-	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
-}
-
-
-
 /*
  * Encode 16 hexadecimals of SO-PIN into binary form
  * Caller must check length of sopin and provide an 8 byte buffer
@@ -199,33 +276,404 @@ static int sc_hsm_encode_sopin(const u8 *sopin, u8 *sopinbin)
 }
 
 
+static int sc_hsm_soc_select_minbioclient(sc_card_t *card)
+{
+	sc_apdu_t apdu;
+	struct sc_aid minBioClient_aid = {
+		{ 0xFF,'m','i','n','B','i','o','C','l','i','e','n','t',0x01 }, 14
+	};
+
+	/* Select MinBioClient */
+	sc_sm_stop(card);
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xA4, 0x04, 0x0C);
+	apdu.data = minBioClient_aid.value;
+	apdu.datalen = minBioClient_aid.len;
+	apdu.lc = minBioClient_aid.len;
+	LOG_TEST_RET(card->ctx,
+			sc_transmit_apdu(card, &apdu),
+			"APDU transmit failed");
+
+	return sc_check_sw(card, apdu.sw1, apdu.sw2);
+}
+
+static int sc_hsm_soc_change(sc_card_t *card, struct sc_pin_cmd_data *data,
+			   int *tries_left)
+{
+	sc_apdu_t apdu;
+	sc_path_t path;
+	int r;
+
+	/* Select MinBioClient */
+	r = sc_hsm_soc_select_minbioclient(card);
+	LOG_TEST_RET(card->ctx, r, "Could not select MinBioClient application");
+
+	/* verify PIN */
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x20, 0x00, 0x80);
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_GOTO_ERR(card->ctx, r, "Could not verify PIN");
+
+	/* change PIN */
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x24, 0x01, 0x80);
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_GOTO_ERR(card->ctx, r, "Could not change PIN");
+
+err:
+	/* Select SC-HSM */
+	sc_path_set(&path, SC_PATH_TYPE_DF_NAME, sc_hsm_aid.value, sc_hsm_aid.len, 0, 0);
+	LOG_TEST_RET(card->ctx,
+			sc_hsm_select_file_ex(card, &path, 1, NULL),
+			"Could not select SmartCard-HSM application");
+
+	return r;
+}
+
+static int sc_hsm_soc_unblock(sc_card_t *card, struct sc_pin_cmd_data *data,
+			   int *tries_left)
+{
+	sc_apdu_t apdu;
+	sc_path_t path;
+	int r;
+
+	/* Select MinBioClient */
+	r = sc_hsm_soc_select_minbioclient(card);
+	LOG_TEST_RET(card->ctx, r, "Could not select MinBioClient application");
+
+	/* verify PUK */
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x20, 0x00, 0x81);
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_GOTO_ERR(card->ctx, r, "Could not verify PUK");
+
+	/* reset retry counter */
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_1, 0x2c, 0x03, 0x00);
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_GOTO_ERR(card->ctx, r, "APDU transmit failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_GOTO_ERR(card->ctx, r, "Could not unblock PIN");
+
+err:
+	/* Select SC-HSM */
+	sc_path_set(&path, SC_PATH_TYPE_DF_NAME, sc_hsm_aid.value, sc_hsm_aid.len, 0, 0);
+	LOG_TEST_RET(card->ctx,
+			sc_hsm_select_file_ex(card, &path, 1, NULL),
+			"Could not select SmartCard-HSM application");
+
+	return r;
+}
+
+static int sc_hsm_soc_biomatch(sc_card_t *card, struct sc_pin_cmd_data *data,
+			   int *tries_left)
+{
+	sc_apdu_t apdu;
+	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
+	int r;
+
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x20, 0x00, 0x85);
+	apdu.cla = 0x80;
+	apdu.data = (unsigned char*)"\x7F\x24\x00";
+	apdu.datalen = 3;
+	apdu.lc = 3;
+	apdu.resplen = 0;
+
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+	/* ignore the actual status bytes */
+
+	/* JCOP's SM accelerator is incapable of using case 1 APDU in SM */
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_2, 0x20, 0x00, 0x81);
+	apdu.resp = rbuf;
+	apdu.resplen = sizeof rbuf;
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+
+	/* now check the status bytes */
+	r =  sc_check_sw(card, apdu.sw1, apdu.sw2);
+	if (r == SC_SUCCESS) {
+		LOG_FUNC_RETURN(card->ctx, r);
+	}
+
+	LOG_FUNC_RETURN(card->ctx, SC_ERROR_PIN_CODE_INCORRECT);
+}
+
+
+
+#ifdef ENABLE_OPENPACE
+#include "sm/sm-eac.h"
+#include <eac/cv_cert.h>
+#include <eac/eac.h>
+#include <eac/ta.h>
+#include <openssl/bio.h>
+#include <openssl/crypto.h>
+
+static int sc_hsm_perform_chip_authentication(sc_card_t *card)
+{
+	int r, protocol;
+	sc_path_t path;
+	u8 all_certs[1024];
+	EAC_CTX *ctx = NULL;
+	size_t all_certs_len = sizeof all_certs, left, device_cert_len, issuer_cert_len;
+	const unsigned char *cert = all_certs, *device_cert, *issuer_cert;
+	BUF_MEM *comp_pub_key = NULL;
+	sc_cvc_t cvc_device, cvc_issuer;
+	/* this is only needed to call sc_pkcs15emu_sc_hsm_decode_cvc */
+	sc_pkcs15_card_t p15card;
+	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
+	/* we know that sc_pkcs15emu_sc_hsm_decode_cvc does not require anything
+	 * else to be initialized than p15card->card */
+	p15card.card = card;
+
+	memset(&cvc_device, 0, sizeof(cvc_device));
+	memset(&cvc_issuer, 0, sizeof(cvc_issuer));
+
+
+	if (priv->EF_C_DevAut && priv->EF_C_DevAut_len) {
+		all_certs_len = priv->EF_C_DevAut_len;
+		cert = priv->EF_C_DevAut;
+	} else {
+		/* get issuer and device certificate from the card */
+		r = sc_path_set(&path, SC_PATH_TYPE_FILE_ID, (u8 *) "\x2F\x02", 2, 0, 0);
+		if (r < 0)
+			goto err;
+		r = sc_select_file(card, &path, NULL);
+		if (r < 0)
+			goto err;
+		r = sc_read_binary(card, 0, all_certs, all_certs_len, 0);
+		if (r < 0)
+			goto err;
+
+		all_certs_len = r;
+
+		/* save EF_C_DevAut for further use */
+		cert = realloc(priv->EF_C_DevAut, all_certs_len);
+		if (cert) {
+			memcpy((unsigned char *) cert, all_certs, all_certs_len);
+			priv->EF_C_DevAut = (unsigned char *) cert;
+			priv->EF_C_DevAut_len = all_certs_len;
+		}
+
+		cert = all_certs;
+	}
+	left = all_certs_len;
+
+	device_cert = cert;
+	r = sc_pkcs15emu_sc_hsm_decode_cvc(&p15card, &cert, &left, &cvc_device);
+	if (r < 0)
+		goto err;
+	device_cert_len = all_certs_len - left;
+
+	issuer_cert = cert;
+	r = sc_pkcs15emu_sc_hsm_decode_cvc(&p15card, &cert, &left, &cvc_issuer);
+	if (r < 0)
+		goto err;
+	issuer_cert_len = all_certs_len - device_cert_len - left;
+
+	ctx = EAC_CTX_new();
+	if (!ctx) {
+		r = SC_ERROR_INTERNAL;
+		goto err;
+	}
+
+
+	/* check all CVCs given of the document's pki */
+	if (!TA_STEP2_import_certificate(ctx, issuer_cert, issuer_cert_len)
+			|| !TA_STEP2_import_certificate(ctx, device_cert, device_cert_len)) {
+		r = SC_ERROR_INTERNAL;
+		goto err;
+	}
+
+	/* XXX on older JCOPs only NID_id_CA_ECDH_3DES_CBC_CBC may be
+	 * supported. The card does not export its capabilities. We hardcode
+	 * NID_id_CA_ECDH_AES_CBC_CMAC_128 here, because we don't have the older
+	 * cards in production. */
+	protocol = NID_id_CA_ECDH_AES_CBC_CMAC_128;
+
+	/* initialize CA domain parameter with the document's public key */
+	if (!EAC_CTX_init_ca(ctx, protocol, 8)) {
+		r = SC_ERROR_INTERNAL;
+		goto err;
+	}
+	EVP_PKEY_free(ctx->ca_ctx->ka_ctx->key);
+	CRYPTO_add(&ctx->ta_ctx->pub_key->references, 1, CRYPTO_LOCK_EVP_PKEY);
+	ctx->ca_ctx->ka_ctx->key = ctx->ta_ctx->pub_key;
+
+	/* generate keys for CA */
+	comp_pub_key = TA_STEP3_generate_ephemeral_key(ctx);
+	r = perform_chip_authentication_ex(card, ctx,
+			cvc_device.publicPoint, cvc_device.publicPointlen);
+
+err:
+	if (r < 0)
+		EAC_CTX_clear_free(ctx);
+	if (comp_pub_key)
+		BUF_MEM_free(comp_pub_key);
+	sc_pkcs15emu_sc_hsm_free_cvc(&cvc_device);
+	sc_pkcs15emu_sc_hsm_free_cvc(&cvc_issuer);
+
+	return r;
+}
+
+#else
+
+static int sc_hsm_perform_chip_authentication(sc_card_t *card)
+{
+	return SC_ERROR_NOT_SUPPORTED;
+}
+#endif
+
+
 
 static int sc_hsm_pin_cmd(sc_card_t *card, struct sc_pin_cmd_data *data,
 			   int *tries_left)
 {
 	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
+	sc_apdu_t apdu;
+	u8 cmdbuff[16];
+	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
 	int r;
+	int cmd = data->cmd;
+	size_t pin2_len = data->pin2.len;
 
-	if (data->cmd == SC_PIN_CMD_GET_INFO) {
-		return sc_hsm_pin_info(card, data, tries_left);
-	}
-	if ((data->cmd == SC_PIN_CMD_VERIFY) && (data->pin_reference == 0x88)) {
-		if (data->pin1.len != 16)
-			return SC_ERROR_INVALID_PIN_LENGTH;
-
-		// Save SO PIN for later use in sc_hsm_init_pin()
-		r = sc_hsm_encode_sopin(data->pin1.data, priv->sopin);
-		LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
-
-		LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+	if (cmd == SC_PIN_CMD_GET_SESSION_PIN) {
+		/* First, perform a standard VERIFY */
+		data->cmd = SC_PIN_CMD_VERIFY;
+		/* we assign pin2.len to 0 early on so that in case of an error we are
+		 * not exiting with an undefined session PIN */
+		data->pin2.len = 0;
 	}
 
-	data->pin1.offset = 5;
-	data->pin1.length_offset = 4;
-	data->pin2.offset = 5;
-	data->pin2.length_offset = 4;
+	if ((card->caps & SC_CARD_CAP_PROTECTED_AUTHENTICATION_PATH)
+		   	&& (data->cmd == SC_PIN_CMD_CHANGE)
+		   	&& (data->pin_reference == 0x81)
+			&& (!data->pin1.data || data->pin1.len <= 0)) {
+		return sc_hsm_soc_change(card, data, tries_left);
+	} else if ((card->caps & SC_CARD_CAP_PROTECTED_AUTHENTICATION_PATH)
+		   	&& (data->cmd == SC_PIN_CMD_UNBLOCK)
+		   	&& (data->pin_reference == 0x81)
+			&& (!data->pin1.data || data->pin1.len <= 0)) {
+		return sc_hsm_soc_unblock(card, data, tries_left);
+	}
 
-	return (*iso_ops->pin_cmd)(card, data, tries_left);
+	/* For contactless cards always establish a secure channel before PIN
+	 * verification. Also, Session PIN generation requires SM. */
+	if ((card->type == SC_CARD_TYPE_SC_HSM_SOC || card->reader->uid.len
+				|| cmd == SC_PIN_CMD_GET_SESSION_PIN)
+			&& (data->cmd != SC_PIN_CMD_GET_INFO)
+			&& card->sm_ctx.sm_mode != SM_MODE_TRANSMIT) {
+		LOG_TEST_RET(card->ctx,
+				sc_hsm_perform_chip_authentication(card),
+				"Could not perform chip authentication");
+	}
+
+	if ((card->caps & SC_CARD_CAP_PROTECTED_AUTHENTICATION_PATH)
+			&& (data->cmd == SC_PIN_CMD_VERIFY)
+			&& (data->pin_reference == 0x81)
+			&& (!data->pin1.data || data->pin1.len <= 0)) {
+		r = sc_hsm_soc_biomatch(card, data, tries_left);
+	} else {
+		if ((data->cmd == SC_PIN_CMD_VERIFY) && (data->pin_reference == 0x88)) {
+			if (data->pin1.len != 16)
+				return SC_ERROR_INVALID_PIN_LENGTH;
+
+			// Save SO PIN for later use in sc_hsm_init_pin()
+			r = sc_hsm_encode_sopin(data->pin1.data, priv->sopin);
+			LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+
+			LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+		}
+
+		if ((data->cmd == SC_PIN_CMD_CHANGE) && (data->pin_reference == 0x88)) {
+			if ((data->pin1.len != 16) || (data->pin2.len != 16))
+				return SC_ERROR_INVALID_PIN_LENGTH;
+
+			r = sc_hsm_encode_sopin(data->pin1.data, cmdbuff);
+			LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+
+			r = sc_hsm_encode_sopin(data->pin2.data, cmdbuff + 8);
+			LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
+
+			sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x24, 0x00, data->pin_reference);
+			apdu.data = cmdbuff;
+			apdu.datalen = sizeof(cmdbuff);
+			apdu.lc = 16;
+			apdu.resplen = 0;
+			data->apdu = &apdu;
+		}
+
+		if ((data->cmd == SC_PIN_CMD_GET_INFO)
+				&& (card->sm_ctx.sm_mode == SM_MODE_TRANSMIT)) {
+			/* JCOP's SM accelerator is incapable of using case 1 APDU in SM */
+			sc_format_apdu(card, &apdu, SC_APDU_CASE_2, 0x20, 0x00, data->pin_reference);
+			apdu.resp = rbuf;
+			apdu.resplen = sizeof rbuf;
+			data->apdu = &apdu;
+		}
+
+		data->pin1.offset = 5;
+		data->pin1.length_offset = 4;
+		data->pin2.offset = 5;
+		data->pin2.length_offset = 4;
+
+		r = (*iso_ops->pin_cmd)(card, data, tries_left);
+	}
+	LOG_TEST_RET(card->ctx, r, "Verification failed");
+
+	if (cmd == SC_PIN_CMD_GET_SESSION_PIN) {
+		/* reset data->cmd to its original value */
+		data->cmd = SC_PIN_CMD_GET_SESSION_PIN;
+		if (data->pin_reference == 0x81) {
+			u8 recvbuf[SC_MAX_APDU_BUFFER_SIZE];
+			if (card->sm_ctx.sm_mode != SM_MODE_TRANSMIT) {
+				sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+						"Session PIN generation only supported in SM");
+				LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+			}
+			sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x5A, 0x01, data->pin_reference);
+			apdu.cla = 0x80;
+			apdu.resp = recvbuf;
+			apdu.resplen = sizeof recvbuf;
+			apdu.le = 0;
+			if (sc_transmit_apdu(card, &apdu) != SC_SUCCESS
+					|| sc_check_sw(card, apdu.sw1, apdu.sw2) != SC_SUCCESS) {
+				sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+						"Generating session PIN failed");
+				LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+			}
+			if (data->pin2.data && pin2_len > 0) {
+				if (pin2_len >= apdu.resplen) {
+					memcpy((unsigned char *) data->pin2.data, apdu.resp,
+							apdu.resplen);
+					data->pin2.len = apdu.resplen;
+				} else {
+					sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+							"Buffer too small for session PIN");
+				}
+			}
+		} else {
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+					"Session PIN not supported for this PIN (0x%02X)",
+					data->pin_reference);
+		}
+	}
+	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+}
+
+
+
+static int sc_hsm_logout(sc_card_t * card)
+{
+	sc_path_t path;
+	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
+	memset(priv->sopin, 0, sizeof(priv->sopin));
+	sc_sm_stop(card);
+
+	sc_path_set(&path, SC_PATH_TYPE_DF_NAME, sc_hsm_aid.value, sc_hsm_aid.len, 0, 0);
+
+	return sc_hsm_select_file_ex(card, &path, 1, NULL);
 }
 
 
@@ -236,7 +684,6 @@ static int sc_hsm_read_binary(sc_card_t *card,
 {
 	sc_context_t *ctx = card->ctx;
 	sc_apdu_t apdu;
-	u8 recvbuf[SC_MAX_APDU_BUFFER_SIZE];
 	u8 cmdbuff[4];
 	int r;
 
@@ -250,14 +697,14 @@ static int sc_hsm_read_binary(sc_card_t *card,
 	cmdbuff[2] = (idx >> 8) & 0xFF;
 	cmdbuff[3] = idx & 0xFF;
 
-	assert(count <= (card->max_recv_size > 0 ? card->max_recv_size : 256));
-	sc_format_apdu(card, &apdu, SC_APDU_CASE_4_SHORT, 0xB1, 0x00, 0x00);
+	assert(count <= sc_get_max_recv_size(card));
+	sc_format_apdu(card, &apdu, SC_APDU_CASE_4, 0xB1, 0x00, 0x00);
 	apdu.data = cmdbuff;
 	apdu.datalen = 4;
 	apdu.lc = 4;
 	apdu.le = count;
 	apdu.resplen = count;
-	apdu.resp = recvbuf;
+	apdu.resp = buf;
 
 	r = sc_transmit_apdu(card, &apdu);
 	LOG_TEST_RET(ctx, r, "APDU transmit failed");
@@ -267,8 +714,6 @@ static int sc_hsm_read_binary(sc_card_t *card,
 		LOG_TEST_RET(ctx, r, "Check SW error");
 	}
 
-	memcpy(buf, recvbuf, apdu.resplen);
-
 	LOG_FUNC_RETURN(ctx, apdu.resplen);
 }
 
@@ -419,7 +864,6 @@ static int sc_hsm_delete_file(sc_card_t *card, const sc_path_t *path)
 }
 
 
-
 static int sc_hsm_set_security_env(sc_card_t *card,
 				   const sc_security_env_t *env,
 				   int se_num)
@@ -496,7 +940,9 @@ static int sc_hsm_decode_ecdsa_signature(sc_card_t *card,
 		fieldsizebytes = 64;
 	}
 
-	sc_log(card->ctx, "Field size %d, signature buffer size %d", fieldsizebytes, outlen);
+	sc_log(card->ctx,
+	       "Field size %"SC_FORMAT_LEN_SIZE_T"u, signature buffer size %"SC_FORMAT_LEN_SIZE_T"u",
+	       fieldsizebytes, outlen);
 
 	if (outlen < (fieldsizebytes * 2)) {
 		LOG_TEST_RET(card->ctx, SC_ERROR_INVALID_DATA, "output too small for EC signature");
@@ -546,14 +992,22 @@ static int sc_hsm_compute_signature(sc_card_t *card,
 	sc_apdu_t apdu;
 	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
 	u8 sbuf[SC_MAX_APDU_BUFFER_SIZE];
-	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
+	sc_hsm_private_data_t *priv;
 
-	assert(card != NULL && data != NULL && out != NULL);
+	if (card == NULL || data == NULL || out == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	priv = (sc_hsm_private_data_t *) card->drv_data;
 
 	if (priv->env == NULL) {
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_OBJECT_NOT_FOUND);
 	}
 
+	// check if datalen exceeds the buffer size
+	if (datalen > SC_MAX_APDU_BUFFER_SIZE) {
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
+	}
+
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_4, 0x68, priv->env->key_ref[0], priv->algorithm);
 	apdu.cla = 0x80;
 	apdu.resp = rbuf;
@@ -565,6 +1019,7 @@ static int sc_hsm_compute_signature(sc_card_t *card,
 	apdu.lc = datalen;
 	apdu.datalen = datalen;
 	r = sc_transmit_apdu(card, &apdu);
+
 	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
 		int len;
@@ -591,10 +1046,13 @@ static int sc_hsm_decipher(sc_card_t *card, const u8 * crgram, size_t crgram_len
 	size_t len;
 	sc_apdu_t apdu;
 	u8 rbuf[SC_MAX_APDU_BUFFER_SIZE];
-	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
+	sc_hsm_private_data_t *priv;
 
-	assert(card != NULL && crgram != NULL && out != NULL);
+	if (card == NULL || crgram == NULL || out == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	LOG_FUNC_CALLED(card->ctx);
+	priv = (sc_hsm_private_data_t *) card->drv_data;
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_4, 0x62, priv->env->key_ref[0], priv->algorithm);
 	apdu.cla = 0x80;
@@ -870,7 +1328,7 @@ static int sc_hsm_init_token(sc_card_t *card, sc_cardctl_pkcs11_init_token_t *pa
 	memset(&ip, 0, sizeof(ip));
 	ip.dkek_shares = -1;
 	ip.options[0] = 0x00;
-	ip.options[0] = 0x01;
+	ip.options[1] = 0x01;
 
 	r = sc_hsm_encode_sopin(params->so_pin, ip.init_code);
 	LOG_TEST_RET(ctx, r, "SO PIN wrong format");
@@ -1026,16 +1484,23 @@ static int sc_hsm_card_ctl(sc_card_t *card, unsigned long cmd, void *ptr)
 
 static int sc_hsm_init(struct sc_card *card)
 {
-	sc_hsm_private_data_t *priv;
+#ifdef _WIN32
+	char expanded_val[PATH_MAX];
+	size_t expanded_len = PATH_MAX;
+#endif
 	int flags,ext_flags;
+	sc_file_t *file;
+	sc_path_t path;
+	sc_hsm_private_data_t *priv = card->drv_data;
 
 	LOG_FUNC_CALLED(card->ctx);
 
-	priv = calloc(1, sizeof(sc_hsm_private_data_t));
-	if (!priv)
-		LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
-
-	card->drv_data = priv;
+	if (!priv) {
+		priv = calloc(1, sizeof(sc_hsm_private_data_t));
+		if (!priv)
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
+		card->drv_data = priv;
+	}
 
 	flags = SC_ALGORITHM_RSA_RAW|SC_ALGORITHM_ONBOARD_KEY_GEN;
 
@@ -1061,9 +1526,58 @@ static int sc_hsm_init(struct sc_card *card)
 	_sc_card_add_ec_alg(card, 256, flags, ext_flags, NULL);
 	_sc_card_add_ec_alg(card, 320, flags, ext_flags, NULL);
 
-	card->caps |= SC_CARD_CAP_RNG|SC_CARD_CAP_APDU_EXT;
+	card->caps |= SC_CARD_CAP_RNG|SC_CARD_CAP_APDU_EXT|SC_CARD_CAP_ISO7816_PIN_INFO;
+
+	sc_path_set(&path, SC_PATH_TYPE_DF_NAME, sc_hsm_aid.value, sc_hsm_aid.len, 0, 0);
+	if (sc_hsm_select_file_ex(card, &path, 0, &file) == SC_SUCCESS
+			&& file->prop_attr && file->prop_attr_len >= 5) {
+		static char card_name[SC_MAX_APDU_BUFFER_SIZE];
+		u8 type = file->prop_attr[2];
+		u8 major = file->prop_attr[3];
+		u8 minor = file->prop_attr[4];
+		char p00[] = "SmartCard-HSM Applet for JCOP";
+		char p01[] = "SmartCard-HSM Demo Applet for JCOP";
+		char *p = "SmartCard-HSM";
+		switch (type) {
+			case 0x00:
+				p = p00;
+				break;
+			case 0x01:
+				p = p01;
+				break;
+			default:
+				break;
+		}
+		snprintf(card_name, sizeof card_name, "%s version %u.%u", p, major, minor);
+		card->name = card_name;
+
+		if (file->prop_attr[1] & 0x04) {
+			card->caps |= SC_CARD_CAP_SESSION_PIN;
+		}
+		sc_file_free(file);
+	}
 
 	card->max_send_size = 1431;		// 1439 buffer size - 8 byte TLV because of odd ins in UPDATE BINARY
+	if (card->type == SC_CARD_TYPE_SC_HSM_SOC) {
+		card->max_recv_size = 0x0630;	// SoC Proxy forces this limit
+	} else {
+		card->max_recv_size = 0;		// Card supports sending with extended length APDU and without limit
+	}
+
+	priv->EF_C_DevAut = NULL;
+	priv->EF_C_DevAut_len = 0;
+
+#ifdef ENABLE_OPENPACE
+	EAC_init();
+#ifdef _WIN32
+	expanded_len = ExpandEnvironmentStringsA(CVCDIR, expanded_val, sizeof expanded_val);
+	if (0 < expanded_len && expanded_len < sizeof expanded_val)
+		EAC_set_cvc_default_dir(expanded_val);
+#else
+	EAC_set_cvc_default_dir(CVCDIR);
+#endif
+#endif
+
 	return 0;
 }
 
@@ -1072,10 +1586,20 @@ static int sc_hsm_init(struct sc_card *card)
 static int sc_hsm_finish(sc_card_t * card)
 {
 	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
+	sc_sm_stop(card);
 	if (priv->serialno) {
 		free(priv->serialno);
 	}
+	if (priv->dffcp) {
+		sc_file_free(priv->dffcp);
+	}
+	free(priv->EF_C_DevAut);
 	free(priv);
+
+#ifdef ENABLE_OPENPACE
+	EAC_cleanup();
+#endif
+
 	return SC_SUCCESS;
 }
 
@@ -1103,6 +1627,7 @@ static struct sc_card_driver * sc_get_driver(void)
 	sc_hsm_ops.finish            = sc_hsm_finish;
 	sc_hsm_ops.card_ctl          = sc_hsm_card_ctl;
 	sc_hsm_ops.pin_cmd           = sc_hsm_pin_cmd;
+	sc_hsm_ops.logout            = sc_hsm_logout;
 
 	/* no record oriented file services */
 	sc_hsm_ops.read_record       = NULL;
