@@ -210,6 +210,17 @@ enum {
 	SC_CARDCTL_PIV_PIN_PREFERENCE,
 	SC_CARDCTL_PIV_OBJECT_PRESENT,
 
+	/*
+	 * CAC specific calls
+	 */
+	SC_CARDCTL_CAC_BASE = _CTL_PREFIX('C', 'A', 'C'),
+	SC_CARDCTL_CAC_INIT_GET_GENERIC_OBJECTS,
+	SC_CARDCTL_CAC_GET_NEXT_GENERIC_OBJECT,
+	SC_CARDCTL_CAC_FINAL_GET_GENERIC_OBJECTS,
+	SC_CARDCTL_CAC_INIT_GET_CERT_OBJECTS,
+	SC_CARDCTL_CAC_GET_NEXT_CERT_OBJECT,
+	SC_CARDCTL_CAC_FINAL_GET_CERT_OBJECTS,
+
         /*
 	 * AuthentIC v3
 	 */
@@ -219,6 +230,17 @@ enum {
         SC_CARDCTL_AUTHENTIC_SDO_STORE,
         SC_CARDCTL_AUTHENTIC_SDO_GENERATE,
 
+	/*
+	 * Coolkey specific calls
+	 */
+	SC_CARDCTL_COOLKEY_BASE = _CTL_PREFIX('C', 'O', 'K'),
+	SC_CARDCTL_COOLKEY_INIT_GET_OBJECTS,
+	SC_CARDCTL_COOLKEY_GET_NEXT_OBJECT,
+	SC_CARDCTL_COOLKEY_FINAL_GET_OBJECTS,
+	SC_CARDCTL_COOLKEY_GET_ATTRIBUTE,
+	SC_CARDCTL_COOLKEY_GET_TOKEN_INFO,
+	SC_CARDCTL_COOLKEY_FIND_OBJECT,
+
         /*
 	 * IAS/ECC
 	 */
@@ -262,7 +284,24 @@ enum {
 	 */
 	SC_CARDCTL_ISOAPPLET_BASE = _CTL_PREFIX('I','S','O'),
 	SC_CARDCTL_ISOAPPLET_GENERATE_KEY,
-	SC_CARDCTL_ISOAPPLET_IMPORT_KEY
+	SC_CARDCTL_ISOAPPLET_IMPORT_KEY,
+
+	/*
+	 * GIDS cards
+	 */
+	SC_CARDCTL_GIDS_BASE = _CTL_PREFIX('G','I','D'),
+	SC_CARDCTL_GIDS_GET_ALL_CONTAINERS,
+	SC_CARDCTL_GIDS_GET_CONTAINER_DETAIL,
+	SC_CARDCTL_GIDS_SELECT_KEY_REFERENCE,
+	SC_CARDCTL_GIDS_CREATE_KEY,
+	SC_CARDCTL_GIDS_GENERATE_KEY,
+	SC_CARDCTL_GIDS_IMPORT_KEY,
+	SC_CARDCTL_GIDS_SAVE_CERT,
+	SC_CARDCTL_GIDS_DELETE_KEY,
+	SC_CARDCTL_GIDS_DELETE_CERT,
+	SC_CARDCTL_GIDS_INITIALIZE,
+	SC_CARDCTL_GIDS_SET_ADMIN_KEY,
+	SC_CARDCTL_GIDS_AUTHENTICATE_ADMIN,
 };
 
 enum {
@@ -523,9 +562,9 @@ typedef struct sc_cardctl_muscle_key_info {
 	int 	keyType;
 	int 	keyLocation;
 	int 	keySize;
-	int 	modLength;
+	size_t 	modLength;
 	u8* 	modValue;
-	int 	expLength;
+	size_t 	expLength;
 	u8* 	expValue;
 	int 	pLength;
 	u8* 	pValue;
@@ -839,7 +878,9 @@ typedef struct sc_rtecp_genkey_data {
  */
 	enum SC_CARDCTL_MYEID_KEY_TYPE {
 		SC_CARDCTL_MYEID_KEY_RSA = 0x11,
-		SC_CARDCTL_MYEID_KEY_EC  = 0x22
+		SC_CARDCTL_MYEID_KEY_DES = 0x19,
+		SC_CARDCTL_MYEID_KEY_EC  = 0x22,
+		SC_CARDCTL_MYEID_KEY_AES = 0x29
 	};
 
 	struct sc_cardctl_myeid_data_obj {
@@ -868,8 +909,8 @@ typedef struct sc_rtecp_genkey_data {
 		unsigned int    invq_len;  
 		unsigned char  *invq;
 		/* new for MyEID > 3.6.0 */
-		unsigned char  *d;                  /* EC private key */
-		unsigned int    d_len;              /* EC */ 
+		unsigned char  *d;                  /* EC private key / Symmetric key */
+		unsigned int    d_len;              /* EC / Symmetric */
 		unsigned char  *ecpublic_point;     /* EC public key */
 		unsigned int    ecpublic_point_len; /* EC */
     };
@@ -1014,6 +1055,40 @@ typedef struct sc_cardctl_isoApplet_import_key {
 	} privkey;
 } sc_cardctl_isoApplet_import_key_t;
 
+/*
+ * coolkey object returned from the card control interface
+ */
+typedef struct sc_cardctl_coolkey_object {
+        sc_path_t path;
+        unsigned long id;
+        size_t length;
+        u8  *data;
+} sc_cardctl_coolkey_object_t;
+
+
+/* data structure to pass attributes through the ctl interface */
+typedef struct sc_cardctl_coolkey_attribute {
+        const sc_cardctl_coolkey_object_t *object;
+	unsigned long attribute_type;
+        u8 attribute_data_type;
+        size_t attribute_length;
+        const u8 *attribute_value;
+} sc_cardctl_coolkey_attribute_t;
+
+#define SC_CARDCTL_COOLKEY_ATTR_TYPE_STRING 0
+#define SC_CARDCTL_COOLKEY_ATTR_TYPE_ULONG 1
+
+typedef struct sc_cardctl_coolkey_find_object {
+	int type; /* in parameter */
+	unsigned long find_id; /* in parameter */
+	sc_cardctl_coolkey_attribute_t *coolkey_template; /* in paramter */
+	int template_count;                       /* in parameter */
+	sc_cardctl_coolkey_object_t *obj; /* out parameter */
+} sc_cardctl_coolkey_find_object_t;
+
+#define SC_CARDCTL_COOLKEY_FIND_BY_ID       0
+#define SC_CARDCTL_COOLKEY_FIND_BY_TEMPLATE 1
+
 #ifdef __cplusplus
 }
 #endif
