@@ -94,6 +94,7 @@ enum {
 	/* starcos driver */
 	SC_CARD_TYPE_STARCOS_BASE = 7000,
 	SC_CARD_TYPE_STARCOS_GENERIC,
+	SC_CARD_TYPE_STARCOS_V3_4,
 
 	/* tcos driver */
 	SC_CARD_TYPE_TCOS_BASE = 8000,
@@ -105,6 +106,7 @@ enum {
 	SC_CARD_TYPE_OPENPGP_BASE = 9000,
 	SC_CARD_TYPE_OPENPGP_V1,
 	SC_CARD_TYPE_OPENPGP_V2,
+	SC_CARD_TYPE_OPENPGP_GNUK,
 
 	/* jcop driver */
 	SC_CARD_TYPE_JCOP_BASE = 10000,
@@ -131,6 +133,9 @@ enum {
 	/* PIV-II type cards */
 	SC_CARD_TYPE_PIV_II_BASE = 14000,
 	SC_CARD_TYPE_PIV_II_GENERIC,
+	SC_CARD_TYPE_PIV_II_HIST,
+	SC_CARD_TYPE_PIV_II_NEO,
+	SC_CARD_TYPE_PIV_II_YUBIKEY4,
 
 	/* MuscleApplet */
 	SC_CARD_TYPE_MUSCLE_BASE = 15000,
@@ -159,6 +164,13 @@ enum {
 	SC_CARD_TYPE_ENTERSAFE_3K,
 	SC_CARD_TYPE_ENTERSAFE_FTCOS_PK_01C,
 	SC_CARD_TYPE_ENTERSAFE_FTCOS_EPASS2003,
+	SC_CARD_TYPE_ENTERSAFE_EJAVA_PK_01C,
+	SC_CARD_TYPE_ENTERSAFE_EJAVA_PK_01C_T0,
+	SC_CARD_TYPE_ENTERSAFE_EJAVA_H10CR_PK_01C_T1,
+	SC_CARD_TYPE_ENTERSAFE_EJAVA_D11CR_PK_01C_T1,
+	SC_CARD_TYPE_ENTERSAFE_EJAVA_C21C_PK_01C_T1,
+	SC_CARD_TYPE_ENTERSAFE_EJAVA_A22CR_PK_01C_T1,
+	SC_CARD_TYPE_ENTERSAFE_EJAVA_A40CR_PK_01C_T1,
 
 	/* MyEID cards */
 	SC_CARD_TYPE_MYEID_BASE = 20000,
@@ -170,10 +182,6 @@ enum {
 	SC_CARD_TYPE_GEMSAFEV1_PTEID,
 	SC_CARD_TYPE_GEMSAFEV1_SEEID,
 
-	/* IAS cards */
-	SC_CARD_TYPE_IAS_BASE = 22000,
-	SC_CARD_TYPE_IAS_PTEID,
-
 	/* Italian CNS cards */
 	SC_CARD_TYPE_ITACNS_BASE = 23000,
 	SC_CARD_TYPE_ITACNS_GENERIC,
@@ -195,6 +203,7 @@ enum {
 
 	/* SmartCard-HSM */
 	SC_CARD_TYPE_SC_HSM = 26000,
+	SC_CARD_TYPE_SC_HSM_SOC = 26001,
 
 	/* Spanish DNIe card */
 	SC_CARD_TYPE_DNIE_BASE = 27000,
@@ -209,7 +218,30 @@ enum {
 
 	/* Masktech cards */
 	SC_CARD_TYPE_MASKTECH_BASE = 29000,
-	SC_CARD_TYPE_MASKTECH_GENERIC
+	SC_CARD_TYPE_MASKTECH_GENERIC,
+
+	/* GIDS cards */
+	SC_CARD_TYPE_GIDS_BASE = 30000,
+	SC_CARD_TYPE_GIDS_GENERIC,
+	SC_CARD_TYPE_GIDS_V1,
+	SC_CARD_TYPE_GIDS_V2,
+
+	/* JPKI cards */
+	SC_CARD_TYPE_JPKI_BASE = 31000,
+
+	SC_CARD_TYPE_COOLKEY_BASE = 32000,
+	SC_CARD_TYPE_COOLKEY_GENERIC,
+
+	/* CAC cards */
+	SC_CARD_TYPE_CAC_BASE = 33000,
+	SC_CARD_TYPE_CAC_GENERIC,
+	SC_CARD_TYPE_CAC_I,
+	SC_CARD_TYPE_CAC_II,
+
+	/* nPA cards */
+	SC_CARD_TYPE_NPA = 34000,
+	SC_CARD_TYPE_NPA_TEST,
+	SC_CARD_TYPE_NPA_ONLINE,
 };
 
 extern sc_card_driver_t *sc_get_default_driver(void);
@@ -239,7 +271,6 @@ extern sc_card_driver_t *sc_get_rutoken_driver(void);
 extern sc_card_driver_t *sc_get_rtecp_driver(void);
 extern sc_card_driver_t *sc_get_westcos_driver(void);
 extern sc_card_driver_t *sc_get_myeid_driver(void);
-extern sc_card_driver_t *sc_get_ias_driver(void);
 extern sc_card_driver_t *sc_get_sc_hsm_driver(void);
 extern sc_card_driver_t *sc_get_itacns_driver(void);
 extern sc_card_driver_t *sc_get_authentic_driver(void);
@@ -248,6 +279,11 @@ extern sc_card_driver_t *sc_get_epass2003_driver(void);
 extern sc_card_driver_t *sc_get_dnie_driver(void);
 extern sc_card_driver_t *sc_get_isoApplet_driver(void);
 extern sc_card_driver_t *sc_get_masktech_driver(void);
+extern sc_card_driver_t *sc_get_gids_driver(void);
+extern sc_card_driver_t *sc_get_jpki_driver(void);
+extern sc_card_driver_t *sc_get_coolkey_driver(void);
+extern sc_card_driver_t *sc_get_cac_driver(void);
+extern sc_card_driver_t *sc_get_npa_driver(void);
 
 #ifdef __cplusplus
 }
