@@ -1,10 +1,10 @@
 /**
  * cwa-dnie.c: DNIe data provider for CWA SM handling.
- * 
+ *
  * Copyright (C) 2010 Juan Antonio Martinez <jonsito@terra.es>
  *
  * This work is derived from many sources at OpenSC Project site,
- * (see references) and the information made public by Spanish 
+ * (see references) and the information made public by Spanish
  * Direccion General de la Policia y de la Guardia Civil
  *
  * This library is free software; you can redistribute it and/or
@@ -27,7 +27,7 @@
 #include "config.h"
 #endif
 
-#ifdef ENABLE_OPENSSL		/* empty file without openssl */
+#if defined(ENABLE_OPENSSL) && defined(ENABLE_SM)	/* empty file without openssl or sm */
 
 #include <stdlib.h>
 #include <string.h>
@@ -39,8 +39,13 @@
 
 #include "cwa-dnie.h"
 
+#include <openssl/ossl_typ.h>
+#include <openssl/bn.h>
 #include <openssl/x509.h>
 #include <openssl/evp.h>
+#include <openssl/rsa.h>
+
+#define MAX_RESP_BUFFER_SIZE 2048
 
 /********************* Keys and certificates as published by DGP ********/
 
@@ -85,6 +90,23 @@ static u8 ifd_modulus[] = {
 	0x6a, 0xe2, 0x36, 0x59, 0x00, 0x16, 0xba, 0x69
 };
 
+/**
+ * Terminal (IFD) key modulus for SM channel creation for PIN channel DNIe 3.0
+ */
+static u8 ifd_pin_modulus[] = {
+	0xF4, 0x27, 0x97, 0x8D, 0xA1, 0x59, 0xBA, 0x02, 0x79, 0x30, 0x8A, 0x6C,
+	0x6A, 0x89, 0x50, 0x5A, 0xDA, 0x5A, 0x67, 0xC3, 0xDA, 0x26, 0x79, 0xEA,
+	0xF4, 0xA1, 0xB0, 0x11, 0x9E, 0xDD, 0x4D, 0xF4, 0x6E, 0x78, 0x04, 0x24,
+	0x71, 0xA9, 0xD1, 0x30, 0x1D, 0x3F, 0xB2, 0x8F, 0x38, 0xC5, 0x7D, 0x08,
+	0x89, 0xF7, 0x31, 0xDB, 0x8E, 0xDD, 0xBC, 0x13, 0x67, 0xC1, 0x34, 0xE1,
+	0xE9, 0x47, 0x78, 0x6B, 0x8E, 0xC8, 0xE4, 0xB9, 0xCA, 0x6A, 0xA7, 0xC2,
+	0x4C, 0x86, 0x91, 0xC7, 0xBE, 0x2F, 0xD8, 0xC1, 0x23, 0x66, 0x0E, 0x98,
+	0x65, 0xE1, 0x4F, 0x19, 0xDF, 0xFB, 0xB7, 0xFF, 0x38, 0x08, 0xC9, 0xF2,
+	0x04, 0xE7, 0x97, 0xD0, 0x6D, 0xD8, 0x33, 0x3A, 0xC5, 0x83, 0x86, 0xEE,
+	0x4E, 0xB6, 0x1E, 0x20, 0xEC, 0xA7, 0xEF, 0x38, 0xD5, 0xB0, 0x5E, 0xB1,
+	0x15, 0x96, 0x6A, 0x5A, 0x89, 0xAD, 0x58, 0xA5
+};
+
 /**
  * Terminal (IFD) public exponent for SM channel creation
  */
@@ -92,6 +114,13 @@ static u8 ifd_public_exponent[] = {
 	0x01, 0x00, 0x01
 };
 
+/**
+ * Terminal (IFD) public exponent for SM channel creation for PIN channel DNIe 3.0
+ */
+static u8 ifd_pin_public_exponent[] = {
+	0x01, 0x00, 0x01
+};
+
 /**
  * Terminal (IFD) private exponent for SM channel establishment
  */
@@ -109,6 +138,23 @@ static u8 ifd_private_exponent[] = {
 	0xbd, 0x9b, 0x00, 0x31, 0x3c, 0x0f, 0x46, 0xed
 };
 
+/**
+ * Terminal (IFD) private exponent for SM channel establishment for PIN channel DNIe 3.0
+ */
+static u8 ifd_pin_private_exponent[] = {
+	0xD2, 0x7A, 0x03, 0x23, 0x7C, 0x72, 0x2E, 0x71, 0x8D, 0x69, 0xF4, 0x1A,
+	0xEC, 0x68, 0xBD, 0x95, 0xE4, 0xE0, 0xC4, 0xCD, 0x49, 0x15, 0x9C, 0x4A,
+	0x99, 0x63, 0x7D, 0xB6, 0x62, 0xFE, 0xA3, 0x02, 0x51, 0xED, 0x32, 0x9C,
+	0xFC, 0x43, 0x89, 0xEB, 0x71, 0x7B, 0x85, 0x02, 0x04, 0xCD, 0xF3, 0x30,
+	0xD6, 0x46, 0xFC, 0x7B, 0x2B, 0x19, 0x29, 0xD6, 0x8C, 0xBE, 0x39, 0x49,
+	0x7B, 0x62, 0x3A, 0x82, 0xC7, 0x64, 0x1A, 0xC3, 0x48, 0x79, 0x57, 0x3D,
+	0xEA, 0x0D, 0xAB, 0xC7, 0xCA, 0x30, 0x9A, 0xE4, 0xB3, 0xED, 0xDA, 0xFA,
+	0xEE, 0x55, 0xD5, 0x42, 0xF7, 0x80, 0x23, 0x03, 0x51, 0xE7, 0x5E, 0x7F,
+	0x32, 0xDC, 0x65, 0x2E, 0xF1, 0xED, 0x47, 0xA5, 0x1C, 0x18, 0xD9, 0xDF,
+	0x9F, 0xF4, 0x8D, 0x87, 0x8D, 0xB6, 0x22, 0xEA, 0x6E, 0x93, 0x70, 0xE9,
+	0xC6, 0x3B, 0x35, 0x8B, 0x7C, 0x11, 0x5A, 0xA1
+};
+
 /**
  *  Intermediate CA certificate in CVC format (Card verifiable certificate)
  */
@@ -133,7 +179,7 @@ static u8 C_CV_CA_CS_AUT_cert[] = {
 	0x52, 0x44, 0x49, 0x60, 0x00, 0x06
 };
 
-/** 
+/**
  * Terminal (IFD) certificate in CVC format (PK.IFD.AUT)
  */
 static u8 C_CV_IFDUser_AUT_cert[] = {
@@ -157,6 +203,30 @@ static u8 C_CV_IFDUser_AUT_cert[] = {
 	0x44, 0x49, 0x60, 0x00, 0x06
 };
 
+/**
+ * Terminal (IFD) certificate in CVC format (PK.IFD.AUT) for the PIN channel in DNIe 3.0
+ */
+static u8 C_CV_IFDUser_AUT_pin_cert[] = {
+	0x7f, 0x21, 0x81, 0xcd, 0x5f, 0x37, 0x81, 0x80, 0x69, 0xc4, 0xe4, 0x94,
+	0xf0, 0x08, 0xe2, 0x42, 0x14, 0xb1, 0xc1, 0x31, 0xb6, 0x1f, 0xce, 0x9c,
+	0x15, 0xfa, 0x3c, 0xb0, 0x61, 0xdd, 0x6f, 0x02, 0xd8, 0xa2, 0xcd, 0x30,
+	0xd7, 0x2f, 0xb6, 0xdf, 0x89, 0x9a, 0xf1, 0x5b, 0x71, 0x78, 0x21, 0xbf,
+	0xb1, 0xaf, 0x7d, 0x75, 0x85, 0x01, 0x6d, 0x8c, 0x36, 0xaf, 0x4a, 0xc2,
+	0xa0, 0xb0, 0xc5, 0x2a, 0xd6, 0x5b, 0x69, 0x25, 0x67, 0x31, 0xc3, 0x4d,
+	0x59, 0x02, 0x0e, 0x87, 0xab, 0x73, 0xa2, 0x30, 0xfa, 0x69, 0xee, 0x82,
+	0xb3, 0x3a, 0x31, 0xdf, 0x04, 0x0c, 0xe9, 0x0f, 0x0a, 0xfc, 0x3a, 0x11,
+	0x1d, 0x35, 0xda, 0x95, 0x66, 0xa8, 0xcd, 0xab, 0xea, 0x0e, 0x3f, 0x75,
+	0x94, 0xc4, 0x40, 0xd3, 0x74, 0x50, 0x7a, 0x94, 0x35, 0x57, 0x59, 0xb3,
+	0x9e, 0xc5, 0xe5, 0xfc, 0xb8, 0x03, 0x8d, 0x79, 0x3d, 0x5f, 0x9b, 0xa8,
+	0xb5, 0xb1, 0x0b, 0x70, 0x5f, 0x38, 0x3c, 0x4c, 0x86, 0x91, 0xc7, 0xbe,
+	0x2f, 0xd8, 0xc1, 0x23, 0x66, 0x0e, 0x98, 0x65, 0xe1, 0x4f, 0x19, 0xdf,
+	0xfb, 0xb7, 0xff, 0x38, 0x08, 0xc9, 0xf2, 0x04, 0xe7, 0x97, 0xd0, 0x6d,
+	0xd8, 0x33, 0x3a, 0xc5, 0x83, 0x86, 0xee, 0x4e, 0xb6, 0x1e, 0x20, 0xec,
+	0xa7, 0xef, 0x38, 0xd5, 0xb0, 0x5e, 0xb1, 0x15, 0x96, 0x6a, 0x5a, 0x89,
+	0xad, 0x58, 0xa5, 0x00, 0x01, 0x00, 0x01, 0x42, 0x08, 0x65, 0x73, 0x53,
+	0x44, 0x49, 0x60, 0x00, 0x06
+};
+
 /**
  * Root CA card key reference
  */
@@ -164,13 +234,13 @@ static u8 root_ca_keyref[] = { 0x02, 0x0f };
 
 
 /**
- * ICC card private key reference 
+ * ICC card private key reference
  */
 static u8 icc_priv_keyref[] = { 0x02, 0x1f };
 
 /**
  * Intermediate CA card key reference
- */ 
+ */
 static u8 cvc_intca_keyref[] =
     { 0x65, 0x73, 0x53, 0x44, 0x49, 0x60, 0x00, 0x06 };
 
@@ -180,16 +250,21 @@ static u8 cvc_intca_keyref[] =
 static u8 cvc_ifd_keyref[] =
     { 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01 };
 
+/**
+ * In memory key reference for selecting IFD sent certificate in PIN channel DNIe 3.0
+ */
+static u8 cvc_ifd_keyref_pin[] =
+    { 0x00, 0x00, 0x00, 0x00, 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01};
+
 /**
  * Serial number for IFD Terminal application
  */
 static u8 sn_ifd[] = { 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01 };
 
 /**
- * Serial number for ICC card.
- * This buffer is to be filled at runtime
+ * Serial number for IFD Terminal application in PIN channel DNIe 3.0
  */
-static u8 sn_icc[] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };
+static u8 sn_ifd_pin[] = { 0x40, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01};
 
 /************ internal functions **********************************/
 
@@ -209,7 +284,7 @@ int dnie_read_file(sc_card_t * card,
 		   const sc_path_t * path,
 		   sc_file_t ** file, u8 ** buffer, size_t * length)
 {
-	u8 *data;
+	u8 *data = NULL;
 	char *msg = NULL;
 	int res = SC_SUCCESS;
 	size_t fsize = 0;	/* file size */
@@ -223,7 +298,7 @@ int dnie_read_file(sc_card_t * card,
 		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
 	/* select file by mean of iso7816 ops */
 	res = card->ops->select_file(card, path, file);
-	if (res != SC_SUCCESS) {
+	if (res != SC_SUCCESS || !file || !(*file)) {
 		msg = "select_file failed";
 		goto dnie_read_file_err;
 	}
@@ -252,7 +327,8 @@ int dnie_read_file(sc_card_t * card,
 		goto dnie_read_file_err;
 	}
 	/* call sc_read_binary() to retrieve data */
-	sc_log(ctx, "read_binary(): expected '%d' bytes", fsize);
+	sc_log(ctx, "read_binary(): expected '%"SC_FORMAT_LEN_SIZE_T"u' bytes",
+	       fsize);
 	res = sc_read_binary(card, 0, data, fsize, 0L);
 	if (res < 0) {		/* read_binary returns number of bytes readed */
 		res = SC_ERROR_CARD_CMD_FAILED;
@@ -265,13 +341,15 @@ int dnie_read_file(sc_card_t * card,
 	res = SC_SUCCESS;
 	goto dnie_read_file_end;
  dnie_read_file_err:
-	if (*file) {
+	if (data)
+		free(data);
+	if (file) {
 		sc_file_free(*file);
 		*file = NULL;
 	}
  dnie_read_file_end:
 	if (msg)
-		sc_log(ctx, msg);
+		sc_log(ctx, "%s", msg);
 	LOG_FUNC_RETURN(ctx, res);
 }
 
@@ -279,52 +357,48 @@ int dnie_read_file(sc_card_t * card,
  * Read SM required certificates from card.
  *
  * This function uses received path to read a certificate file from
- * card. 
+ * card.
  * No validation is done except that received data is effectively a certificate
  * @param card Pointer to card driver structure
  * @param certpat path to requested certificate
  * @param cert where to store resultig data
- * @return SC_SUCCESS if ok, else error code 
+ * @return SC_SUCCESS if ok, else error code
  */
 static int dnie_read_certificate(sc_card_t * card, char *certpath, X509 ** cert)
 {
 	sc_file_t *file = NULL;
-	sc_path_t *path = NULL;
-	u8 *buffer = NULL;
+	sc_path_t path;
+	u8 *buffer = NULL, *buffer2 = NULL;
 	char *msg = NULL;
 	size_t bufferlen = 0;
 	int res = SC_SUCCESS;
 
 	LOG_FUNC_CALLED(card->ctx);
-	path = (sc_path_t *) calloc(1, sizeof(sc_path_t));
-	if (!path) {
-		msg = "Cannot allocate path data for cert read";
-		res = SC_ERROR_OUT_OF_MEMORY;
-		goto read_cert_end;
-	}
-	sc_format_path(certpath, path);
-	res = dnie_read_file(card, path, &file, &buffer, &bufferlen);
+	sc_format_path(certpath, &path);
+	res = dnie_read_file(card, &path, &file, &buffer, &bufferlen);
 	if (res != SC_SUCCESS) {
 		msg = "Cannot get intermediate CA cert";
 		goto read_cert_end;
 	}
-	*cert = d2i_X509(NULL, (const unsigned char **)&buffer, bufferlen);
+	buffer2 = buffer;
+	*cert = d2i_X509(NULL, (const unsigned char **)&buffer2, bufferlen);
 	if (*cert == NULL) {	/* received data is not a certificate */
 		res = SC_ERROR_OBJECT_NOT_VALID;
-		msg = "Readed data is not a certificate";
+		msg = "Read data is not a certificate";
 		goto read_cert_end;
 	}
 	res = SC_SUCCESS;
 
  read_cert_end:
-	if (file) {
-		sc_file_free(file);
-		file = NULL;
+	if (buffer) {
+		free(buffer);
 		buffer = NULL;
 		bufferlen = 0;
 	}
+	sc_file_free(file);
+	file = NULL;
 	if (msg)
-		sc_log(card->ctx, msg);
+		sc_log(card->ctx, "%s", msg);
 	LOG_FUNC_RETURN(card->ctx, res);
 }
 
@@ -342,6 +416,7 @@ static int dnie_get_root_ca_pubkey(sc_card_t * card, EVP_PKEY ** root_ca_key)
 {
 	int res=SC_SUCCESS;
 	RSA *root_ca_rsa=NULL;
+	BIGNUM *root_ca_rsa_n, *root_ca_rsa_e;
 	LOG_FUNC_CALLED(card->ctx);
 
 	/* compose root_ca_public key with data provided by Dnie Manual */
@@ -351,11 +426,20 @@ static int dnie_get_root_ca_pubkey(sc_card_t * card, EVP_PKEY ** root_ca_key)
 		sc_log(card->ctx, "Cannot create data for root CA public key");
 		return SC_ERROR_OUT_OF_MEMORY;
 	}
-	root_ca_rsa->n = BN_bin2bn(icc_root_ca_modulus,
-				   sizeof(icc_root_ca_modulus), root_ca_rsa->n);
-	root_ca_rsa->e = BN_bin2bn(icc_root_ca_public_exponent,
-				   sizeof(icc_root_ca_public_exponent),
-				   root_ca_rsa->e);
+
+	root_ca_rsa_n = BN_bin2bn(icc_root_ca_modulus, sizeof(icc_root_ca_modulus), NULL);
+	root_ca_rsa_e = BN_bin2bn(icc_root_ca_public_exponent, sizeof(icc_root_ca_public_exponent), NULL);
+	if (RSA_set0_key(root_ca_rsa, root_ca_rsa_n, root_ca_rsa_e, NULL) != 1) {
+		BN_free(root_ca_rsa_n);
+		BN_free(root_ca_rsa_e);
+		if (*root_ca_key)
+			EVP_PKEY_free(*root_ca_key);
+		if (root_ca_rsa)
+			RSA_free(root_ca_rsa);
+		sc_log(card->ctx, "Cannot set RSA valuses for CA public key");
+		return SC_ERROR_INTERNAL;
+	}
+
 	res = EVP_PKEY_assign_RSA(*root_ca_key, root_ca_rsa);
 	if (!res) {
 		if (*root_ca_key)
@@ -411,21 +495,49 @@ static int dnie_get_cvc_ifd_cert(sc_card_t * card, u8 ** cert, size_t * length)
 }
 
 /**
- * Get IFD (Terminal) private key data.
- * 
- * As this is a local (in memory) provider, just get data specified in
- * DNIe's manual and compose an OpenSSL private key structure
+ * Retrieve IFD (application) CVC certificate and length for
+ * the PIN channel.
  *
- * Notice that resulting data should be keept in memory as little as possible
- * Erasing them once used
+ * Returns a byte array with the application's certificate
+ * (in CardVerifiable Certificate format) to be sent to the
+ * card in External Authentication process
+ * As this is local provider, just points to provided static data,
+ * and allways return success
+ *
+ * @param card Pointer to card driver Certificate
+ * @param cert Where to store resulting byte array
+ * @param length len of returned byte array
+ * @return SC_SUCCESS if ok; else error code
+ */
+static int dnie_get_cvc_ifd_cert_pin(sc_card_t * card, u8 ** cert, size_t * length)
+{
+	LOG_FUNC_CALLED(card->ctx);
+	*cert = C_CV_IFDUser_AUT_pin_cert;
+	*length = sizeof(C_CV_IFDUser_AUT_pin_cert);
+	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+}
+
+/**
+ * Get IFD (Terminal) private key data passing the three
+ * arguments (modulus, public and private exponent).
  *
  * @param card pointer to card driver structure
  * @param ifd_privkey where to store IFD private key
+ * @param modulus the byte array used as the modulus of the key
+ * @param modulus_len the length of the modulus
+ * @param public_exponent the byte array for the public exponent
+ * @param public_exponent_len the length of the public exponent
+ * @param private_exponent the byte array for the private exponent
+ * @param private_exponent_len the length of the private exponent
  * @return SC_SUCCESS if ok; else error code
  */
-static int dnie_get_ifd_privkey(sc_card_t * card, EVP_PKEY ** ifd_privkey)
+static int dnie_get_privkey(sc_card_t * card, EVP_PKEY ** ifd_privkey,
+                            u8 * modulus, int modulus_len,
+                            u8 * public_exponent, int public_exponent_len,
+                            u8 * private_exponent, int private_exponent_len)
 {
 	RSA *ifd_rsa=NULL;
+	BIGNUM *ifd_rsa_n, *ifd_rsa_e, *ifd_rsa_d = NULL;
 	int res=SC_SUCCESS;
 
 	LOG_FUNC_CALLED(card->ctx);
@@ -437,13 +549,19 @@ static int dnie_get_ifd_privkey(sc_card_t * card, EVP_PKEY ** ifd_privkey)
 		sc_log(card->ctx, "Cannot create data for IFD private key");
 		return SC_ERROR_OUT_OF_MEMORY;
 	}
-	ifd_rsa->n = BN_bin2bn(ifd_modulus, sizeof(ifd_modulus), ifd_rsa->n);
-	ifd_rsa->e =
-	    BN_bin2bn(ifd_public_exponent, sizeof(ifd_public_exponent),
-		      ifd_rsa->e);
-	ifd_rsa->d =
-	    BN_bin2bn(ifd_private_exponent, sizeof(ifd_private_exponent),
-		      ifd_rsa->d);
+	ifd_rsa_n = BN_bin2bn(modulus, modulus_len, NULL);
+	ifd_rsa_e = BN_bin2bn(public_exponent, public_exponent_len, NULL);
+	ifd_rsa_d = BN_bin2bn(private_exponent, private_exponent_len, NULL);
+	if (RSA_set0_key(ifd_rsa, ifd_rsa_n, ifd_rsa_e, ifd_rsa_d) != 1) {
+		BN_free(ifd_rsa_n);
+		BN_free(ifd_rsa_e);
+		BN_free(ifd_rsa_d);
+		RSA_free(ifd_rsa);
+		EVP_PKEY_free(*ifd_privkey);
+		sc_log(card->ctx, "Cannot set RSA values for IFD private key");
+		return SC_ERROR_INTERNAL;
+	}
+
 	res = EVP_PKEY_assign_RSA(*ifd_privkey, ifd_rsa);
 	if (!res) {
 		if (*ifd_privkey)
@@ -454,6 +572,40 @@ static int dnie_get_ifd_privkey(sc_card_t * card, EVP_PKEY ** ifd_privkey)
 	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 }
 
+/**
+ * Get IFD (Terminal) private key data
+ *
+ * As this is a local (in memory) provider, just get data specified in
+ * DNIe's manual and compose an OpenSSL private key structure
+ *
+ * @param card pointer to card driver structure
+ * @param ifd_privkey where to store IFD private key
+ * @return SC_SUCCESS if ok; else error code
+ */
+static int dnie_get_ifd_privkey(sc_card_t * card, EVP_PKEY ** ifd_privkey)
+{
+	return dnie_get_privkey(card, ifd_privkey, ifd_modulus, sizeof(ifd_modulus),
+				ifd_public_exponent, sizeof(ifd_public_exponent),
+				ifd_private_exponent, sizeof(ifd_private_exponent));
+}
+
+/**
+ * Get IFD (Terminal) private key data for the PIN channel DNIe 3.0
+ *
+ * As this is a local (in memory) provider, just get data specified in
+ * DNIe's manual and compose an OpenSSL private key structure
+ *
+ * @param card pointer to card driver structure
+ * @param ifd_privkey where to store IFD private key
+ * @return SC_SUCCESS if ok; else error code
+ */
+static int dnie_get_ifd_privkey_pin(sc_card_t * card, EVP_PKEY ** ifd_privkey)
+{
+        return dnie_get_privkey(card, ifd_privkey, ifd_pin_modulus, sizeof(ifd_pin_modulus),
+                                ifd_pin_public_exponent, sizeof(ifd_pin_public_exponent),
+                                ifd_pin_private_exponent, sizeof(ifd_pin_private_exponent));
+}
+
 /**
  * Get ICC intermediate CA Certificate from card.
  *
@@ -533,12 +685,32 @@ static int dnie_get_ifd_pubkey_ref(sc_card_t * card, u8 ** buf, size_t * len)
 	return SC_SUCCESS;
 }
 
+/**
+ *  Retrieve public key reference for IFD certificate for the PIN channel.
+ *
+ * This tells the card with in memory key reference is to be used
+ * when CVC cert is sent for external auth procedure
+ * As this driver is for local SM authentication SC_SUCCESS is allways returned
+ *
+ * @param card pointer to card driver structure
+ * @param buf where to store data to be sent
+ * @param len where to store data length
+ * @return SC_SUCCESS if ok; else error code
+ */
+static int dnie_get_ifd_pubkey_ref_pin(sc_card_t * card, u8 ** buf, size_t * len)
+{
+	LOG_FUNC_CALLED(card->ctx);
+	*buf = cvc_ifd_keyref_pin;
+	*len = sizeof(cvc_ifd_keyref_pin);
+	return SC_SUCCESS;
+}
+
 /**
  * Retrieve key reference for ICC privkey.
- * 
- * In local SM stablishment, just retrieve key reference from static 
+ *
+ * In local SM stablishment, just retrieve key reference from static
  * data tables and just return success
- * 
+ *
  * @param card pointer to card driver structure
  * @param buf where to store data
  * @param len where to store data length
@@ -561,9 +733,27 @@ static int dnie_get_icc_privkey_ref(sc_card_t * card, u8 ** buf, size_t * len)
  * @param buf where to store result (8 bytes)
  * @return SC_SUCCESS if ok; else error
  */
-static int dnie_get_sn_ifd(sc_card_t * card, u8 ** buf)
+static int dnie_get_sn_ifd(sc_card_t * card)
 {
-	*buf = sn_ifd;
+	struct sm_cwa_session * sm = &card->sm_ctx.info.session.cwa;
+	memcpy(sm->ifd.sn, sn_ifd, sizeof(sm->ifd.sn));
+	return SC_SUCCESS;
+}
+
+/**
+ * Retrieve SN.IFD (8 bytes left padded with zeroes if required)
+ * for the PIN channel DNIe 3.0.
+ *
+ * In DNIe local SM procedure, just read it from static data and
+ * return SC_SUCCESS
+ *
+ * @param card pointer to card structure
+ * @return SC_SUCCESS if ok; else error
+ */
+static int dnie_get_sn_ifd_pin(sc_card_t * card)
+{
+	struct sm_cwa_session * sm = &card->sm_ctx.info.session.cwa;
+	memcpy(sm->ifd.sn, sn_ifd_pin, sizeof(sm->ifd.sn));
 	return SC_SUCCESS;
 }
 
@@ -573,21 +763,19 @@ static int dnie_get_sn_ifd(sc_card_t * card, u8 ** buf)
  * Just retrieve it from cache and return success
  *
  * @param card pointer to card structure
- * @param buf where to store result (8 bytes)
  * @return SC_SUCCESS if ok; else error
  */
-static int dnie_get_sn_icc(sc_card_t * card, u8 ** buf)
+static int dnie_get_sn_icc(sc_card_t * card)
 {
 	int res=SC_SUCCESS;
 	sc_serial_number_t serial;
+	struct sm_cwa_session * sm = &card->sm_ctx.info.session.cwa;
 
 	res = sc_card_ctl(card, SC_CARDCTL_GET_SERIALNR, &serial);
 	LOG_TEST_RET(card->ctx, res, "Error in gettting serial number");
 	/* copy into sn_icc buffer.Remember that dnie sn has 7 bytes length */
-	memset(&sn_icc[0], 0, sizeof(sn_icc));
-	memcpy(&sn_icc[1], serial.value, 7);
-	/* return data */
-	*buf = &sn_icc[0];
+	memset(sm->icc.sn, 0, sizeof(sm->icc.sn));
+	memcpy(&sm->icc.sn[1], serial.value, 7);
 	return SC_SUCCESS;
 }
 
@@ -595,11 +783,11 @@ static int dnie_get_sn_icc(sc_card_t * card, u8 ** buf)
  * CWA-14890 SM stablisment pre-operations.
  *
  * DNIe needs to get icc serial number at the begin of the sm creation
- * (to avoid breaking key references) so get it an store into serialnr 
+ * (to avoid breaking key references) so get it an store into serialnr
  * cache here.
  *
  * In this way if get_sn_icc is called(), we make sure that no APDU
- * command is to be sent to card, just retrieve it from cache 
+ * command is to be sent to card, just retrieve it from cache
  *
  * @param card pointer to card driver structure
  * @param provider pointer to SM data provider for DNIe
@@ -621,7 +809,7 @@ static int dnie_create_pre_ops(sc_card_t * card, cwa_provider_t * provider)
  * Main entry point for DNIe CWA14890 SM data provider.
  *
  * Return a pointer to DNIe data provider with proper function pointers
- * 
+ *
  * @param card pointer to card driver data structure
  * @return cwa14890 DNIe data provider if success, null on error
  */
@@ -636,11 +824,9 @@ cwa_provider_t *dnie_get_cwa_provider(sc_card_t * card)
 
 	/* pre and post operations */
 	res->cwa_create_pre_ops = dnie_create_pre_ops;
-	res->cwa_create_post_ops = NULL;
 
 	/* Get ICC intermediate CA  path */
-	res->cwa_get_icc_intermediate_ca_cert =
-	    dnie_get_icc_intermediate_ca_cert;
+	res->cwa_get_icc_intermediate_ca_cert = dnie_get_icc_intermediate_ca_cert;
 	/* Get ICC certificate path */
 	res->cwa_get_icc_cert = dnie_get_icc_cert;
 
@@ -658,8 +844,7 @@ cwa_provider_t *dnie_get_cwa_provider(sc_card_t * card)
 	res->cwa_get_root_ca_pubkey_ref = dnie_get_root_ca_pubkey_ref;
 
 	/* Get public key reference for IFD intermediate CA certificate */
-	res->cwa_get_intermediate_ca_pubkey_ref =
-	    dnie_get_intermediate_ca_pubkey_ref;
+	res->cwa_get_intermediate_ca_pubkey_ref = dnie_get_intermediate_ca_pubkey_ref;
 
 	/* Get public key reference for IFD CVC certificate */
 	res->cwa_get_ifd_pubkey_ref = dnie_get_ifd_pubkey_ref;
@@ -673,212 +858,59 @@ cwa_provider_t *dnie_get_cwa_provider(sc_card_t * card)
 	/* Get ICC Serial Number */
 	res->cwa_get_sn_icc = dnie_get_sn_icc;
 
-    /************** operations related with APDU encoding ******************/
-
-	/* pre and post operations */
-	res->cwa_encode_pre_ops = NULL;
-	res->cwa_encode_post_ops = NULL;
-
-    /************** operations related APDU response decoding **************/
-
-	/* pre and post operations */
-	res->cwa_decode_pre_ops = NULL;
-	res->cwa_decode_post_ops = NULL;
-
 	return res;
 }
 
-static int dnie_transmit_apdu_internal(sc_card_t * card, sc_apdu_t * apdu)
+/**
+ * Changes the provider to use the common secure (DNIe 2.0)
+ * channel.
+ *
+ * @param card the card to change the cwa provider for
+ */
+void dnie_change_cwa_provider_to_secure(sc_card_t * card)
 {
-	u8 buf[2048];		/* use for store partial le responses */
-	int res = SC_SUCCESS;
-	cwa_provider_t *provider = NULL;
-	if ((card == NULL) || (card->ctx == NULL) || (apdu == NULL))
-		return SC_ERROR_INVALID_ARGUMENTS;
-	LOG_FUNC_CALLED(card->ctx);
-	provider = GET_DNIE_PRIV_DATA(card)->cwa_provider;
-	memset(buf, 0, sizeof(buf));
+	cwa_provider_t * res = GET_DNIE_PRIV_DATA(card)->cwa_provider;
 
-	/* check if envelope is needed */
-	if (apdu->lc <= card->max_send_size) {
-		int tmp;
-		/* no envelope needed */
-		sc_log(card->ctx, "envelope tx is not required");
-
-		tmp = apdu->cse;	/* save original apdu type */
-		/* if SM is on, assure rx buffer exists and force get_response */
-		if (provider->status.session.state == CWA_SM_ACTIVE) {
-			if (tmp == SC_APDU_CASE_3_SHORT)
-				apdu->cse = SC_APDU_CASE_4_SHORT;
-			if (apdu->resplen == 0) {	/* no response buffer: create */
-				apdu->resp = buf;
-				apdu->resplen = 2048;
-				apdu->le = card->max_recv_size;
-			}
-		}
-		/* call std sc_transmit_apdu */
-		res = sc_transmit_apdu(card, apdu);
-		/* and restore original apdu type */
-		apdu->cse = tmp;
-	} else {
-
-		size_t e_txlen = 0;
-		size_t index = 0;
-		sc_apdu_t *e_apdu = NULL;
-		u8 *e_tx = NULL;
-
-		/* envelope needed */
-		sc_log(card->ctx, "envelope tx required: lc:%d", apdu->lc);
-
-		e_apdu = calloc(1, sizeof(sc_apdu_t));	/* enveloped apdu */
-		e_tx = calloc(7 + apdu->datalen, sizeof(u8));	/* enveloped data */
-		if (!e_apdu || !e_tx)
-			LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
-
-		/* copy apdu info into enveloped data */
-		*(e_tx + 0) = apdu->cla;	/* apdu header */
-		*(e_tx + 1) = apdu->ins;
-		*(e_tx + 2) = apdu->p1;
-		*(e_tx + 3) = apdu->p2;
-		*(e_tx + 4) = 0x00;	/* length in extended format */
-		*(e_tx + 5) = 0xff & (apdu->lc >> 8);
-		*(e_tx + 6) = 0xff & apdu->lc;
-		memcpy(e_tx + 7, apdu->data, apdu->lc);
-		e_txlen = 7 + apdu->lc;
-		/* sc_log(card->ctx, "Data to be enveloped & sent: (%d bytes)\n%s\n===============================================================",e_txlen,sc_dump_hex(e_tx,e_txlen)); */
-		/* split apdu in n chunks of max_send_size len */
-		for (index = 0; index < e_txlen; index += card->max_send_size) {
-			size_t len = MIN(card->max_send_size, e_txlen - index);
-			sc_log(card->ctx, "envelope tx offset:%04X size:%02X",
-			       index, len);
-
-			/* compose envelope apdu command */
-			sc_format_apdu(card, e_apdu, apdu->cse, 0xC2, 0x00,
-				       0x00);
-			e_apdu->cla = 0x90;	/* propietary CLA */
-			e_apdu->data = e_tx + index;
-			e_apdu->lc = len;
-			e_apdu->datalen = len;
-			e_apdu->le = apdu->le;
-			e_apdu->resp = apdu->resp;
-			e_apdu->resplen = apdu->resplen;
-			/* if SM is ON, ensure resp exists, and force getResponse() */
-			if (provider->status.session.state == CWA_SM_ACTIVE) {
-				/* set up proper apdu type */
-				if (e_apdu->cse == SC_APDU_CASE_3_SHORT)
-					e_apdu->cse = SC_APDU_CASE_4_SHORT;
-				/* if no response buffer: create */
-				if (apdu->resplen == 0) {
-					e_apdu->resp = buf;
-					e_apdu->resplen = 2048;
-					e_apdu->le = card->max_recv_size;
-				}
-			}
-			/* send data chunk bypassing apdu wrapping */
-			res = sc_transmit_apdu(card, e_apdu);
-			LOG_TEST_RET(card->ctx, res,
-				     "Error in envelope() send apdu");
-		}		/* for */
-		/* last apdu sent contains response to enveloped cmd */
-		apdu->resp = e_apdu->resp;
-		apdu->resplen = e_apdu->resplen;
-		res = SC_SUCCESS;
-	}
-	LOG_FUNC_RETURN(card->ctx, res);
+	/* redefine different IFD data for secure channel */
+	res->cwa_get_cvc_ifd_cert = dnie_get_cvc_ifd_cert;
+	res->cwa_get_ifd_privkey = dnie_get_ifd_privkey;
+	res->cwa_get_ifd_pubkey_ref = dnie_get_ifd_pubkey_ref;
+	res->cwa_get_sn_ifd = dnie_get_sn_ifd;
 }
 
 /**
- * APDU Wrapping routine.
+ * Changes the provider to use the new PIN (DNIe 3.0)
+ * channel.
  *
- * Called before sc_transmit_apdu() to allowing APDU wrapping
- * If set to NULL no wrapping process will be done
- * Usefull on Secure Messaging APDU encode/decode
- * If returned value is greater than zero, do_single_transmit() 
- * will be called, else means either SC_SUCCESS or error code 
- *
- * NOTE:
- * DNIe doesn't handle apdu chaining; instead apdus with
- * lc>max_send_size are sent by mean of envelope() apdu command
- * So we use this method for
- * - encode and decode SM if SM is on
- * - use envelope instead of apdu chain if lc>max_send_size
- *
- * @param card Pointer to Card Structure
- * @param apdu to be wrapped
- * @return 
- * - positive: use OpenSC's sc_transmit_apdu()
- * - negative: error
- * - zero: success: no need to further transmission
+ * @param card the card to change the cwa provider for
  */
-static int dnie_wrap_apdu(sc_card_t * card, sc_apdu_t * apdu)
+void dnie_change_cwa_provider_to_pin(sc_card_t * card)
 {
-	int res = SC_SUCCESS;
-	sc_apdu_t wrapped;
-	sc_context_t *ctx;
-	cwa_provider_t *provider = NULL;
-	int retries = 3;
+	cwa_provider_t * res = GET_DNIE_PRIV_DATA(card)->cwa_provider;
 
-	if ((card == NULL) || (card->ctx == NULL) || (apdu == NULL))
-		return SC_ERROR_INVALID_ARGUMENTS;
-	ctx=card->ctx;
-	LOG_FUNC_CALLED(ctx);
-	provider = GET_DNIE_PRIV_DATA(card)->cwa_provider;
-	for (retries=3; retries>0; retries--) {
-		/* preserve original apdu to take care of retransmission */
-		memcpy(&wrapped, apdu, sizeof(sc_apdu_t));
-		/* SM is active, encode apdu */
-		if (provider->status.session.state == CWA_SM_ACTIVE) {
-			wrapped.resp = NULL;
-			wrapped.resplen = 0;	/* let get_response() assign space */
-			res = cwa_encode_apdu(card, provider, apdu, &wrapped);
-			LOG_TEST_RET(ctx, res,
-				     "Error in cwa_encode_apdu process");
-		}
-		/* send apdu via envelope() cmd if needed */
-		res = dnie_transmit_apdu_internal(card, &wrapped);
-		/* check for tx errors */
-		LOG_TEST_RET(ctx, res, "Error in dnie_transmit_apdu process");
-
-		/* parse response and handle SM related errors */
-		res=card->ops->check_sw(card,wrapped.sw1,wrapped.sw2);
-		if ( res == SC_ERROR_SM ) {
-			sc_log(ctx,"Detected SM error/collision. Try %d",retries);
-			switch(provider->status.session.state) {
-				/* No SM or creating: collision with other process
-				   just retry as SM error reset ICC SM state */
-				case CWA_SM_NONE: 
-				case CWA_SM_INPROGRESS: 
-					continue;
-				/* SM was active: force restart SM and retry */
-				case CWA_SM_ACTIVE:
-					res=cwa_create_secure_channel(card, provider, CWA_SM_COLD);
-					LOG_TEST_RET(ctx,res,"Cannot re-enable SM");
-					continue;
-			}
-		}
-
-		/* if SM is active; decode apdu */
-		if (provider->status.session.state == CWA_SM_ACTIVE) {
-			apdu->resp = NULL;
-			apdu->resplen = 0;	/* let cwa_decode_response() eval & create size */
-			res = cwa_decode_response(card, provider, &wrapped, apdu);
-			LOG_TEST_RET(ctx, res, "Error in cwa_decode_response process");
-		} else {
-			/* memcopy result to original apdu */
-			memcpy(apdu, &wrapped, sizeof(sc_apdu_t));
-		}
-		LOG_FUNC_RETURN(ctx, res);
-	}
-	sc_log(ctx,"Too many retransmissions. Abort and return");
-	LOG_FUNC_RETURN(ctx, SC_ERROR_INTERNAL);
+	/* redefine different IFD data for PIN channel */
+	res->cwa_get_cvc_ifd_cert = dnie_get_cvc_ifd_cert_pin;
+	res->cwa_get_ifd_privkey = dnie_get_ifd_privkey_pin;
+	res->cwa_get_ifd_pubkey_ref = dnie_get_ifd_pubkey_ref_pin;
+	res->cwa_get_sn_ifd = dnie_get_sn_ifd_pin;
 }
 
-int dnie_transmit_apdu(sc_card_t * card, sc_apdu_t * apdu)
+void dnie_format_apdu(sc_card_t *card, sc_apdu_t *apdu,
+			int cse, int ins, int p1, int p2, int le, int lc,
+			unsigned char * resp, size_t resplen,
+			const unsigned char * data, size_t datalen)
 {
-	int res = SC_SUCCESS;
-	res = dnie_wrap_apdu(card, apdu);
-	if (res <= 0) return res;
-	return sc_transmit_apdu(card, apdu);
+	sc_format_apdu(card, apdu, cse, ins, p1, p2);
+	apdu->le = le;
+	apdu->lc = lc;
+	if (resp != NULL) {
+		apdu->resp = resp;
+		apdu->resplen = resplen;
+	}
+	if (data != NULL) {
+		apdu->data = data;
+		apdu->datalen = datalen;
+	}
 }
 
 #endif				/* HAVE_OPENSSL */
