@@ -91,7 +91,7 @@ static void sc_do_log_va(sc_context_t *ctx, int level, const char *file, int lin
 	gettimeofday (&tv, NULL);
 	tm = localtime (&tv.tv_sec);
 	strftime (time_string, sizeof(time_string), "%H:%M:%S", tm);
-	r = snprintf(p, left, "0x%lx %s.%03ld ", (unsigned long)pthread_self(), time_string, tv.tv_usec / 1000);
+	r = snprintf(p, left, "0x%lx %s.%03ld ", (unsigned long)pthread_self(), time_string, (long)tv.tv_usec / 1000);
 #endif
 	p += r;
 	left -= r;
@@ -112,13 +112,11 @@ static void sc_do_log_va(sc_context_t *ctx, int level, const char *file, int lin
 	if (r < 0)
 		return;
 
-#ifdef _WIN32
-	if (ctx->debug_filename)   {
+	if (ctx->reopen_log_file)   {
 		r = sc_ctx_log_to_file(ctx, ctx->debug_filename);
 		if (r < 0)
 			return;
 	}
-#endif
 
 	outf = ctx->debug_file;
 	if (outf == NULL)
@@ -130,15 +128,11 @@ static void sc_do_log_va(sc_context_t *ctx, int level, const char *file, int lin
 		fprintf(outf, "\n");
 	fflush(outf);
 
-#ifdef _WIN32
-	if (ctx->debug_filename)   {
-		if (ctx->debug_file && (ctx->debug_file != stderr && ctx->debug_file != stdout))   {
+	if (ctx->reopen_log_file)   {
+		if (ctx->debug_file && (ctx->debug_file != stderr && ctx->debug_file != stdout))
 			fclose(ctx->debug_file);
-			ctx->debug_file = NULL;
-		}
+		ctx->debug_file = NULL;
 	}
-#endif
-
 
 	return;
 }
@@ -147,9 +141,9 @@ void _sc_debug(struct sc_context *ctx, int level, const char *format, ...)
 {
 	va_list ap;
 
-        va_start(ap, format);
-        sc_do_log_va(ctx, level, NULL, 0, NULL, format, ap);
-        va_end(ap);
+	va_start(ap, format);
+	sc_do_log_va(ctx, level, NULL, 0, NULL, format, ap);
+	va_end(ap);
 }
 
 void _sc_log(struct sc_context *ctx, const char *format, ...)
@@ -161,6 +155,28 @@ void _sc_log(struct sc_context *ctx, const char *format, ...)
 	va_end(ap);
 }
 
+void _sc_debug_hex(sc_context_t *ctx, int type, const char *file, int line,
+		const char *func, const char *label, const u8 *data, size_t len)
+{
+	size_t blen = len * 5 + 128;
+	char *buf = malloc(blen);
+	if (buf == NULL)
+		return;
+
+	sc_hex_dump(ctx, type, data, len, buf, blen);
+
+	if (label)
+		sc_do_log(ctx, type, file, line, func,
+			"\n%s (%u byte%s):\n%s",
+			label, (unsigned int) len, len==1?"":"s", buf);
+	else
+		sc_do_log(ctx, type, file, line, func,
+			"%u byte%s:\n%s",
+			(unsigned int) len, len==1?"":"s", buf);
+
+	free(buf);
+}
+
 /* Although not used, we need this for consistent exports */
 void sc_hex_dump(struct sc_context *ctx, int level, const u8 * in, size_t count, char *buf, size_t len)
 {
@@ -170,7 +186,9 @@ void sc_hex_dump(struct sc_context *ctx, int level, const u8 * in, size_t count,
 	if (!ctx || ctx->debug < level)
 		return;
 
-	assert(buf != NULL && (in != NULL || count == 0));
+	if (buf == NULL || (in == NULL && count != 0)) {
+		return;
+	}
 	buf[0] = 0;
 	if ((count * 5) > len)
 		return;
