@@ -1,7 +1,7 @@
 /*
  * pkc15-algo.c: ASN.1 handling for algorithm IDs and parameters
  *
- * Copyright (C) 2001, 2002  Olaf Kirch <okir@lst.de>
+ * Copyright (C) 2001, 2002  Olaf Kirch <okir@suse.de>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -267,7 +267,9 @@ asn1_decode_ec_params(sc_context_t *ctx, void **paramp,
 	struct sc_asn1_entry asn1_ec_params[4];
 	struct sc_ec_parameters *ecp;
 
-	sc_debug(ctx, SC_LOG_DEBUG_ASN1, "DEE - asn1_decode_ec_params %p:%d %d", buf, buflen, depth);
+	sc_debug(ctx, SC_LOG_DEBUG_ASN1,
+		 "DEE - asn1_decode_ec_params %p:%"SC_FORMAT_LEN_SIZE_T"u %d",
+		 buf, buflen, depth);
 
 	memset(&curve, 0, sizeof(curve));
 
@@ -421,7 +423,7 @@ static struct sc_asn1_pkcs15_algorithm_info algorithm_table[] = {
 			asn1_encode_ec_params,
 			asn1_free_ec_params },
 #endif
-/* TODO: -DEE Not clear of we need the next five or not */
+/* TODO: -DEE Not clear if we need the next five or not */
 #ifdef SC_ALGORITHM_ECDSA_SHA1
 	/* Note RFC 3279 says no ecParameters */
 	{ SC_ALGORITHM_ECDSA_SHA1, {{ 1, 2, 840, 10045, 4, 1, -1}}, NULL, NULL, NULL},
