@@ -35,11 +35,7 @@
 #include "internal.h"
 #include "asn1.h"
 #include "pkcs15.h"
-
-static const struct sc_asn1_entry     c_asn1_data_object[] = {
-        { "dataObject", SC_ASN1_OCTET_STRING, SC_ASN1_TAG_OCTET_STRING, 0, NULL, NULL },
-        { NULL, 0, 0, 0, NULL, NULL }
-};
+#include "common/compat_strnlen.h"
 
 
 int
@@ -157,7 +153,7 @@ int sc_pkcs15_encode_dodf_entry(sc_context_t *ctx,
 	size_t label_len;
 
 	info = (struct sc_pkcs15_data_info *) obj->data;
-	label_len = strlen(info->app_label);
+	label_len = strnlen(info->app_label, sizeof info->app_label);
 
 	sc_copy_asn1_entry(c_asn1_com_data_attr, asn1_com_data_attr);
 	sc_copy_asn1_entry(c_asn1_type_data_attr, asn1_type_data_attr);
