@@ -22,17 +22,20 @@
 #include "config.h"
 #endif
 
+#include <assert.h>
+#include <fcntl.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
-#include <stdio.h>
-#include <fcntl.h>
-#include <assert.h>
 
-#include "internal.h"
-#include "asn1.h"
-#include "pkcs15.h"
+#ifdef _WIN32
+#include <io.h>
+#else
+#include <unistd.h>
+#endif
 
 #ifdef ENABLE_OPENSSL
+#include <openssl/bn.h>
 #include <openssl/rsa.h>
 #include <openssl/dsa.h>
 #include <openssl/evp.h>
@@ -44,6 +47,11 @@
 #endif
 #endif
 
+#include "internal.h"
+#include "asn1.h"
+#include "pkcs15.h"
+
+
 #define C_ASN1_PKINFO_ATTR_SIZE 3
 static const struct sc_asn1_entry c_asn1_pkinfo[C_ASN1_PKINFO_ATTR_SIZE] = {
 		{ "algorithm", SC_ASN1_ALGORITHM_ID,  SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
@@ -718,7 +726,8 @@ sc_pkcs15_decode_pubkey_ec(sc_context_t *ctx,
 	if (*ecpoint_data != 0x04)
 		LOG_TEST_RET(ctx, SC_ERROR_NOT_SUPPORTED, "Supported only uncompressed EC pointQ value");
 
-	sc_log(ctx, "decode-EC key=%p, buf=%p, buflen=%d", key, buf, buflen);
+	sc_log(ctx, "decode-EC key=%p, buf=%p, buflen=%"SC_FORMAT_LEN_SIZE_T"u",
+	       key, buf, buflen);
 
 	key->ecpointQ.len = ecpoint_len;
 	key->ecpointQ.value = ecpoint_data;
@@ -726,7 +735,7 @@ sc_pkcs15_decode_pubkey_ec(sc_context_t *ctx,
 	/*
 	 * Only get here if raw point is stored in pkcs15 without curve name
 	 * spki has the curvename, so we can get the field_length
-	 * Following only true for curves that are multiple of 8 
+	 * Following only true for curves that are multiple of 8
 	 */
 	key->params.field_length = (ecpoint_len - 1)/2 * 8;
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
@@ -747,7 +756,9 @@ sc_pkcs15_encode_pubkey_ec(sc_context_t *ctx, struct sc_pkcs15_pubkey_ec *key,
 	r = sc_asn1_encode(ctx, asn1_ec_pointQ, buf, buflen);
 	LOG_TEST_RET(ctx, r, "ASN.1 encoding failed");
 
-	sc_log(ctx, "EC key->ecpointQ=%p:%d *buf=%p:%d", key->ecpointQ.value, key->ecpointQ.len, *buf, *buflen);
+	sc_log(ctx,
+	       "EC key->ecpointQ=%p:%"SC_FORMAT_LEN_SIZE_T"u *buf=%p:%"SC_FORMAT_LEN_SIZE_T"u",
+	       key->ecpointQ.value, key->ecpointQ.len, *buf, *buflen);
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
 
@@ -818,7 +829,7 @@ sc_pkcs15_encode_pubkey_as_spki(sc_context_t *ctx, struct sc_pkcs15_pubkey *pubk
 		pkey.value = pubkey->u.ec.ecpointQ.value;
 		pkey.len = 0; /* flag as do not delete */
 
-	        if (pubkey->u.ec.params.named_curve || pubkey->u.ec.params.der.value)   {
+		if (pubkey->u.ec.params.named_curve || pubkey->u.ec.params.der.value)   {
 			struct sc_ec_parameters *ec_params = NULL;
 
 			r = sc_pkcs15_fix_ec_parameters(ctx, &pubkey->u.ec.params);
@@ -892,14 +903,19 @@ int
 sc_pkcs15_read_pubkey(struct sc_pkcs15_card *p15card, const struct sc_pkcs15_object *obj,
 		struct sc_pkcs15_pubkey **out)
 {
-	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_context *ctx;
 	const struct sc_pkcs15_pubkey_info *info = NULL;
 	struct sc_pkcs15_pubkey *pubkey = NULL;
 	unsigned char *data = NULL;
 	size_t	len;
 	int	algorithm, r;
 
-	assert(p15card != NULL && obj != NULL && out != NULL);
+	if (p15card == NULL || p15card->card == NULL || p15card->card->ops == NULL
+			|| obj == NULL || out == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	ctx = p15card->card->ctx;
+
 	LOG_FUNC_CALLED(ctx);
 	sc_log(ctx, "Public key type 0x%X", obj->type);
 
@@ -986,7 +1002,9 @@ err:
 static int
 sc_pkcs15_dup_bignum (struct sc_pkcs15_bignum *dst, struct sc_pkcs15_bignum *src)
 {
-	assert(dst && src);
+	if (!dst || !src) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	if (src->data && src->len)   {
 		dst->data = calloc(1, src->len);
@@ -1007,7 +1025,9 @@ sc_pkcs15_pubkey_from_prvkey(struct sc_context *ctx, struct sc_pkcs15_prkey *prv
 	struct sc_pkcs15_pubkey *pubkey = NULL;
 	int rv = SC_SUCCESS;
 
-	assert(prvkey && out);
+	if (!prvkey || !out) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	*out = NULL;
 	pubkey = calloc(1, sizeof(struct sc_pkcs15_pubkey));
@@ -1034,8 +1054,10 @@ sc_pkcs15_pubkey_from_prvkey(struct sc_context *ctx, struct sc_pkcs15_prkey *prv
 		break;
 	case SC_ALGORITHM_EC:
 		pubkey->u.ec.ecpointQ.value = malloc(prvkey->u.ec.ecpointQ.len);
-		if (!pubkey->u.ec.ecpointQ.value)
+		if (!pubkey->u.ec.ecpointQ.value) {
+			sc_pkcs15_free_pubkey(pubkey);
 			LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+		}
 		memcpy(pubkey->u.ec.ecpointQ.value, prvkey->u.ec.ecpointQ.value, prvkey->u.ec.ecpointQ.len);
 		pubkey->u.ec.ecpointQ.len = prvkey->u.ec.ecpointQ.len;
 		break;
@@ -1061,7 +1083,9 @@ sc_pkcs15_dup_pubkey(struct sc_context *ctx, struct sc_pkcs15_pubkey *key, struc
 	u8* alg;
 	size_t alglen;
 
-	assert(key && out);
+	if (!key || !out) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	*out = NULL;
 	pubkey = calloc(1, sizeof(struct sc_pkcs15_pubkey));
@@ -1074,6 +1098,8 @@ sc_pkcs15_dup_pubkey(struct sc_context *ctx, struct sc_pkcs15_pubkey *key, struc
 		rv = sc_asn1_encode_algorithm_id(ctx, &alg, &alglen,key->alg_id, 0);
 		if (rv == SC_SUCCESS) {
 			pubkey->alg_id = (struct sc_algorithm_id *)calloc(1, sizeof(struct sc_algorithm_id));
+			if (pubkey->alg_id == NULL)
+				LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 			rv = sc_asn1_decode_algorithm_id(ctx, alg, alglen, pubkey->alg_id, 0);
 			free(alg);
 		}
@@ -1136,7 +1162,9 @@ sc_pkcs15_dup_pubkey(struct sc_context *ctx, struct sc_pkcs15_pubkey *key, struc
 void
 sc_pkcs15_erase_pubkey(struct sc_pkcs15_pubkey *key)
 {
-	assert(key != NULL);
+	if (key == NULL) {
+		return;
+	}
 	if (key->alg_id) {
 		sc_asn1_clear_algorithm_id(key->alg_id);
 		free(key->alg_id);
@@ -1190,6 +1218,10 @@ sc_pkcs15_free_pubkey_info(sc_pkcs15_pubkey_info_t *info)
 {
 	if (info->subject.value)
 		free(info->subject.value);
+	if (info->direct.spki.value)
+		free(info->direct.spki.value);
+	if (info->direct.raw.value)
+		free(info->direct.raw.value);
 	sc_pkcs15_free_key_params(&info->params);
 	free(info);
 }
@@ -1201,16 +1233,21 @@ sc_pkcs15_read_der_file(sc_context_t *ctx, char * filename,
 {
 	int r;
 	int f = -1;
-	size_t len;
-	u8 tagbuf[16]; /* enough to read in the tag and length */
+	size_t len, offs;
+	u8 tagbuf[1024]; /* enough to read in the tag and length */
 	u8 * rbuf = NULL;
-	size_t rbuflen;
-	const u8 * body;
+	size_t rbuflen = 0;
+	const u8 * body = NULL;
 	size_t bodylen;
 	unsigned int cla_out, tag_out;
-	*buf = NULL;
 
 	LOG_FUNC_CALLED(ctx);
+	if (!buf || !buflen)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
+
+	*buf = NULL;
+	*buflen = 0;
+
 	f = open(filename, O_RDONLY);
 	if (f < 0) {
 		r = SC_ERROR_FILE_NOT_FOUND;
@@ -1224,14 +1261,24 @@ sc_pkcs15_read_der_file(sc_context_t *ctx, char * filename,
 		goto out;
 	}
 	len = r;
+
 	body = tagbuf;
-	if (sc_asn1_read_tag(&body, 0xfffff, &cla_out, &tag_out, &bodylen) != SC_SUCCESS) {
-		sc_log(ctx, "DER problem");
+	r = sc_asn1_read_tag(&body, len, &cla_out, &tag_out, &bodylen);
+	if (r != SC_SUCCESS)
+		goto out;
+
+	if (tag_out == SC_ASN1_TAG_EOC || body == NULL)   {
+		r = SC_SUCCESS;
+		goto out;
+	}
+
+	offs = body - tagbuf;
+	if (offs > len || offs < 2 || offs > offs + bodylen)   {
 		r = SC_ERROR_INVALID_ASN1_OBJECT;
 		goto out;
 	}
 
-	rbuflen = body - tagbuf + bodylen;
+	rbuflen = offs + bodylen;
 	rbuf = malloc(rbuflen);
 	if (rbuf == NULL) {
 		r = SC_ERROR_OUT_OF_MEMORY;
@@ -1252,7 +1299,7 @@ sc_pkcs15_read_der_file(sc_context_t *ctx, char * filename,
 	*buf = rbuf;
 	rbuf = NULL;
 	r = rbuflen;
-	out:
+out:
 	if (f >= 0)
 		close(f);
 
@@ -1275,11 +1322,15 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 	unsigned char *tmp_buf = NULL;
 	int r;
 
-	sc_log(ctx, "sc_pkcs15_pubkey_from_spki_fields() called: %p:%d\n%s", buf, buflen, sc_dump_hex(buf, buflen));
+	sc_log(ctx,
+	       "sc_pkcs15_pubkey_from_spki_fields() called: %p:%"SC_FORMAT_LEN_SIZE_T"u\n%s",
+	       buf, buflen, sc_dump_hex(buf, buflen));
 
 	tmp_buf = malloc(buflen);
-	if (!tmp_buf)
-		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+	if (!tmp_buf) {
+		r = SC_ERROR_OUT_OF_MEMORY;
+		LOG_TEST_GOTO_ERR(ctx, r, "");
+	}
 	memcpy(tmp_buf, buf, buflen);
 
 	if ((*tmp_buf & SC_ASN1_TAG_CONTEXT))
@@ -1287,8 +1338,10 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 
 	memset(&pk_alg, 0, sizeof(pk_alg));
 	pubkey = calloc(1, sizeof(sc_pkcs15_pubkey_t));
-	if (pubkey == NULL)
-		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+	if (pubkey == NULL) {
+		r = SC_ERROR_OUT_OF_MEMORY;
+		LOG_TEST_GOTO_ERR(ctx, r, "");
+	}
 	*outpubkey = pubkey;
 
 	sc_copy_asn1_entry(c_asn1_pkinfo, asn1_pkinfo);
@@ -1297,11 +1350,13 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 	sc_format_asn1_entry(asn1_pkinfo + 1, &pk.value, &pk.len, 0);
 
 	r = sc_asn1_decode(ctx, asn1_pkinfo, tmp_buf, buflen, NULL, NULL);
-	LOG_TEST_RET(ctx, r, "ASN.1 parsing of subjectPubkeyInfo failed");
+	LOG_TEST_GOTO_ERR(ctx, r, "ASN.1 parsing of subjectPubkeyInfo failed");
 
 	pubkey->alg_id = calloc(1, sizeof(struct sc_algorithm_id));
-	if (pubkey->alg_id == NULL)
-		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+	if (pubkey->alg_id == NULL) {
+		r = SC_ERROR_OUT_OF_MEMORY;
+		LOG_TEST_GOTO_ERR(ctx, r, "");
+	}
 
 	memcpy(pubkey->alg_id, &pk_alg, sizeof(struct sc_algorithm_id));
 	pubkey->algorithm = pk_alg.algorithm;
@@ -1314,7 +1369,7 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 		/* EC public key is not encapsulated into BIT STRING -- it's a BIT STRING */
 		/*
 		 * sc_pkcs15_fix_ec_parameters below will set field_length from curve.
-		 * if no alg_id->params, assume field_length is multiple of 8 
+		 * if no alg_id->params, assume field_length is multiple of 8
 		 */
 		pubkey->u.ec.params.field_length = (pk.len - 1) / 2 * 8;
 
@@ -1322,13 +1377,15 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 			struct sc_ec_parameters *ecp = (struct sc_ec_parameters *)pubkey->alg_id->params;
 
 			pubkey->u.ec.params.der.value = malloc(ecp->der.len);
-			if (pubkey->u.ec.params.der.value == NULL)
-				LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+			if (pubkey->u.ec.params.der.value == NULL) {
+				r = SC_ERROR_OUT_OF_MEMORY;
+				LOG_TEST_GOTO_ERR(ctx, r, "");
+			}
 
 			memcpy(pubkey->u.ec.params.der.value, ecp->der.value, ecp->der.len);
 			pubkey->u.ec.params.der.len = ecp->der.len;
 			r = sc_pkcs15_fix_ec_parameters(ctx, &pubkey->u.ec.params);
-			LOG_TEST_RET(ctx, r, "failed to fix EC parameters");
+			LOG_TEST_GOTO_ERR(ctx, r, "failed to fix EC parameters");
 		}
 
 		pubkey->u.ec.ecpointQ.value = malloc(pk.len);
@@ -1340,15 +1397,16 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 	else   {
 		/* Public key is expected to be encapsulated into BIT STRING */
 		r = sc_pkcs15_decode_pubkey(ctx, pubkey, pk.value, pk.len);
-		LOG_TEST_RET(ctx, r, "ASN.1 parsing of subjectPubkeyInfo failed");
+		LOG_TEST_GOTO_ERR(ctx, r, "ASN.1 parsing of subjectPubkeyInfo failed");
 	}
 
+err:
 	if (pk.value)
 		free(pk.value);
 	if (tmp_buf)
 		free(tmp_buf);
 
-	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
+	LOG_FUNC_RETURN(ctx, r);
 }
 
 
@@ -1369,6 +1427,8 @@ sc_pkcs15_pubkey_from_spki_sequence(struct sc_context *ctx, const unsigned char
 
 	if(outpubkey)
 		*outpubkey = pubkey;
+	else
+		free(pubkey);
 
 	LOG_FUNC_RETURN(ctx, r);
 }
@@ -1475,7 +1535,8 @@ sc_pkcs15_fix_ec_parameters(struct sc_context *ctx, struct sc_ec_parameters *ecp
 			sc_format_oid(&ecparams->id, ec_curve_infos[ii].oid_str);
 
 		ecparams->field_length = ec_curve_infos[ii].size;
-		sc_log(ctx, "Curve length %i", ecparams->field_length);
+		sc_log(ctx, "Curve length %"SC_FORMAT_LEN_SIZE_T"u",
+		       ecparams->field_length);
 	}
 	else if (ecparams->named_curve)   {	/* it can be name of curve or OID in ASCII form */
 		for (ii=0; ec_curve_infos[ii].name; ii++)   {
@@ -1512,14 +1573,19 @@ sc_pkcs15_convert_pubkey(struct sc_pkcs15_pubkey *pkcs15_key, void *evp_key)
 {
 #ifdef ENABLE_OPENSSL
 	EVP_PKEY *pk = (EVP_PKEY *)evp_key;
+	int pk_type;
+	pk_type = EVP_PKEY_base_id(pk);
 
-	switch (pk->type) {
+	switch (pk_type) {
 	case EVP_PKEY_RSA: {
 		struct sc_pkcs15_pubkey_rsa *dst = &pkcs15_key->u.rsa;
 		RSA *src = EVP_PKEY_get1_RSA(pk);
+		const BIGNUM *src_n, *src_e;
+
+		RSA_get0_key(src, &src_n, &src_e, NULL);
 
 		pkcs15_key->algorithm = SC_ALGORITHM_RSA;
-		if (!sc_pkcs15_convert_bignum(&dst->modulus, src->n) || !sc_pkcs15_convert_bignum(&dst->exponent, src->e))
+		if (!sc_pkcs15_convert_bignum(&dst->modulus, src_n) || !sc_pkcs15_convert_bignum(&dst->exponent, src_e))
 			return SC_ERROR_INVALID_DATA;
 		RSA_free(src);
 		break;
@@ -1527,12 +1593,16 @@ sc_pkcs15_convert_pubkey(struct sc_pkcs15_pubkey *pkcs15_key, void *evp_key)
 	case EVP_PKEY_DSA: {
 		struct sc_pkcs15_pubkey_dsa *dst = &pkcs15_key->u.dsa;
 		DSA *src = EVP_PKEY_get1_DSA(pk);
+		const BIGNUM *src_pub_key, *src_priv_key, *src_p, *src_q, *src_g;
+
+		DSA_get0_key(src, &src_pub_key, &src_priv_key);
+		DSA_get0_pqg(src, &src_p, &src_q, &src_g);
 
 		pkcs15_key->algorithm = SC_ALGORITHM_DSA;
-		sc_pkcs15_convert_bignum(&dst->pub, src->pub_key);
-		sc_pkcs15_convert_bignum(&dst->p, src->p);
-		sc_pkcs15_convert_bignum(&dst->q, src->q);
-		sc_pkcs15_convert_bignum(&dst->g, src->g);
+		sc_pkcs15_convert_bignum(&dst->pub, src_pub_key);
+		sc_pkcs15_convert_bignum(&dst->p, src_p);
+		sc_pkcs15_convert_bignum(&dst->q, src_q);
+		sc_pkcs15_convert_bignum(&dst->g, src_g);
 		DSA_free(src);
 		break;
 	}
