@@ -30,6 +30,7 @@
 #include "pkcs15.h"
 #include "asn1.h"
 #include "common/compat_strlcpy.h"
+#include "common/compat_strnlen.h"
 
 #include "card-sc-hsm.h"
 
@@ -179,6 +180,53 @@ static const struct sc_asn1_entry c_asn1_req[C_ASN1_REQ_SIZE] = {
 
 
 
+static int read_file(sc_pkcs15_card_t * p15card, u8 fid[2],
+		u8 *efbin, size_t *len, int optional)
+{
+	sc_path_t path;
+	int r;
+
+	sc_path_set(&path, SC_PATH_TYPE_FILE_ID, fid, 2, 0, 0);
+	/* look this up with our AID */
+	path.aid = sc_hsm_aid;
+	/* we don't have a pre-known size of the file */
+	path.count = -1;
+	if (!p15card->opts.use_file_cache
+			|| SC_SUCCESS != sc_pkcs15_read_cached_file(p15card, &path, &efbin, len)) {
+		/* avoid re-selection of SC-HSM */
+		path.aid.len = 0;
+		r = sc_select_file(p15card->card, &path, NULL);
+		if (r < 0) {
+			sc_log(p15card->card->ctx, "Could not select EF");
+		} else {
+			r = sc_read_binary(p15card->card, 0, efbin, *len, 0);
+		}
+
+		if (r < 0) {
+			sc_log(p15card->card->ctx, "Could not read EF");
+			if (!optional) {
+				return r;
+			}
+			/* optional files are saved as empty files to avoid card
+			 * transactions. Parsing the file's data will reveal that they were
+			 * missing. */
+			*len = 0;
+		} else {
+			*len = r;
+		}
+
+		if (p15card->opts.use_file_cache) {
+			/* save this with our AID */
+			path.aid = sc_hsm_aid;
+			sc_pkcs15_cache_file(p15card, &path, efbin, *len);
+		}
+	}
+
+	return SC_SUCCESS;
+}
+
+
+
 /*
  * Decode a card verifiable certificate as defined in TR-03110.
  */
@@ -298,10 +346,10 @@ int sc_pkcs15emu_sc_hsm_encode_cvc(sc_pkcs15_card_t * p15card,
 	}
 
 	sc_format_asn1_entry(asn1_cvc_body    , &cvc->cpi, NULL, 1);
-	lencar = strlen(cvc->car);
+	lencar = strnlen(cvc->car, sizeof cvc->car);
 	sc_format_asn1_entry(asn1_cvc_body + 1, &cvc->car, &lencar, 1);
 	sc_format_asn1_entry(asn1_cvc_body + 2, &asn1_cvc_pubkey, NULL, 1);
-	lenchr = strlen(cvc->chr);
+	lenchr = strnlen(cvc->chr, sizeof cvc->chr);
 	sc_format_asn1_entry(asn1_cvc_body + 3, &cvc->chr, &lenchr, 1);
 
 	sc_format_asn1_entry(asn1_cvcert    , &asn1_cvc_body, NULL, 1);
@@ -392,8 +440,10 @@ static int sc_pkcs15emu_sc_hsm_get_ec_public_key(struct sc_context *ctx, sc_cvc_
 
 	ecp->der.len = oid->len + 2;
 	ecp->der.value = calloc(ecp->der.len, 1);
-	if (!ecp->der.value)
+	if (!ecp->der.value) {
+		free(ecp);
 		return SC_ERROR_OUT_OF_MEMORY;
+	}
 
 	*(ecp->der.value + 0) = 0x06;
 	*(ecp->der.value + 1) = (u8)oid->len;
@@ -401,8 +451,11 @@ static int sc_pkcs15emu_sc_hsm_get_ec_public_key(struct sc_context *ctx, sc_cvc_
 	ecp->type = 1;		// Named curve
 
 	pubkey->alg_id = (struct sc_algorithm_id *)calloc(1, sizeof(struct sc_algorithm_id));
-	if (!pubkey->alg_id)
+	if (!pubkey->alg_id) {
+		free(ecp->der.value);
+		free(ecp);
 		return SC_ERROR_OUT_OF_MEMORY;
+	}
 
 	pubkey->alg_id->algorithm = SC_ALGORITHM_EC;
 	pubkey->alg_id->params = ecp;
@@ -419,6 +472,7 @@ static int sc_pkcs15emu_sc_hsm_get_ec_public_key(struct sc_context *ctx, sc_cvc_
 	memcpy(pubkey->u.ec.params.der.value, ecp->der.value, ecp->der.len);
 	pubkey->u.ec.params.der.len = ecp->der.len;
 
+	/* FIXME: check return value? */
 	sc_pkcs15_fix_ec_parameters(ctx, &pubkey->u.ec.params);
 
 	return SC_SUCCESS;
@@ -475,28 +529,21 @@ void sc_pkcs15emu_sc_hsm_free_cvc(sc_cvc_t *cvc)
 
 
 
-static int sc_pkcs15emu_sc_hsm_add_pubkey(sc_pkcs15_card_t *p15card, sc_pkcs15_prkey_info_t *key_info, char *label)
+static int sc_pkcs15emu_sc_hsm_add_pubkey(sc_pkcs15_card_t *p15card, u8 *efbin, size_t len, sc_pkcs15_prkey_info_t *key_info, char *label)
 {
 	struct sc_context *ctx = p15card->card->ctx;
 	sc_card_t *card = p15card->card;
 	sc_pkcs15_pubkey_info_t pubkey_info;
 	sc_pkcs15_object_t pubkey_obj;
 	struct sc_pkcs15_pubkey pubkey;
-	u8 efbin[1024];
 	sc_cvc_t cvc;
 	u8 *cvcpo;
-	size_t cvclen;
 	int r;
 
-	/* EF.CERT is selected */
-	r = sc_read_binary(p15card->card, 0, efbin, sizeof(efbin), 0);
-	LOG_TEST_RET(ctx, r, "Could not read CSR from EF");
-
 	cvcpo = efbin;
-	cvclen = r;
 
 	memset(&cvc, 0, sizeof(cvc));
-	r = sc_pkcs15emu_sc_hsm_decode_cvc(p15card, (const u8 **)&cvcpo, &cvclen, &cvc);
+	r = sc_pkcs15emu_sc_hsm_decode_cvc(p15card, (const u8 **)&cvcpo, &len, &cvc);
 	LOG_TEST_RET(ctx, r, "Could decode certificate signing request");
 
 	memset(&pubkey, 0, sizeof(pubkey));
@@ -518,10 +565,12 @@ static int sc_pkcs15emu_sc_hsm_add_pubkey(sc_pkcs15_card_t *p15card, sc_pkcs15_p
 
 	if (pubkey.algorithm == SC_ALGORITHM_RSA) {
 		pubkey_info.modulus_length = pubkey.u.rsa.modulus.len << 3;
+		pubkey_info.usage = SC_PKCS15_PRKEY_USAGE_ENCRYPT|SC_PKCS15_PRKEY_USAGE_VERIFY|SC_PKCS15_PRKEY_USAGE_WRAP;
 		r = sc_pkcs15emu_add_rsa_pubkey(p15card, &pubkey_obj, &pubkey_info);
 	} else {
 		/* TODO fix if support of non multiple of 8 curves are added */
 		pubkey_info.field_length = cvc.primeOrModuluslen << 3;
+		pubkey_info.usage = SC_PKCS15_PRKEY_USAGE_VERIFY;
 		r = sc_pkcs15emu_add_ec_pubkey(p15card, &pubkey_obj, &pubkey_info);
 	}
 	LOG_TEST_RET(ctx, r, "Could not add public key");
@@ -544,10 +593,9 @@ static int sc_pkcs15emu_sc_hsm_add_prkd(sc_pkcs15_card_t * p15card, u8 keyid) {
 	sc_pkcs15_object_t cert_obj;
 	struct sc_pkcs15_object prkd;
 	sc_pkcs15_prkey_info_t *key_info;
-	sc_file_t *file = NULL;
-	sc_path_t path;
 	u8 fid[2];
-	u8 efbin[512];
+	/* enough to hold a complete certificate */
+	u8 efbin[4096];
 	u8 *ptr;
 	size_t len;
 	int r;
@@ -556,21 +604,13 @@ static int sc_pkcs15emu_sc_hsm_add_prkd(sc_pkcs15_card_t * p15card, u8 keyid) {
 	fid[1] = keyid;
 
 	/* Try to select a related EF containing the PKCS#15 description of the key */
-	sc_path_set(&path, SC_PATH_TYPE_FILE_ID, fid, sizeof(fid), 0, 0);
-	r = sc_select_file(card, &path, &file);
-
-	if (r != SC_SUCCESS) {
-		return SC_SUCCESS;
-	}
-
-	sc_file_free(file);
-	r = sc_read_binary(p15card->card, 0, efbin, sizeof(efbin), 0);
+	len = sizeof efbin;
+	r = read_file(p15card, fid, efbin, &len, 1);
 	LOG_TEST_RET(card->ctx, r, "Could not read EF.PRKD");
 
-	memset(&prkd, 0, sizeof(prkd));
 	ptr = efbin;
-	len = r;
 
+	memset(&prkd, 0, sizeof(prkd));
 	r = sc_pkcs15_decode_prkdf_entry(p15card, &prkd, (const u8 **)&ptr, &len);
 	LOG_TEST_RET(card->ctx, r, "Could not decode EF.PRKD");
 
@@ -585,6 +625,7 @@ static int sc_pkcs15emu_sc_hsm_add_prkd(sc_pkcs15_card_t * p15card, u8 keyid) {
 
 	key_info = (sc_pkcs15_prkey_info_t *)prkd.data;
 	key_info->key_reference = keyid;
+	key_info->path.aid.len = 0;
 
 	if (prkd.type == SC_PKCS15_TYPE_PRKEY_RSA) {
 		r = sc_pkcs15emu_add_rsa_prkey(p15card, &prkd, key_info);
@@ -597,28 +638,20 @@ static int sc_pkcs15emu_sc_hsm_add_prkd(sc_pkcs15_card_t * p15card, u8 keyid) {
 	/* Check if we also have a certificate for the private key */
 	fid[0] = EE_CERTIFICATE_PREFIX;
 
-	sc_path_set(&path, SC_PATH_TYPE_FILE_ID, fid, sizeof(fid), 0, 0);
-	r = sc_select_file(card, &path, &file);
+	len = sizeof efbin;
+	r = read_file(p15card, fid, efbin, &len, 0);
+	LOG_TEST_RET(card->ctx, r, "Could not read EF");
 
-	if (r != SC_SUCCESS) {
-		return SC_SUCCESS;
-	}
-
-	sc_file_free(file);
-
-	/* Check if the certificate is a X.509 certificate */
-	r = sc_read_binary(p15card->card, 0, efbin, 1, 0);
-
-	if (r < 0) {
-		return SC_SUCCESS;
-	}
+	LOG_TEST_RET(card->ctx, r, "Could not read EF");
 
 	if (efbin[0] == 0x67) {		/* Decode CSR and create public key object */
-		sc_pkcs15emu_sc_hsm_add_pubkey(p15card, key_info, prkd.label);
+		sc_pkcs15emu_sc_hsm_add_pubkey(p15card, efbin, len, key_info, prkd.label);
+		free(key_info);
 		return SC_SUCCESS;		/* Ignore any errors */
 	}
 
 	if (efbin[0] != 0x30) {
+		free(key_info);
 		return SC_SUCCESS;
 	}
 
@@ -626,11 +659,22 @@ static int sc_pkcs15emu_sc_hsm_add_prkd(sc_pkcs15_card_t * p15card, u8 keyid) {
 	memset(&cert_obj, 0, sizeof(cert_obj));
 
 	cert_info.id = key_info->id;
-	cert_info.path = path;
+	sc_path_set(&cert_info.path, SC_PATH_TYPE_FILE_ID, fid, 2, 0, 0);
 	cert_info.path.count = -1;
+	if (p15card->opts.use_file_cache) {
+		/* look this up with our AID, which should already be cached from the
+		 * call to `read_file`. This may have the side effect that OpenSC's
+		 * caching layer re-selects our applet *if the cached file cannot be
+		 * found/used* and we may loose the authentication status. We assume
+		 * that caching works perfectly without this side effect. */
+		cert_info.path.aid = sc_hsm_aid;
+	}
 
 	strlcpy(cert_obj.label, prkd.label, sizeof(cert_obj.label));
 	r = sc_pkcs15emu_add_x509_cert(p15card, &cert_obj, &cert_info);
+
+	free(key_info);
+
 	LOG_TEST_RET(card->ctx, r, "Could not add certificate");
 
 	return SC_SUCCESS;
@@ -646,8 +690,6 @@ static int sc_pkcs15emu_sc_hsm_add_dcod(sc_pkcs15_card_t * p15card, u8 id) {
 	sc_card_t *card = p15card->card;
 	sc_pkcs15_data_info_t *data_info;
 	sc_pkcs15_object_t data_obj;
-	sc_file_t *file = NULL;
-	sc_path_t path;
 	u8 fid[2];
 	u8 efbin[512];
 	const u8 *ptr;
@@ -658,21 +700,13 @@ static int sc_pkcs15emu_sc_hsm_add_dcod(sc_pkcs15_card_t * p15card, u8 id) {
 	fid[1] = id;
 
 	/* Try to select a related EF containing the PKCS#15 description of the data */
-	sc_path_set(&path, SC_PATH_TYPE_FILE_ID, fid, sizeof(fid), 0, 0);
-	r = sc_select_file(card, &path, &file);
-
-	if (r != SC_SUCCESS) {
-		return SC_SUCCESS;
-	}
-
-	sc_file_free(file);
-	r = sc_read_binary(p15card->card, 0, efbin, sizeof(efbin), 0);
+	len = sizeof efbin;
+	r = read_file(p15card, fid, efbin, &len, 1);
 	LOG_TEST_RET(card->ctx, r, "Could not read EF.DCOD");
 
-	memset(&data_obj, 0, sizeof(data_obj));
 	ptr = efbin;
-	len = r;
 
+	memset(&data_obj, 0, sizeof(data_obj));
 	r = sc_pkcs15_decode_dodf_entry(p15card, &data_obj, &ptr, &len);
 	LOG_TEST_RET(card->ctx, r, "Could not decode EF.DCOD");
 
@@ -695,8 +729,6 @@ static int sc_pkcs15emu_sc_hsm_add_cd(sc_pkcs15_card_t * p15card, u8 id) {
 	sc_card_t *card = p15card->card;
 	sc_pkcs15_cert_info_t *cert_info;
 	sc_pkcs15_object_t obj;
-	sc_file_t *file = NULL;
-	sc_path_t path;
 	u8 fid[2];
 	u8 efbin[512];
 	const u8 *ptr;
@@ -707,21 +739,13 @@ static int sc_pkcs15emu_sc_hsm_add_cd(sc_pkcs15_card_t * p15card, u8 id) {
 	fid[1] = id;
 
 	/* Try to select a related EF containing the PKCS#15 description of the data */
-	sc_path_set(&path, SC_PATH_TYPE_FILE_ID, fid, sizeof(fid), 0, 0);
-	r = sc_select_file(card, &path, &file);
-
-	if (r != SC_SUCCESS) {
-		return SC_SUCCESS;
-	}
-
-	sc_file_free(file);
-	r = sc_read_binary(p15card->card, 0, efbin, sizeof(efbin), 0);
+	len = sizeof efbin;
+	r = read_file(p15card, fid, efbin, &len, 1);
 	LOG_TEST_RET(card->ctx, r, "Could not read EF.DCOD");
 
-	memset(&obj, 0, sizeof(obj));
 	ptr = efbin;
-	len = r;
 
+	memset(&obj, 0, sizeof(obj));
 	r = sc_pkcs15_decode_cdf_entry(p15card, &obj, &ptr, &len);
 	LOG_TEST_RET(card->ctx, r, "Could not decode EF.CD");
 
@@ -739,23 +763,18 @@ static int sc_pkcs15emu_sc_hsm_add_cd(sc_pkcs15_card_t * p15card, u8 id) {
 static int sc_pkcs15emu_sc_hsm_read_tokeninfo (sc_pkcs15_card_t * p15card)
 {
 	sc_card_t *card = p15card->card;
-	sc_file_t *file = NULL;
-	sc_path_t path;
 	int r;
 	u8 efbin[512];
+	size_t len;
 
 	LOG_FUNC_CALLED(card->ctx);
 
 	/* Read token info */
-	sc_path_set(&path, SC_PATH_TYPE_FILE_ID, (u8 *) "\x2F\x03", 2, 0, 0);
-	r = sc_select_file(card, &path, &file);
-	LOG_TEST_RET(card->ctx, r, "Could not select EF.TokenInfo");
-	sc_file_free(file);
-
-	r = sc_read_binary(p15card->card, 0, efbin, sizeof(efbin), 0);
+	len = sizeof efbin;
+	r = read_file(p15card, (u8 *) "\x2F\x03", efbin, &len, 1);
 	LOG_TEST_RET(card->ctx, r, "Could not read EF.TokenInfo");
 
-	r = sc_pkcs15_parse_tokeninfo(card->ctx, p15card->tokeninfo, efbin, r);
+	r = sc_pkcs15_parse_tokeninfo(card->ctx, p15card->tokeninfo, efbin, len);
 	LOG_TEST_RET(card->ctx, r, "Could not decode EF.TokenInfo");
 
 	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
@@ -770,6 +789,7 @@ static int sc_pkcs15emu_sc_hsm_read_tokeninfo (sc_pkcs15_card_t * p15card)
 static int sc_pkcs15emu_sc_hsm_init (sc_pkcs15_card_t * p15card)
 {
 	sc_card_t *card = p15card->card;
+	sc_hsm_private_data_t *priv = (sc_hsm_private_data_t *) card->drv_data;
 	sc_file_t *file = NULL;
 	sc_path_t path;
 	u8 filelist[MAX_EXT_APDU_LENGTH];
@@ -779,7 +799,8 @@ static int sc_pkcs15emu_sc_hsm_init (sc_pkcs15_card_t * p15card)
 	struct sc_app_info *appinfo;
 	struct sc_pkcs15_auth_info pin_info;
 	struct sc_pkcs15_object pin_obj;
-	u8 efbin[512];
+	struct sc_pin_cmd_data pindata;
+	u8 efbin[1024];
 	u8 *ptr;
 	size_t len;
 
@@ -808,16 +829,24 @@ static int sc_pkcs15emu_sc_hsm_init (sc_pkcs15_card_t * p15card)
 	sc_file_free(file);
 
 	/* Read device certificate to determine serial number */
-	sc_path_set(&path, SC_PATH_TYPE_FILE_ID, (u8 *) "\x2F\x02", 2, 0, 0);
-	r = sc_select_file(card, &path, &file);
-	LOG_TEST_RET(card->ctx, r, "Could not select EF.C_DevAut");
-	sc_file_free(file);
+	if (priv->EF_C_DevAut && priv->EF_C_DevAut_len) {
+		ptr = priv->EF_C_DevAut;
+		len = priv->EF_C_DevAut_len;
+	} else {
+		len = sizeof efbin;
+		r = read_file(p15card, (u8 *) "\x2F\x02", efbin, &len, 1);
+		LOG_TEST_RET(card->ctx, r, "Could not select EF.C_DevAut");
 
-	r = sc_read_binary(p15card->card, 0, efbin, sizeof(efbin), 0);
-	LOG_TEST_RET(card->ctx, r, "Could not read EF.C_DevAut");
+		/* save EF_C_DevAut for further use */
+		ptr = realloc(priv->EF_C_DevAut, len);
+		if (ptr) {
+			memcpy(ptr, efbin, len);
+			priv->EF_C_DevAut = ptr;
+			priv->EF_C_DevAut_len = len;
+		}
 
-	ptr = efbin;
-	len = r;
+		ptr = efbin;
+	}
 
 	memset(&devcert, 0 ,sizeof(devcert));
 	r = sc_pkcs15emu_sc_hsm_decode_cvc(p15card, (const u8 **)&ptr, &len, &devcert);
@@ -846,7 +875,7 @@ static int sc_pkcs15emu_sc_hsm_init (sc_pkcs15_card_t * p15card)
 	if (appinfo->label == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
 
-	len = strlen(devcert.chr);		/* Strip last 5 digit sequence number from CHR */
+	len = strnlen(devcert.chr, sizeof devcert.chr);		/* Strip last 5 digit sequence number from CHR */
 	assert(len >= 8);
 	len -= 5;
 
@@ -866,9 +895,10 @@ static int sc_pkcs15emu_sc_hsm_init (sc_pkcs15_card_t * p15card)
 
 	pin_info.auth_id.len = 1;
 	pin_info.auth_id.value[0] = 1;
+	pin_info.path.aid = sc_hsm_aid;
 	pin_info.auth_type = SC_PKCS15_PIN_AUTH_TYPE_PIN;
 	pin_info.attrs.pin.reference = 0x81;
-	pin_info.attrs.pin.flags = SC_PKCS15_PIN_FLAG_LOCAL|SC_PKCS15_PIN_FLAG_INITIALIZED|SC_PKCS15_PIN_FLAG_UNBLOCK_DISABLED|SC_PKCS15_PIN_FLAG_EXCHANGE_REF_DATA;
+	pin_info.attrs.pin.flags = SC_PKCS15_PIN_FLAG_LOCAL|SC_PKCS15_PIN_FLAG_INITIALIZED|SC_PKCS15_PIN_FLAG_EXCHANGE_REF_DATA;
 	pin_info.attrs.pin.type = SC_PKCS15_PIN_TYPE_ASCII_NUMERIC;
 	pin_info.attrs.pin.min_length = 6;
 	pin_info.attrs.pin.stored_length = 0;
@@ -877,6 +907,8 @@ static int sc_pkcs15emu_sc_hsm_init (sc_pkcs15_card_t * p15card)
 	pin_info.tries_left = 3;
 	pin_info.max_tries = 3;
 
+	pin_obj.auth_id.len = 1;
+	pin_obj.auth_id.value[0] = 2;
 	strlcpy(pin_obj.label, "UserPIN", sizeof(pin_obj.label));
 	pin_obj.flags = SC_PKCS15_CO_FLAG_PRIVATE|SC_PKCS15_CO_FLAG_MODIFIABLE;
 
@@ -884,22 +916,22 @@ static int sc_pkcs15emu_sc_hsm_init (sc_pkcs15_card_t * p15card)
 	if (r < 0)
 		LOG_FUNC_RETURN(card->ctx, r);
 
-
 	memset(&pin_info, 0, sizeof(pin_info));
 	memset(&pin_obj, 0, sizeof(pin_obj));
 
 	pin_info.auth_id.len = 1;
 	pin_info.auth_id.value[0] = 2;
+	pin_info.path.aid = sc_hsm_aid;
 	pin_info.auth_type = SC_PKCS15_PIN_AUTH_TYPE_PIN;
 	pin_info.attrs.pin.reference = 0x88;
-	pin_info.attrs.pin.flags = SC_PKCS15_PIN_FLAG_LOCAL|SC_PKCS15_PIN_FLAG_CHANGE_DISABLED|SC_PKCS15_PIN_FLAG_INITIALIZED|SC_PKCS15_PIN_FLAG_UNBLOCK_DISABLED|SC_PKCS15_PIN_FLAG_SO_PIN;
+	pin_info.attrs.pin.flags = SC_PKCS15_PIN_FLAG_LOCAL|SC_PKCS15_PIN_FLAG_INITIALIZED|SC_PKCS15_PIN_FLAG_UNBLOCK_DISABLED|SC_PKCS15_PIN_FLAG_SO_PIN;
 	pin_info.attrs.pin.type = SC_PKCS15_PIN_TYPE_BCD;
 	pin_info.attrs.pin.min_length = 16;
 	pin_info.attrs.pin.stored_length = 0;
 	pin_info.attrs.pin.max_length = 16;
 	pin_info.attrs.pin.pad_char = '\0';
-	pin_info.tries_left = 3;
-	pin_info.max_tries = 3;
+	pin_info.tries_left = 15;
+	pin_info.max_tries = 15;
 
 	strlcpy(pin_obj.label, "SOPIN", sizeof(pin_obj.label));
 	pin_obj.flags = SC_PKCS15_CO_FLAG_PRIVATE;
@@ -909,6 +941,30 @@ static int sc_pkcs15emu_sc_hsm_init (sc_pkcs15_card_t * p15card)
 		LOG_FUNC_RETURN(card->ctx, r);
 
 
+	if (card->type == SC_CARD_TYPE_SC_HSM_SOC) {
+		/* SC-HSM of this type always has a PIN-Pad */
+		r = SC_SUCCESS;
+	} else {
+		memset(&pindata, 0, sizeof(pindata));
+		pindata.cmd = SC_PIN_CMD_GET_INFO;
+		pindata.pin_type = SC_AC_CHV;
+		pindata.pin_reference = 0x85;
+
+		r = sc_pin_cmd(card, &pindata, NULL);
+	}
+	if (r == SC_ERROR_DATA_OBJECT_NOT_FOUND) {
+		memset(&pindata, 0, sizeof(pindata));
+		pindata.cmd = SC_PIN_CMD_GET_INFO;
+		pindata.pin_type = SC_AC_CHV;
+		pindata.pin_reference = 0x86;
+
+		r = sc_pin_cmd(card, &pindata, NULL);
+	}
+
+	if ((r != SC_ERROR_DATA_OBJECT_NOT_FOUND) && (r != SC_ERROR_INCORRECT_PARAMETERS))
+		card->caps |= SC_CARD_CAP_PROTECTED_AUTHENTICATION_PATH;
+
+
 	filelistlength = sc_list_files(card, filelist, sizeof(filelist));
 	LOG_TEST_RET(card->ctx, filelistlength, "Could not enumerate file and key identifier");
 
@@ -935,12 +991,14 @@ static int sc_pkcs15emu_sc_hsm_init (sc_pkcs15_card_t * p15card)
 
 
 int sc_pkcs15emu_sc_hsm_init_ex(sc_pkcs15_card_t *p15card,
+				struct sc_aid *aid,
 				sc_pkcs15emu_opt_t *opts)
 {
 	if (opts && (opts->flags & SC_PKCS15EMU_FLAGS_NO_CHECK)) {
 		return sc_pkcs15emu_sc_hsm_init(p15card);
 	} else {
-		if (p15card->card->type != SC_CARD_TYPE_SC_HSM) {
+		if (p15card->card->type != SC_CARD_TYPE_SC_HSM
+				&& p15card->card->type != SC_CARD_TYPE_SC_HSM_SOC) {
 			return SC_ERROR_WRONG_CARD;
 		}
 		return sc_pkcs15emu_sc_hsm_init(p15card);
