@@ -35,6 +35,7 @@
 
 int sc_pkcs15emu_tcos_init_ex(
 	sc_pkcs15_card_t   *p15card,
+	struct sc_aid *,
 	sc_pkcs15emu_opt_t *opts);
 
 static int insert_cert(
@@ -207,6 +208,7 @@ static int insert_pin(
 	pin_info.attrs.pin.stored_length    = 16;
 	pin_info.attrs.pin.max_length       = 16;
 	pin_info.attrs.pin.pad_char         = '\0';
+	pin_info.logged_in = SC_PIN_STATE_UNKNOWN;
 	sc_format_path(path, &pin_info.path);
 
 	memset(&pin_obj, 0, sizeof(pin_obj));
@@ -488,6 +490,7 @@ static int detect_unicard(
 
 int sc_pkcs15emu_tcos_init_ex(
 	sc_pkcs15_card_t   *p15card,
+	struct sc_aid *aid,
 	sc_pkcs15emu_opt_t *opts
 ){
 	sc_card_t         *card = p15card->card;
