@@ -88,7 +88,7 @@ static const struct sc_asn1_entry c_asn1_profile_indication[C_ASN1_PROFILE_INDIC
 
 #define C_ASN1_TOKI_ATTRS_SIZE 15
 static const struct sc_asn1_entry c_asn1_toki_attrs[C_ASN1_TOKI_ATTRS_SIZE] = {
-	{ "version",        SC_ASN1_INTEGER,		SC_ASN1_TAG_INTEGER, 0, NULL, NULL },
+	{ "version",	    SC_ASN1_INTEGER,		SC_ASN1_TAG_INTEGER, 0, NULL, NULL },
 	{ "serialNumber",   SC_ASN1_OCTET_STRING,	SC_ASN1_TAG_OCTET_STRING, SC_ASN1_OPTIONAL, NULL, NULL },
 	{ "manufacturerID", SC_ASN1_UTF8STRING,		SC_ASN1_TAG_UTF8STRING, SC_ASN1_OPTIONAL, NULL, NULL },
 	{ "label",	    SC_ASN1_UTF8STRING,		SC_ASN1_CTX | 0, SC_ASN1_OPTIONAL, NULL, NULL },
@@ -111,9 +111,11 @@ static const struct sc_asn1_entry c_asn1_tokeninfo[] = {
 	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
-static void sc_pkcs15_free_unusedspace(struct sc_pkcs15_card *p15card);
-static void sc_pkcs15_remove_dfs(struct sc_pkcs15_card *p15card);
-static void sc_pkcs15_remove_objects(struct sc_pkcs15_card *p15card);
+static void sc_pkcs15_free_unusedspace(struct sc_pkcs15_card *);
+static void sc_pkcs15_remove_dfs(struct sc_pkcs15_card *);
+static void sc_pkcs15_remove_objects(struct sc_pkcs15_card *);
+static int sc_pkcs15_aux_get_md_guid(struct sc_pkcs15_card *, const struct sc_pkcs15_object *,
+		unsigned, unsigned char *, size_t *);
 
 int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
 	sc_pkcs15_tokeninfo_t *ti, const u8 *buf, size_t blen)
@@ -132,7 +134,7 @@ int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
 	u8 preferred_language[3];
 	size_t lang_length = sizeof(preferred_language);
 	struct sc_asn1_entry asn1_supported_algorithms[SC_MAX_SUPPORTED_ALGORITHMS + 1],
-			     asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7];
+			asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7];
 	size_t reference_len = sizeof(ti->supported_algos[0].reference);
 	size_t mechanism_len = sizeof(ti->supported_algos[0].mechanism);
 	size_t operations_len = sizeof(ti->supported_algos[0].operations);
@@ -201,6 +203,7 @@ int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
 			sprintf(byte, "%02X", serial[ii]);
 			strcat(ti->serial_number, byte);
 		}
+		sc_log(ctx, "TokenInfo.serialNunmber '%s'", ti->serial_number);
 	}
 
 	if (ti->manufacturer_id == NULL) {
@@ -267,7 +270,7 @@ sc_pkcs15_encode_tokeninfo(sc_context_t *ctx, sc_pkcs15_tokeninfo_t *ti,
 	struct sc_asn1_entry asn1_toki_attrs[C_ASN1_TOKI_ATTRS_SIZE];
 	struct sc_asn1_entry asn1_tokeninfo[2];
 	struct sc_asn1_entry asn1_supported_algorithms[SC_MAX_SUPPORTED_ALGORITHMS + 1],
-			     asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7];
+			asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7];
 	size_t reference_len = sizeof(ti->supported_algos[0].reference);
 	size_t mechanism_len = sizeof(ti->supported_algos[0].mechanism);
 	size_t operations_len = sizeof(ti->supported_algos[0].operations);
@@ -403,14 +406,10 @@ fix_authentic_ddo(struct sc_pkcs15_card *p15card)
 	 * Cleanup this attributes -- default values must be OK.
 	 */
 	if (p15card->card->type == SC_CARD_TYPE_OBERTHUR_AUTHENTIC_3_2)   {
-		if (p15card->file_odf != NULL) {
-			sc_file_free(p15card->file_odf);
-			p15card->file_odf = NULL;
-		}
-	        if (p15card->file_tokeninfo != NULL) {
-			sc_file_free(p15card->file_tokeninfo);
-			p15card->file_tokeninfo = NULL;
-		}
+		sc_file_free(p15card->file_odf);
+		p15card->file_odf = NULL;
+		sc_file_free(p15card->file_tokeninfo);
+		p15card->file_tokeninfo = NULL;
 	}
 }
 
@@ -421,11 +420,12 @@ fix_starcos_pkcs15_card(struct sc_pkcs15_card *p15card)
 	struct sc_context *ctx = p15card->card->ctx;
 
 	/* set special flags based on card meta data */
-	if (strcmp(p15card->card->driver->short_name,"cardos") == 0) {
+	if (strcmp(p15card->card->driver->short_name,"cardos") == 0
+			&& p15card->tokeninfo && p15card->tokeninfo->label) {
 
 		/* D-Trust cards (D-TRUST, D-SIGN) */
 		if (strstr(p15card->tokeninfo->label,"D-TRUST") != NULL
-			|| strstr(p15card->tokeninfo->label,"D-SIGN") != NULL) {
+				|| strstr(p15card->tokeninfo->label,"D-SIGN") != NULL) {
 
 			/* D-TRUST Card 2.0 2cc (standard cards, which always add
 			 * SHA1 prefix itself */
@@ -514,18 +514,12 @@ parse_ddo(struct sc_pkcs15_card *p15card, const u8 * buf, size_t buflen)
 	fix_authentic_ddo(p15card);
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 mem_err:
-	if (p15card->file_odf != NULL) {
-		sc_file_free(p15card->file_odf);
-		p15card->file_odf = NULL;
-	}
-	if (p15card->file_tokeninfo != NULL) {
-		sc_file_free(p15card->file_tokeninfo);
-		p15card->file_tokeninfo = NULL;
-	}
-	if (p15card->file_unusedspace != NULL) {
-		sc_file_free(p15card->file_unusedspace);
-		p15card->file_unusedspace = NULL;
-	}
+	sc_file_free(p15card->file_odf);
+	p15card->file_odf = NULL;
+	sc_file_free(p15card->file_tokeninfo);
+	p15card->file_tokeninfo = NULL;
+	sc_file_free(p15card->file_unusedspace);
+	p15card->file_unusedspace = NULL;
 	LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 }
 
@@ -536,8 +530,8 @@ sc_pkcs15_get_lastupdate(struct sc_pkcs15_card *p15card)
 	struct sc_context *ctx  = p15card->card->ctx;
 	struct sc_file *file = NULL;
 	struct sc_asn1_entry asn1_last_update[C_ASN1_LAST_UPDATE_SIZE];
-        unsigned char *content, last_update[32];
-        size_t lupdate_len = sizeof(last_update) - 1;
+	unsigned char *content, last_update[32];
+	size_t lupdate_len = sizeof(last_update) - 1;
 	int r, content_len;
 	size_t size;
 
@@ -547,15 +541,12 @@ sc_pkcs15_get_lastupdate(struct sc_pkcs15_card *p15card)
 	if (!p15card->tokeninfo->last_update.path.len)
 		return NULL;
 
-        r = sc_select_file(p15card->card, &p15card->tokeninfo->last_update.path, &file);
+	r = sc_select_file(p15card->card, &p15card->tokeninfo->last_update.path, &file);
 	if (r < 0)
 		return NULL;
 
-	if (file->size) {
-		size = 1024;
-	} else {
-		size = file->size;
-	}
+	size = file->size ? file->size : 1024;
+
 	content = calloc(size, 1);
 	if (!content)
 		return NULL;
@@ -776,9 +767,8 @@ sc_pkcs15_free_app(struct sc_pkcs15_card *p15card)
 void
 sc_pkcs15_card_free(struct sc_pkcs15_card *p15card)
 {
-	if (p15card == NULL)
+	if (p15card == NULL || p15card->magic != SC_PKCS15_CARD_MAGIC)
 		return;
-	assert(p15card->magic == SC_PKCS15_CARD_MAGIC);
 
 	if (p15card->ops.clear)
 		p15card->ops.clear(p15card);
@@ -793,14 +783,10 @@ sc_pkcs15_card_free(struct sc_pkcs15_card *p15card)
 	sc_pkcs15_free_unusedspace(p15card);
 	p15card->unusedspace_read = 0;
 
-	if (p15card->file_app != NULL)
-		sc_file_free(p15card->file_app);
-	if (p15card->file_tokeninfo != NULL)
-		sc_file_free(p15card->file_tokeninfo);
-	if (p15card->file_odf != NULL)
-		sc_file_free(p15card->file_odf);
-	if (p15card->file_unusedspace != NULL)
-		sc_file_free(p15card->file_unusedspace);
+	sc_file_free(p15card->file_app);
+	sc_file_free(p15card->file_tokeninfo);
+	sc_file_free(p15card->file_odf);
+	sc_file_free(p15card->file_unusedspace);
 
 	p15card->magic = 0;
 	sc_pkcs15_free_tokeninfo(p15card);
@@ -826,22 +812,14 @@ sc_pkcs15_card_clear(struct sc_pkcs15_card *p15card)
 	sc_pkcs15_remove_dfs(p15card);
 
 	p15card->df_list = NULL;
-	if (p15card->file_app != NULL) {
-		sc_file_free(p15card->file_app);
-		p15card->file_app = NULL;
-	}
-	if (p15card->file_tokeninfo != NULL) {
-		sc_file_free(p15card->file_tokeninfo);
-		p15card->file_tokeninfo = NULL;
-	}
-	if (p15card->file_odf != NULL) {
-		sc_file_free(p15card->file_odf);
-		p15card->file_odf = NULL;
-	}
-	if (p15card->file_unusedspace != NULL) {
-		sc_file_free(p15card->file_unusedspace);
-		p15card->file_unusedspace = NULL;
-	}
+	sc_file_free(p15card->file_app);
+	p15card->file_app = NULL;
+	sc_file_free(p15card->file_tokeninfo);
+	p15card->file_tokeninfo = NULL;
+	sc_file_free(p15card->file_odf);
+	p15card->file_odf = NULL;
+	sc_file_free(p15card->file_unusedspace);
+	p15card->file_unusedspace = NULL;
 	if (p15card->tokeninfo->label != NULL) {
 		free(p15card->tokeninfo->label);
 		p15card->tokeninfo->label = NULL;
@@ -1073,17 +1051,28 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 	if(buf == NULL)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 
-	err = sc_read_binary(card, 0, buf, len, 0);
-	if (err < 0)   {
-		sc_log(ctx, "read EF(ODF) file error: %s", sc_strerror(err));
-		goto end;
+	err = -1; /* file state: not in cache */
+	if (p15card->opts.use_file_cache) {
+		err = sc_pkcs15_read_cached_file(p15card, &tmppath, &buf, &len);
+		if (err == SC_SUCCESS)
+			err = len;
 	}
-	else if (err < 2) {
-		err = SC_ERROR_PKCS15_APP_NOT_FOUND;
-		sc_log(ctx, "Invalid content of EF(ODF): %s", sc_strerror(err));
-		goto end;
+	if (err < 0) {
+		err = sc_read_binary(card, 0, buf, len, 0);
+		if (err < 0)   {
+			sc_log(ctx, "read EF(ODF) file error: %s", sc_strerror(err));
+			goto end;
+		}
+		else if (err < 2) {
+			err = SC_ERROR_PKCS15_APP_NOT_FOUND;
+			sc_log(ctx, "Invalid content of EF(ODF): %s", sc_strerror(err));
+			goto end;
+		}
+
+		if (p15card->opts.use_file_cache) {
+			sc_pkcs15_cache_file(p15card, &tmppath, buf, len);
+		}
 	}
-	len = err;
 
 	if (parse_odf(buf, len, p15card)) {
 		err = SC_ERROR_PKCS15_APP_NOT_FOUND;
@@ -1128,15 +1117,28 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 	if(buf == NULL)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 
-	err = sc_read_binary(card, 0, buf, len, 0);
-	if (err < 0)   {
-		sc_log(ctx, "read EF(TokenInfo) file error: %s", sc_strerror(err));
-		goto end;
+	err = -1; /* file state: not in cache */
+	if (p15card->opts.use_file_cache) {
+		err = sc_pkcs15_read_cached_file(p15card, &tmppath, &buf, &len);
+		if (err == SC_SUCCESS)
+			err = len;
 	}
-	if (err <= 2) {
-		err = SC_ERROR_PKCS15_APP_NOT_FOUND;
-		sc_log(ctx, "Invalid content of EF(TokenInfo): %s", sc_strerror(err));
-		goto end;
+	if (err < 0) {
+		err = sc_read_binary(card, 0, buf, len, 0);
+		if (err < 0)   {
+			sc_log(ctx, "read EF(TokenInfo) file error: %s", sc_strerror(err));
+			goto end;
+		}
+		if (err <= 2) {
+			err = SC_ERROR_PKCS15_APP_NOT_FOUND;
+			sc_log(ctx, "Invalid content of EF(TokenInfo): %s", sc_strerror(err));
+			goto end;
+		}
+
+		if (p15card->opts.use_file_cache) {
+			sc_pkcs15_cache_file(p15card, &tmppath, buf, len);
+		}
+		err = len;
 	}
 
 	memset(&tokeninfo, 0, sizeof(tokeninfo));
@@ -1151,6 +1153,10 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 	if (!p15card->tokeninfo->serial_number && card->serialnr.len)   {
 		char *serial = calloc(1, card->serialnr.len*2 + 1);
 		size_t ii;
+		if (!serial) {
+			err = SC_ERROR_OUT_OF_MEMORY;
+			goto end;
+		}
 
 		for(ii=0;ii<card->serialnr.len;ii++)
 			sprintf(serial + ii*2, "%02X", *(card->serialnr.value + ii));
@@ -1186,7 +1192,9 @@ sc_pkcs15_bind(struct sc_card *card, struct sc_aid *aid,
 	LOG_FUNC_CALLED(ctx);
 	sc_log(ctx, "application(aid:'%s')", aid ? sc_dump_hex(aid->value, aid->len) : "empty");
 
-	assert(p15card_out != NULL);
+	if (p15card_out == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	p15card = sc_pkcs15_card_new();
 	if (p15card == NULL)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
@@ -1207,8 +1215,8 @@ sc_pkcs15_bind(struct sc_card *card, struct sc_aid *aid,
 				p15card->opts.pin_cache_ignore_user_consent);
 	}
 	sc_log(ctx, "PKCS#15 options: use_file_cache=%d use_pin_cache=%d pin_cache_counter=%d pin_cache_ignore_user_consent=%d",
-	         p15card->opts.use_file_cache, p15card->opts.use_pin_cache,
-		 p15card->opts.pin_cache_counter, p15card->opts.pin_cache_ignore_user_consent);
+			p15card->opts.use_file_cache, p15card->opts.use_pin_cache,p15card->opts.pin_cache_counter,
+			p15card->opts.pin_cache_ignore_user_consent);
 
 	r = sc_lock(card);
 	if (r) {
@@ -1222,7 +1230,7 @@ sc_pkcs15_bind(struct sc_card *card, struct sc_aid *aid,
 		sc_log(ctx, "PKCS#15 emulation enabled");
 		emu_first = scconf_get_bool(conf_block, "try_emulation_first", 0);
 		if (emu_first || sc_pkcs15_is_emulation_only(card)) {
-			r = sc_pkcs15_bind_synthetic(p15card);
+			r = sc_pkcs15_bind_synthetic(p15card, aid);
 			if (r == SC_SUCCESS)
 				goto done;
 			r = sc_pkcs15_bind_internal(p15card, aid);
@@ -1232,7 +1240,7 @@ sc_pkcs15_bind(struct sc_card *card, struct sc_aid *aid,
 			r = sc_pkcs15_bind_internal(p15card, aid);
 			if (r == SC_SUCCESS)
 				goto done;
-			r = sc_pkcs15_bind_synthetic(p15card);
+			r = sc_pkcs15_bind_synthetic(p15card, aid);
 			if (r < 0)
 				goto error;
 		}
@@ -1258,7 +1266,9 @@ error:
 int
 sc_pkcs15_unbind(struct sc_pkcs15_card *p15card)
 {
-	assert(p15card != NULL && p15card->magic == SC_PKCS15_CARD_MAGIC);
+	if (p15card == NULL || p15card->magic != SC_PKCS15_CARD_MAGIC) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	LOG_FUNC_CALLED(p15card->card->ctx);
 	if (p15card->dll_handle)
@@ -1278,6 +1288,7 @@ __sc_pkcs15_search_objects(struct sc_pkcs15_card *p15card, unsigned int class_ma
 	struct sc_pkcs15_df	*df = NULL;
 	unsigned int	df_mask = 0;
 	size_t		match_count = 0;
+	int r;
 
 	if (type)
 		class_mask |= SC_PKCS15_TYPE_TO_CLASS(type);
@@ -1314,9 +1325,12 @@ __sc_pkcs15_search_objects(struct sc_pkcs15_card *p15card, unsigned int class_ma
 		}
 		if (df->enumerated)
 			continue;
-		/* Enumerate the DF's, so p15card->obj_list is
-		 * populated. */
-		if (SC_SUCCESS != sc_pkcs15_parse_df(p15card, df))
+		/* Enumerate the DF's, so p15card->obj_list is populated. */
+		if (p15card->ops.parse_df)
+			r = p15card->ops.parse_df(p15card, df);
+		else
+			r = sc_pkcs15_parse_df(p15card, df);
+		if (r != SC_SUCCESS)
 			continue;
 	}
 
@@ -1359,26 +1373,16 @@ compare_obj_id(struct sc_pkcs15_object *obj, const struct sc_pkcs15_id *id)
 {
 	void *data = obj->data;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_CERT_X509:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_CERT:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_cert_info *) data)->id, id);
-	case SC_PKCS15_TYPE_PRKEY_RSA:
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PRKEY_EC:
+	case SC_PKCS15_TYPE_PRKEY:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_prkey_info *) data)->id, id);
-	case SC_PKCS15_TYPE_PUBKEY_RSA:
-	case SC_PKCS15_TYPE_PUBKEY_DSA:
-	case SC_PKCS15_TYPE_PUBKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PUBKEY_EC:
+	case SC_PKCS15_TYPE_PUBKEY:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_pubkey_info *) data)->id, id);
-	case SC_PKCS15_TYPE_SKEY_DES:
-	case SC_PKCS15_TYPE_SKEY_2DES:
-	case SC_PKCS15_TYPE_SKEY_3DES:
+	case SC_PKCS15_TYPE_SKEY:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_skey_info *) data)->id, id);
-	case SC_PKCS15_TYPE_AUTH_PIN:
-	case SC_PKCS15_TYPE_AUTH_BIO:
-	case SC_PKCS15_TYPE_AUTH_AUTHKEY:
+	case SC_PKCS15_TYPE_AUTH:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_auth_info *) data)->auth_id, id);
 	case SC_PKCS15_TYPE_DATA_OBJECT:
 		return sc_pkcs15_compare_id(&((struct sc_pkcs15_data_info *) data)->id, id);
@@ -1390,7 +1394,7 @@ compare_obj_id(struct sc_pkcs15_object *obj, const struct sc_pkcs15_id *id)
 static int
 sc_obj_app_oid(struct sc_pkcs15_object *obj, const struct sc_object_id *app_oid)
 {
-	if (obj->type & SC_PKCS15_TYPE_DATA_OBJECT)
+	if ((obj->type & SC_PKCS15_TYPE_CLASS_MASK) == SC_PKCS15_TYPE_DATA_OBJECT)
 		return sc_compare_oid(&((struct sc_pkcs15_data_info *) obj->data)->app_oid, app_oid);
 	return 0;
 }
@@ -1402,17 +1406,11 @@ compare_obj_usage(struct sc_pkcs15_object *obj, unsigned int mask, unsigned int
 	void		*data = obj->data;
 	unsigned int	usage;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_PRKEY_RSA:
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PRKEY_EC:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_PRKEY:
 		usage = ((struct sc_pkcs15_prkey_info *) data)->usage;
 		break;
-	case SC_PKCS15_TYPE_PUBKEY_RSA:
-	case SC_PKCS15_TYPE_PUBKEY_DSA:
-	case SC_PKCS15_TYPE_PUBKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PUBKEY_EC:
+	case SC_PKCS15_TYPE_PUBKEY:
 		usage = ((struct sc_pkcs15_pubkey_info *) data)->usage;
 		break;
 	default:
@@ -1428,8 +1426,8 @@ compare_obj_flags(struct sc_pkcs15_object *obj, unsigned int mask, unsigned int
 	struct sc_pkcs15_auth_info *auth_info;
 	unsigned int	flags;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_AUTH_PIN:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_AUTH:
 		auth_info = (struct sc_pkcs15_auth_info *) obj->data;
 		if (auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
 			return 0;
@@ -1449,17 +1447,14 @@ compare_obj_reference(struct sc_pkcs15_object *obj, int value)
 	void		*data = obj->data;
 	int		reference;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_AUTH_PIN:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_AUTH:
 		auth_info = (struct sc_pkcs15_auth_info *) obj->data;
 		if (auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
 			return 0;
 		reference = auth_info->attrs.pin.reference;
 		break;
-	case SC_PKCS15_TYPE_PRKEY_RSA:
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PRKEY_EC:
+	case SC_PKCS15_TYPE_PRKEY:
 		reference = ((struct sc_pkcs15_prkey_info *) data)->key_reference;
 		break;
 	default:
@@ -1474,20 +1469,16 @@ compare_obj_path(struct sc_pkcs15_object *obj, const struct sc_path *path)
 {
 	void *data = obj->data;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_CERT_X509:
-		return sc_compare_path(&((struct sc_pkcs15_cert_info *) data)->path, path);
-	case SC_PKCS15_TYPE_PRKEY_RSA:
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PRKEY_EC:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_PRKEY:
 		return sc_compare_path(&((struct sc_pkcs15_prkey_info *) data)->path, path);
-	case SC_PKCS15_TYPE_PUBKEY_RSA:
-	case SC_PKCS15_TYPE_PUBKEY_DSA:
-	case SC_PKCS15_TYPE_PUBKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PUBKEY_EC:
+	case SC_PKCS15_TYPE_PUBKEY:
 		return sc_compare_path(&((struct sc_pkcs15_pubkey_info *) data)->path, path);
-	case SC_PKCS15_TYPE_AUTH_PIN:
+	case SC_PKCS15_TYPE_SKEY:
+		return sc_compare_path(&((struct sc_pkcs15_skey_info *) data)->path, path);
+	case SC_PKCS15_TYPE_CERT:
+		return sc_compare_path(&((struct sc_pkcs15_cert_info *) data)->path, path);
+	case SC_PKCS15_TYPE_AUTH:
 		return sc_compare_path(&((struct sc_pkcs15_auth_info *) data)->path, path);
 	case SC_PKCS15_TYPE_DATA_OBJECT:
 		return sc_compare_path(&((struct sc_pkcs15_data_info *) data)->path, path);
@@ -1504,8 +1495,8 @@ compare_obj_data_name(struct sc_pkcs15_object *obj, const char *app_label, const
 	if (obj->type != SC_PKCS15_TYPE_DATA_OBJECT)
 		return 0;
 
-	return !strcmp(cinfo->app_label, app_label) &&
-		!strcmp(obj->label, label);
+	return !strncmp(cinfo->app_label, app_label, sizeof cinfo->app_label) &&
+		!strncmp(obj->label, label, sizeof obj->label);
 }
 
 
@@ -1565,9 +1556,8 @@ sc_pkcs15_search_objects(struct sc_pkcs15_card *p15card, struct sc_pkcs15_search
 
 int
 sc_pkcs15_get_objects_cond(struct sc_pkcs15_card *p15card, unsigned int type,
-			       int (* func)(struct sc_pkcs15_object *, void *),
-			       void *func_arg,
-			       struct sc_pkcs15_object **ret, size_t ret_size)
+		int (* func)(struct sc_pkcs15_object *, void *),
+		void *func_arg, struct sc_pkcs15_object **ret, size_t ret_size)
 {
 	return __sc_pkcs15_search_objects(p15card, 0, type,
 			func, func_arg, ret, ret_size);
@@ -1790,8 +1780,7 @@ sc_pkcs15_find_data_object_by_name(struct sc_pkcs15_card *p15card, const char *a
 
 int
 sc_pkcs15_find_prkey_by_id_usage(struct sc_pkcs15_card *p15card, const struct sc_pkcs15_id *id,
-			       unsigned int usage,
-			       struct sc_pkcs15_object **out)
+		unsigned int usage, struct sc_pkcs15_object **out)
 {
 	struct sc_pkcs15_search_key sk;
 
@@ -1955,7 +1944,9 @@ sc_pkcs15_encode_df(struct sc_context *ctx, struct sc_pkcs15_card *p15card, stru
 		     unsigned char **nbuf, size_t *nbufsize) = NULL;
 	int r;
 
-	assert(p15card != NULL && p15card->magic == SC_PKCS15_CARD_MAGIC);
+	if (p15card == NULL || p15card->magic != SC_PKCS15_CARD_MAGIC) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	switch (df->type) {
 	case SC_PKCS15_PRKDF:
 		func = sc_pkcs15_encode_prkdf_entry;
@@ -1964,6 +1955,9 @@ sc_pkcs15_encode_df(struct sc_context *ctx, struct sc_pkcs15_card *p15card, stru
 	case SC_PKCS15_PUKDF_TRUSTED:
 		func = sc_pkcs15_encode_pukdf_entry;
 		break;
+	case SC_PKCS15_SKDF:
+		func = sc_pkcs15_encode_skdf_entry;
+		break;
 	case SC_PKCS15_CDF:
 	case SC_PKCS15_CDF_TRUSTED:
 	case SC_PKCS15_CDF_USEFUL:
@@ -2023,11 +2017,6 @@ sc_pkcs15_parse_df(struct sc_pkcs15_card *p15card, struct sc_pkcs15_df *df)
 
 	sc_log(ctx, "called; path=%s, type=%d, enum=%d", sc_print_path(&df->path), df->type, df->enumerated);
 
-	if (p15card->ops.parse_df)   {
-		r = p15card->ops.parse_df(p15card, df);
-		LOG_FUNC_RETURN(ctx, r);
-	}
-
 	if (df->enumerated)
 		LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 
@@ -2223,7 +2212,8 @@ sc_pkcs15_encode_unusedspace(struct sc_context *ctx, struct sc_pkcs15_card *p15c
 		sc_format_asn1_entry(asn1_unusedspace + c, asn1_values + 3*c, NULL, 1);
 		sc_copy_asn1_entry(c_asn1_unusedspace_values, asn1_values + 3*c);
 		sc_format_asn1_entry(asn1_values + 3*c, &unusedspace->path, NULL, 1);
-		sc_format_asn1_entry(asn1_values + 3*c+1, &unusedspace->auth_id, NULL, unusedspace->auth_id.len);
+		sc_format_asn1_entry(asn1_values + 3*c+1, &unusedspace->auth_id, NULL,
+			   unusedspace->auth_id.len > 0 ? 1 : 0);
 		c++;
 	}
 	asn1_unusedspace[c].name = NULL;
@@ -2302,13 +2292,16 @@ int
 sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_path,
 		unsigned char **buf, size_t *buflen)
 {
-	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_context *ctx;
 	struct sc_file *file = NULL;
 	unsigned char *data = NULL;
 	size_t	len = 0, offset = 0;
 	int	r;
 
-	assert(p15card != NULL && in_path != NULL && buf != NULL);
+	if (p15card == NULL || p15card->card == NULL || in_path == NULL || buf == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	ctx = p15card->card->ctx;
 
 	LOG_FUNC_CALLED(ctx);
 	sc_log(ctx, "path=%s, index=%u, count=%d", sc_print_path(in_path), in_path->index, in_path->count);
@@ -2316,7 +2309,16 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 	r = -1; /* file state: not in cache */
 	if (p15card->opts.use_file_cache) {
 		r = sc_pkcs15_read_cached_file(p15card, in_path, &data, &len);
+
+		if (!r && in_path->aid.len > 0 && in_path->len >= 2)   {
+			struct sc_path parent = *in_path;
+
+			parent.len -= 2;
+			parent.type = SC_PATH_TYPE_PATH;
+			r = sc_select_file(p15card->card, &parent, NULL);
+		}
 	}
+
 	if (r) {
 		r = sc_lock(p15card->card);
 		LOG_TEST_RET(ctx, r, "sc_lock() failed");
@@ -2360,7 +2362,6 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 				if (r == SC_ERROR_RECORD_NOT_FOUND)
 					break;
 				if (r < 0) {
-					free(data);
 					goto fail_unlock;
 				}
 				if (r < 2)
@@ -2382,7 +2383,6 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 		else {
 			r = sc_read_binary(p15card->card, offset, data, len, 0);
 			if (r < 0) {
-				free(data);
 				goto fail_unlock;
 			}
 			/* sc_read_binary may return less than requested */
@@ -2391,14 +2391,18 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 		sc_unlock(p15card->card);
 
 		sc_file_free(file);
+
+		if (p15card->opts.use_file_cache) {
+			sc_pkcs15_cache_file(p15card, in_path, data, len);
+		}
 	}
 	*buf = data;
 	*buflen = len;
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 
 fail_unlock:
-	if (file)
-		sc_file_free(file);
+	free(data);
+	sc_file_free(file);
 	sc_unlock(p15card->card);
 	LOG_FUNC_RETURN(ctx, r);
 }
@@ -2407,7 +2411,8 @@ fail_unlock:
 int
 sc_pkcs15_compare_id(const struct sc_pkcs15_id *id1, const struct sc_pkcs15_id *id2)
 {
-	assert(id1 != NULL && id2 != NULL);
+	if (id1 == NULL || id2 == NULL)
+		return 0;
 	if (id1->len != id2->len)
 		return 0;
 	return memcmp(id1->value, id2->value, id1->len) == 0;
@@ -2417,9 +2422,15 @@ sc_pkcs15_compare_id(const struct sc_pkcs15_id *id1, const struct sc_pkcs15_id *
 void
 sc_pkcs15_format_id(const char *str, struct sc_pkcs15_id *id)
 {
-	size_t len = sizeof(id->value);
+	size_t len;
 
-	if (sc_hex_to_bin(str, id->value, &len) >= 0)
+	if (!id)
+		return;
+	len = sizeof(id->value);
+
+	if (sc_hex_to_bin(str, id->value, &len) != SC_SUCCESS)
+		id->len = 0;
+	else
 		id->len = len;
 }
 
@@ -2595,13 +2606,16 @@ sc_pkcs15_add_supported_algo_ref(struct sc_pkcs15_object *obj, struct sc_support
 	if (!algo)
 		return SC_SUCCESS;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_PRKEY_RSA:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_PRKEY:
 		algo_refs = ((struct sc_pkcs15_prkey_info *)obj->data)->algo_refs;
 		break;
-	case SC_PKCS15_TYPE_PUBKEY_RSA:
+	case SC_PKCS15_TYPE_PUBKEY:
 		algo_refs = ((struct sc_pkcs15_pubkey_info *)obj->data)->algo_refs;
 		break;
+	case SC_PKCS15_TYPE_SKEY:
+		algo_refs = ((struct sc_pkcs15_skey_info *)obj->data)->algo_refs;
+		break;
 	}
 	if (!algo_refs)
 		return SC_ERROR_NOT_SUPPORTED;
@@ -2627,23 +2641,20 @@ sc_pkcs15_get_object_id(const struct sc_pkcs15_object *obj, struct sc_pkcs15_id
 	if (!obj || !out)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
-	switch (obj->type) {
-	case SC_PKCS15_TYPE_CERT_X509:
+	switch (obj->type & SC_PKCS15_TYPE_CLASS_MASK) {
+	case SC_PKCS15_TYPE_CERT:
 		*out = ((struct sc_pkcs15_cert_info *) obj->data)->id;
 		break;
-	case SC_PKCS15_TYPE_PRKEY_RSA:
-	case SC_PKCS15_TYPE_PRKEY_DSA:
-	case SC_PKCS15_TYPE_PRKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PRKEY_EC:
+	case SC_PKCS15_TYPE_PRKEY:
 		*out = ((struct sc_pkcs15_prkey_info *) obj->data)->id;
 		break;
-	case SC_PKCS15_TYPE_PUBKEY_RSA:
-	case SC_PKCS15_TYPE_PUBKEY_DSA:
-	case SC_PKCS15_TYPE_PUBKEY_GOSTR3410:
-	case SC_PKCS15_TYPE_PUBKEY_EC:
+	case SC_PKCS15_TYPE_PUBKEY:
 		*out = ((struct sc_pkcs15_pubkey_info *) obj->data)->id;
 		break;
-	case SC_PKCS15_TYPE_AUTH_PIN:
+	case SC_PKCS15_TYPE_SKEY:
+		*out = ((struct sc_pkcs15_skey_info *) obj->data)->id;
+		break;
+	case SC_PKCS15_TYPE_AUTH:
 		*out = ((struct sc_pkcs15_auth_info *) obj->data)->auth_id;
 		break;
 	case SC_PKCS15_TYPE_DATA_OBJECT:
@@ -2696,7 +2707,7 @@ sc_pkcs15_serialize_guid(unsigned char *in, size_t in_size, unsigned flags,
 
 int
 sc_pkcs15_get_object_guid(struct sc_pkcs15_card *p15card, const struct sc_pkcs15_object *obj,
-		                unsigned flags, unsigned char *out, size_t *out_size)
+		unsigned flags, unsigned char *out, size_t *out_size)
 {
 	struct sc_context *ctx = p15card->card->ctx;
 	struct sc_serial_number serialnr;
@@ -2713,19 +2724,13 @@ sc_pkcs15_get_object_guid(struct sc_pkcs15_card *p15card, const struct sc_pkcs15
 		LOG_FUNC_RETURN(ctx, rv);
 	}
 
+	rv = sc_pkcs15_aux_get_md_guid(p15card, obj, flags, out, out_size);
+	if (rv == SC_SUCCESS)
+		LOG_FUNC_RETURN(ctx, SC_SUCCESS);
+	else if (rv != SC_ERROR_NOT_SUPPORTED)
+		LOG_TEST_RET(ctx, rv, "Failed to get alternative object GUID");
+
 	memset(out, 0, *out_size);
-	if ((obj->type & SC_PKCS15_TYPE_CLASS_MASK) == SC_PKCS15_TYPE_PRKEY)   {
-		struct sc_pkcs15_prkey_info *info = (struct sc_pkcs15_prkey_info *)obj->data;
-
-		if (info->cmap_record.guid && info->cmap_record.guid_len)   {
-			if (*out_size < info->cmap_record.guid_len)
-				LOG_FUNC_RETURN(ctx, SC_ERROR_BUFFER_TOO_SMALL);
-
-			memcpy(out, info->cmap_record.guid, info->cmap_record.guid_len);
-			*out_size = info->cmap_record.guid_len;
-			LOG_FUNC_RETURN(ctx, SC_SUCCESS);
-		}
-	}
 
 	rv = sc_pkcs15_get_object_id(obj, &id);
 	LOG_TEST_RET(ctx, rv, "Cannot get object's ID");
@@ -2755,9 +2760,9 @@ sc_pkcs15_get_object_guid(struct sc_pkcs15_card *p15card, const struct sc_pkcs15
 	memcpy(guid_bin + id.len, serialnr.value, serialnr.len);
 	guid_bin_size = id.len + serialnr.len;
 
-        /*
+	/*
 	 * If OpenSSL is available (SHA1), then rather use the hash of the data
-         * - this also protects against data being too short
+	 * - this also protects against data being too short
 	 */
 #ifdef ENABLE_OPENSSL
 	SHA1(guid_bin, guid_bin_size, guid_bin);
@@ -2779,6 +2784,31 @@ sc_pkcs15_get_object_guid(struct sc_pkcs15_card *p15card, const struct sc_pkcs15
 }
 
 
+static int
+sc_pkcs15_aux_get_md_guid(struct sc_pkcs15_card *p15card, const struct sc_pkcs15_object *obj,
+		unsigned flags,
+		unsigned char *out, size_t *out_size)
+{
+	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_pkcs15_prkey_info *prkey_info = NULL;
+	int rv;
+
+	LOG_FUNC_CALLED(ctx);
+	if(!out || !out_size)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
+
+	if ((obj->type & SC_PKCS15_TYPE_CLASS_MASK) != SC_PKCS15_TYPE_PRKEY)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_SUPPORTED);
+
+	prkey_info = (struct sc_pkcs15_prkey_info *)obj->data;
+	if (!prkey_info->aux_data || prkey_info->aux_data->type != SC_AUX_DATA_TYPE_MD_CMAP_RECORD)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_SUPPORTED);
+
+	rv = sc_aux_data_get_md_guid(ctx, prkey_info->aux_data, flags, out, out_size);
+	LOG_FUNC_RETURN(ctx, rv);
+}
+
+
 void
 sc_pkcs15_free_key_params(struct sc_pkcs15_key_params *params)
 {
