@@ -26,6 +26,7 @@ extern "C" {
 #endif
 
 #include "libopensc/opensc.h"
+#include "libopensc/aux-data.h"
 
 #define SC_PKCS15_CACHE_DIR		".eid"
 
@@ -128,7 +129,7 @@ struct sc_pkcs15_auth_info {
 	/* authentication method: CHV, SEN, SYMBOLIC, ... */
 	unsigned int  auth_method;
 
-	int tries_left, max_tries;
+	int tries_left, max_tries, logged_in;
 	int max_unlocks;
  };
 typedef struct sc_pkcs15_auth_info sc_pkcs15_auth_info_t;
@@ -161,6 +162,15 @@ struct sc_pkcs15_u8 {
 };
 typedef struct sc_pkcs15_u8 sc_pkcs15_u8_t;
 
+struct sc_pkcs15_data {
+	u8 *data;	/* DER encoded raw data object */
+	size_t data_len;
+};
+typedef struct sc_pkcs15_data sc_pkcs15_data_t;
+
+#define sc_pkcs15_skey sc_pkcs15_data
+#define sc_pkcs15_skey_t sc_pkcs15_data_t
+
 struct sc_pkcs15_pubkey_rsa {
 	sc_pkcs15_bignum_t modulus;
 	sc_pkcs15_bignum_t exponent;
@@ -250,6 +260,7 @@ struct sc_pkcs15_prkey {
 		struct sc_pkcs15_prkey_dsa dsa;
 		struct sc_pkcs15_prkey_ec ec;
 		struct sc_pkcs15_prkey_gostr3410 gostr3410;
+		struct sc_pkcs15_skey secret;
 	} u;
 };
 typedef struct sc_pkcs15_prkey sc_pkcs15_prkey_t;
@@ -275,8 +286,8 @@ struct sc_pkcs15_cert {
 	size_t issuer_len;
 	u8 *subject;
 	size_t subject_len;
-	u8 *crl;
-	size_t crl_len;
+	u8 *extensions;
+	size_t extensions_len;
 
 	struct sc_pkcs15_pubkey * key;
 
@@ -295,12 +306,6 @@ struct sc_pkcs15_cert_info {
 };
 typedef struct sc_pkcs15_cert_info sc_pkcs15_cert_info_t;
 
-struct sc_pkcs15_data {
-	u8 *data;	/* DER encoded raw data object */
-	size_t data_len;
-};
-typedef struct sc_pkcs15_data sc_pkcs15_data_t;
-
 struct sc_pkcs15_data_info {
 	/* FIXME: there is no pkcs15 ID in DataType */
 	struct sc_pkcs15_id id;
@@ -372,56 +377,6 @@ struct sc_pkcs15_key_params {
 	void (*free_params)(void *);
 };
 
-/* From Windows Smart Card Minidriver Specification
- * Version 7.06
- *
- * #define MAX_CONTAINER_NAME_LEN       39
- * #define CONTAINER_MAP_VALID_CONTAINER        1
- * #define CONTAINER_MAP_DEFAULT_CONTAINER      2
- * typedef struct _CONTAINER_MAP_RECORD
- * {
- *      WCHAR wszGuid [MAX_CONTAINER_NAME_LEN + 1];
- *      BYTE bFlags;
- *      BYTE bReserved;
- *      WORD wSigKeySizeBits;
- *      WORD wKeyExchangeKeySizeBits;
- * } CONTAINER_MAP_RECORD, *PCONTAINER_MAP_RECORD;
- */
-#define SC_MD_MAX_CONTAINER_NAME_LEN 39
-#define SC_MD_CONTAINER_MAP_VALID_CONTAINER	0x01
-#define SC_MD_CONTAINER_MAP_DEFAULT_CONTAINER	0x02
-struct sc_md_cmap_record {
-	unsigned char *guid;
-	size_t guid_len;
-	unsigned flags;
-	unsigned keysize_sign;
-	unsigned keysize_keyexchange;
-};
-
-/* From Windows Smart Card Minidriver Specification
- * Version 7.06
- *
- * typedef struct _CARD_CACHE_FILE_FORMAT
- * {
- *	BYTE bVersion;		// Cache version
- *	BYTE bPinsFreshness;	// Card PIN
- *	WORD wContainersFreshness;
- *	WORD wFilesFreshness;
- * } CARD_CACHE_FILE_FORMAT, *PCARD_CACHE_FILE_FORMAT;
- */
-struct sc_md_cardcf {
-	unsigned char version;
-	unsigned char pin_freshness;
-	unsigned cont_freshness;
-	unsigned files_freshness;
-
-};
-
-struct sc_md_data {
-	struct sc_md_cardcf cardcf;
-	void *prop_data;
-};
-
 struct sc_pkcs15_prkey_info {
 	struct sc_pkcs15_id id;	/* correlates to public certificate id */
 	unsigned int usage, access_flags;
@@ -438,8 +393,8 @@ struct sc_pkcs15_prkey_info {
 
 	struct sc_path path;
 
-	/* Used by minidriver and its on-card support */
-	struct sc_md_cmap_record cmap_record;
+	/* Non-pkcs15 data, like MD CMAP record */
+	struct sc_auxiliary_data *aux_data;
 };
 typedef struct sc_pkcs15_prkey_info sc_pkcs15_prkey_info_t;
 
@@ -472,15 +427,12 @@ struct sc_pkcs15_skey_info {
 	int native, key_reference;
 	size_t value_len;
 	unsigned long key_type;
-	int algo_refs[SC_MAX_SUPPORTED_ALGORITHMS];
+	unsigned int algo_refs[SC_MAX_SUPPORTED_ALGORITHMS];
 	struct sc_path path; /* if on card */
 	struct sc_pkcs15_der data;
 };
 typedef struct sc_pkcs15_skey_info sc_pkcs15_skey_info_t;
 
-#define sc_pkcs15_skey sc_pkcs15_data
-#define sc_pkcs15_skey_t sc_pkcs15_data_t
-
 #define SC_PKCS15_TYPE_CLASS_MASK		0xF00
 
 #define SC_PKCS15_TYPE_PRKEY			0x100
@@ -659,6 +611,18 @@ typedef struct sc_pkcs15_card {
 /* flags suitable for struct sc_pkcs15_card */
 #define SC_PKCS15_CARD_FLAG_EMULATED			0x02000000
 
+/* X509 bits for certificate usage extansion */
+#define SC_X509_DIGITAL_SIGNATURE     0x0001UL
+#define SC_X509_NON_REPUDIATION       0x0002UL
+#define SC_X509_KEY_ENCIPHERMENT      0x0004UL
+#define SC_X509_DATA_ENCIPHERMENT     0x0008UL
+#define SC_X509_KEY_AGREEMENT         0x0010UL
+#define SC_X509_KEY_CERT_SIGN         0x0020UL
+#define SC_X509_CRL_SIGN              0x0040UL
+#define SC_X509_ENCIPHER_ONLY         0x0080UL
+#define SC_X509_DECIPHER_ONLY         0x0100UL
+
+
 /* sc_pkcs15_bind:  Binds a card object to a PKCS #15 card object
  * and initializes a new PKCS #15 card object.  Will return
  * SC_ERROR_PKCS15_APP_NOT_FOUND, if the card hasn't got a
@@ -762,6 +726,20 @@ void sc_pkcs15_free_certificate(struct sc_pkcs15_cert *cert);
 int sc_pkcs15_find_cert_by_id(struct sc_pkcs15_card *card,
 			      const struct sc_pkcs15_id *id,
 			      struct sc_pkcs15_object **out);
+int sc_pkcs15_get_name_from_dn(struct sc_context *ctx,
+                              const u8 *dn, size_t dn_len,
+                              const struct sc_object_id *type,
+                              u8 **name, size_t *name_len);
+int sc_pkcs15_get_extension(struct sc_context *ctx,
+                            struct sc_pkcs15_cert *cert,
+                            const struct sc_object_id *type,
+                            u8 **ext_val, size_t *ext_val_len,
+                            int *is_critical);
+int sc_pkcs15_get_bitstring_extension(struct sc_context *ctx,
+                                      struct sc_pkcs15_cert *cert,
+                                      const struct sc_object_id *type,
+                                      unsigned long long *value,
+                                      int *is_critical);
 /* sc_pkcs15_create_cdf:  Creates a new certificate DF on a card pointed
  * by <card>.  Information about the file, such as the file ID, is read
  * from <file>.  <certs> has to be NULL-terminated. */
@@ -790,6 +768,10 @@ int sc_pkcs15_find_skey_by_id(struct sc_pkcs15_card *card,
 int sc_pkcs15_verify_pin(struct sc_pkcs15_card *card,
 			 struct sc_pkcs15_object *pin_obj,
 			 const u8 *pincode, size_t pinlen);
+int sc_pkcs15_verify_pin_with_session_pin(struct sc_pkcs15_card *p15card,
+			 struct sc_pkcs15_object *pin_obj,
+			 const unsigned char *pincode, size_t pinlen,
+			 const unsigned char *sessionpin, size_t *sessionpinlen);
 int sc_pkcs15_change_pin(struct sc_pkcs15_card *card,
 			 struct sc_pkcs15_object *pin_obj,
 			 const u8 *oldpincode, size_t oldpinlen,
@@ -798,6 +780,8 @@ int sc_pkcs15_unblock_pin(struct sc_pkcs15_card *card,
 			 struct sc_pkcs15_object *pin_obj,
 			 const u8 *puk, size_t puklen,
 			 const u8 *newpin, size_t newpinlen);
+int sc_pkcs15_get_pin_info(struct sc_pkcs15_card *card,
+			 struct sc_pkcs15_object *pin_obj);
 int sc_pkcs15_find_pin_by_auth_id(struct sc_pkcs15_card *card,
 				  const struct sc_pkcs15_id *id,
 				  struct sc_pkcs15_object **out);
@@ -845,6 +829,9 @@ int sc_pkcs15_encode_prkdf_entry(struct sc_context *ctx,
 int sc_pkcs15_encode_pukdf_entry(struct sc_context *ctx,
 			const struct sc_pkcs15_object *obj, u8 **buf,
 			size_t *bufsize);
+int sc_pkcs15_encode_skdf_entry(struct sc_context *ctx,
+			const struct sc_pkcs15_object *obj, u8 **buf,
+			size_t *buflen);
 int sc_pkcs15_encode_dodf_entry(struct sc_context *ctx,
 			const struct sc_pkcs15_object *obj, u8 **buf,
 			size_t *bufsize);
@@ -994,7 +981,7 @@ typedef struct sc_pkcs15emu_opt {
 
 #define SC_PKCS15EMU_FLAGS_NO_CHECK	0x00000001
 
-extern int sc_pkcs15_bind_synthetic(struct sc_pkcs15_card *);
+extern int sc_pkcs15_bind_synthetic(struct sc_pkcs15_card *, struct sc_aid *);
 extern int sc_pkcs15_is_emulation_only(sc_card_t *);
 
 int sc_pkcs15emu_object_add(struct sc_pkcs15_card *, unsigned int,
