@@ -50,10 +50,13 @@ const char *sc_get_version(void)
 int sc_hex_to_bin(const char *in, u8 *out, size_t *outlen)
 {
 	int err = SC_SUCCESS;
-	size_t left, count = 0;
+	size_t left, count = 0, in_len;
 
-	assert(in != NULL && out != NULL && outlen != NULL);
-        left = *outlen;
+	if (in == NULL || out == NULL || outlen == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	left = *outlen;
+	in_len = strlen(in);
 
 	while (*in != '\0') {
 		int byte = 0, nybbles = 2;
@@ -76,10 +79,17 @@ int sc_hex_to_bin(const char *in, u8 *out, size_t *outlen)
 			}
 			byte |= c;
 		}
+
+		/* Detect premature end of string before byte is complete */
+		if (in_len > 1 && *in == '\0' && nybbles >= 0) {
+			err = SC_ERROR_INVALID_ARGUMENTS;
+			break;
+		}
+
 		if (*in == ':' || *in == ' ')
 			in++;
 		if (left <= 0) {
-                        err = SC_ERROR_BUFFER_TOO_SMALL;
+			err = SC_ERROR_BUFFER_TOO_SMALL;
 			break;
 		}
 		out[count++] = (u8) byte;
@@ -118,15 +128,20 @@ int sc_bin_to_hex(const u8 *in, size_t in_len, char *out, size_t out_len,
  */
 size_t sc_right_trim(u8 *buf, size_t len) {
 
-	long i;
+	size_t i;
 
-	for(i=len-1; i >=0; i--) {
-		if(!isprint(buf[i])) {
-			buf[i] = '\0';
-			len--;
-			continue;
+	if (!buf)
+		return 0;
+
+	if (len > 0) {
+		for(i = len-1; i > 0; i--) {
+			if(!isprint(buf[i])) {
+				buf[i] = '\0';
+				len--;
+				continue;
+			}
+			break;
 		}
-		break;
 	}
 	return len;
 }
@@ -165,6 +180,13 @@ unsigned short bebytes2ushort(const u8 *buf)
 	return (unsigned short) (buf[0] << 8 | buf[1]);
 }
 
+unsigned short lebytes2ushort(const u8 *buf)
+{
+	if (buf == NULL)
+		return 0U;
+	return (unsigned short)buf[1] << 8 | (unsigned short)buf[0];
+}
+
 void sc_init_oid(struct sc_object_id *oid)
 {
 	int ii;
@@ -213,7 +235,9 @@ int sc_compare_oid(const struct sc_object_id *oid1, const struct sc_object_id *o
 {
 	int i;
 
-	assert(oid1 != NULL && oid2 != NULL);
+	if (oid1 == NULL || oid2 == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	for (i = 0; i < SC_MAX_OBJECT_ID_OCTETS; i++)   {
 		if (oid1->value[i] != oid2->value[i])
@@ -412,8 +436,9 @@ int sc_file_add_acl_entry(sc_file_t *file, unsigned int operation,
 {
 	sc_acl_entry_t *p, *_new;
 
-	assert(file != NULL);
-	assert(operation < SC_MAX_AC_OPS);
+	if (file == NULL || operation >= SC_MAX_AC_OPS) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	switch (method) {
 	case SC_AC_NEVER:
@@ -479,8 +504,9 @@ const sc_acl_entry_t * sc_file_get_acl_entry(const sc_file_t *file,
 		SC_AC_UNKNOWN, SC_AC_KEY_REF_NONE, {{0, 0, 0, {0}}}, NULL
 	};
 
-	assert(file != NULL);
-	assert(operation < SC_MAX_AC_OPS);
+	if (file == NULL || operation >= SC_MAX_AC_OPS) {
+		return NULL;
+	}
 
 	p = file->acl[operation];
 	if (p == (sc_acl_entry_t *) 1)
@@ -497,8 +523,9 @@ void sc_file_clear_acl_entries(sc_file_t *file, unsigned int operation)
 {
 	sc_acl_entry_t *e;
 
-	assert(file != NULL);
-	assert(operation < SC_MAX_AC_OPS);
+	if (file == NULL || operation >= SC_MAX_AC_OPS) {
+		return;
+	}
 
 	e = file->acl[operation];
 	if (e == (sc_acl_entry_t *) 1 ||
@@ -529,7 +556,8 @@ sc_file_t * sc_file_new(void)
 void sc_file_free(sc_file_t *file)
 {
 	unsigned int i;
-	assert(sc_file_valid(file));
+	if (file == NULL || !sc_file_valid(file))
+		return;
 	file->magic = 0;
 	for (i = 0; i < SC_MAX_AC_OPS; i++)
 		sc_file_clear_acl_entries(file, i);
@@ -550,7 +578,8 @@ void sc_file_dup(sc_file_t **dest, const sc_file_t *src)
 	const sc_acl_entry_t *e;
 	unsigned int op;
 
-	assert(sc_file_valid(src));
+	if (!sc_file_valid(src))
+		return;
 	*dest = NULL;
 	newf = sc_file_new();
 	if (newf == NULL)
@@ -587,8 +616,7 @@ void sc_file_dup(sc_file_t **dest, const sc_file_t *src)
 		goto err;
 	return;
 err:
-	if (newf != NULL)
-		sc_file_free(newf);
+	sc_file_free(newf);
 	*dest = NULL;
 }
 
@@ -596,7 +624,9 @@ int sc_file_set_sec_attr(sc_file_t *file, const u8 *sec_attr,
 			 size_t sec_attr_len)
 {
 	u8 *tmp;
-	assert(sc_file_valid(file));
+	if (!sc_file_valid(file)) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	if (sec_attr == NULL) {
 		if (file->sec_attr != NULL)
@@ -624,7 +654,9 @@ int sc_file_set_prop_attr(sc_file_t *file, const u8 *prop_attr,
 			 size_t prop_attr_len)
 {
 	u8 *tmp;
-	assert(sc_file_valid(file));
+	if (!sc_file_valid(file)) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	if (prop_attr == NULL) {
 		if (file->prop_attr != NULL)
@@ -652,7 +684,9 @@ int sc_file_set_type_attr(sc_file_t *file, const u8 *type_attr,
 			 size_t type_attr_len)
 {
 	u8 *tmp;
-	assert(sc_file_valid(file));
+	if (!sc_file_valid(file)) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	if (type_attr == NULL) {
 		if (file->type_attr != NULL)
@@ -681,7 +715,9 @@ int sc_file_set_content(sc_file_t *file, const u8 *content,
 			 size_t content_len)
 {
 	u8 *tmp;
-	assert(sc_file_valid(file));
+	if (!sc_file_valid(file)) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	if (content == NULL) {
 		if (file->encoded_content != NULL)
@@ -709,9 +745,8 @@ int sc_file_set_content(sc_file_t *file, const u8 *content,
 
 
 int sc_file_valid(const sc_file_t *file) {
-#ifndef NDEBUG
-	assert(file != NULL);
-#endif
+	if (file == NULL)
+		return 0;
 	return file->magic == SC_FILE_MAGIC;
 }
 
@@ -798,34 +833,40 @@ void *sc_mem_alloc_secure(sc_context_t *ctx, size_t len)
 
     pointer = calloc(len, sizeof(unsigned char));
     if (!pointer)
-        return NULL;
+	return NULL;
 #ifdef HAVE_SYS_MMAN_H
-    /* TODO Windows support and mprotect too */
+    /* TODO mprotect */
     /* Do not swap the memory */
     if (mlock(pointer, len) >= 0)
-        locked = 1;
+	locked = 1;
+#endif
+#ifdef _WIN32
+	/* Do not swap the memory */
+	if (VirtualLock(pointer, len) != 0)
+		locked = 1;
 #endif
     if (!locked) {
-        if (ctx->paranoid_memory) {
-            sc_do_log (ctx, 0, NULL, 0, NULL, "cannot lock memory, failing allocation because paranoid set");
-            free (pointer);
-            pointer = NULL;
-        } else {
-            sc_do_log (ctx, 0, NULL, 0, NULL, "cannot lock memory, sensitive data may be paged to disk");
-        }
+	if (ctx->flags & SC_CTX_FLAG_PARANOID_MEMORY) {
+	    sc_do_log (ctx, 0, NULL, 0, NULL, "cannot lock memory, failing allocation because paranoid set");
+	    free (pointer);
+	    pointer = NULL;
+	} else {
+	    sc_do_log (ctx, 0, NULL, 0, NULL, "cannot lock memory, sensitive data may be paged to disk");
+	}
     }
     return pointer;
 }
 
 void sc_mem_clear(void *ptr, size_t len)
 {
-#ifdef ENABLE_OPENSSL
 	/* FIXME: Bug in 1.0.0-beta series crashes with 0 length */
-	if (len > 0)
+	if (len > 0)   {
+#ifdef ENABLE_OPENSSL
 		OPENSSL_cleanse(ptr, len);
 #else
-	memset(ptr, 0, len);
+		memset(ptr, 0, len);
 #endif
+	}
 }
 
 int sc_mem_reverse(unsigned char *buf, size_t len)
