@@ -26,6 +26,14 @@
 #include <sys/time.h>
 #endif
 
+#ifdef PKCS11_THREAD_LOCKING
+#if defined(HAVE_PTHREAD)
+#include <pthread.h>
+#elif defined(_WIN32)
+#include <windows.h>
+#endif
+#endif /* PKCS11_THREAD_LOCKING */
+
 #include "sc-pkcs11.h"
 
 #ifndef MODULE_APP_NAME
@@ -42,11 +50,15 @@ pid_t initialized_pid = (pid_t)-1;
 static int in_finalize = 0;
 extern CK_FUNCTION_LIST pkcs11_function_list;
 
-#if defined(HAVE_PTHREAD) && defined(PKCS11_THREAD_LOCKING)
-#include <pthread.h>
+#ifdef PKCS11_THREAD_LOCKING
+
+#if defined(HAVE_PTHREAD)
+
 CK_RV mutex_create(void **mutex)
 {
-	pthread_mutex_t *m = calloc(1, sizeof(*mutex));
+	pthread_mutex_t *m;
+
+	m = calloc(1, sizeof(*m));
 	if (m == NULL)
 		return CKR_GENERAL_ERROR;;
 	pthread_mutex_init(m, NULL);
@@ -79,7 +91,10 @@ CK_RV mutex_destroy(void *p)
 
 static CK_C_INITIALIZE_ARGS _def_locks = {
 	mutex_create, mutex_destroy, mutex_lock, mutex_unlock, 0, NULL };
-#elif defined(_WIN32) && defined (PKCS11_THREAD_LOCKING)
+#define HAVE_OS_LOCKING
+
+#elif defined(_WIN32)
+
 CK_RV mutex_create(void **mutex)
 {
 	CRITICAL_SECTION *m;
@@ -112,14 +127,18 @@ CK_RV mutex_destroy(void *p)
 	free(p);
 	return CKR_OK;
 }
+
 static CK_C_INITIALIZE_ARGS _def_locks = {
 	mutex_create, mutex_destroy, mutex_lock, mutex_unlock, 0, NULL };
+#define HAVE_OS_LOCKING
+
 #endif
 
+#endif /* PKCS11_THREAD_LOCKING */
+
 static CK_C_INITIALIZE_ARGS_PTR	global_locking;
-static void *			global_lock = NULL;
-#if (defined(HAVE_PTHREAD) || defined(_WIN32)) && defined(PKCS11_THREAD_LOCKING)
-#define HAVE_OS_LOCKING
+static void *global_lock = NULL;
+#ifdef HAVE_OS_LOCKING
 static CK_C_INITIALIZE_ARGS_PTR default_mutex_funcs = &_def_locks;
 #else
 static CK_C_INITIALIZE_ARGS_PTR default_mutex_funcs = NULL;
@@ -202,9 +221,11 @@ CK_RV C_Initialize(CK_VOID_PTR pInitArgs)
 	unsigned int i;
 	sc_context_param_t ctx_opts;
 
-	/* Handle fork() exception */
 #if !defined(_WIN32)
+	/* Handle fork() exception */
 	if (current_pid != initialized_pid) {
+		if (context)
+			context->flags |= SC_CTX_FLAG_TERMINATE;
 		C_Finalize(NULL_PTR);
 	}
 	initialized_pid = current_pid;
@@ -236,24 +257,23 @@ CK_RV C_Initialize(CK_VOID_PTR pInitArgs)
 	load_pkcs11_parameters(&sc_pkcs11_conf, context);
 
 	/* List of sessions */
-	list_init(&sessions);
+	if (0 != list_init(&sessions)) {
+		rv = CKR_HOST_MEMORY;
+		goto out;
+	}
 	list_attributes_seeker(&sessions, session_list_seeker);
 
 	/* List of slots */
 	list_init(&virtual_slots);
+	if (0 != list_init(&virtual_slots)) {
+		rv = CKR_HOST_MEMORY;
+		goto out;
+	}
 	list_attributes_seeker(&virtual_slots, slot_list_seeker);
 
-	/* Create a slot for a future "PnP" stuff. */
-	if (sc_pkcs11_conf.plug_and_play) {
-		create_slot(NULL);
-	}
-
 	/* Create slots for readers found on initialization, only if in 2.11 mode */
-	if (!sc_pkcs11_conf.plug_and_play) {
-		for (i=0; i<sc_ctx_get_reader_count(context); i++) {
+	for (i=0; i<sc_ctx_get_reader_count(context); i++)
 			initialize_reader(sc_ctx_get_reader(context, i));
-		}
-	}
 
 out:
 	if (context != NULL)
@@ -303,6 +323,7 @@ CK_RV C_Finalize(CK_VOID_PTR pReserved)
 
 	while ((slot = list_fetch(&virtual_slots))) {
 		list_destroy(&slot->objects);
+		list_destroy(&slot->logins);
 		free(slot);
 	}
 	list_destroy(&virtual_slots);
@@ -331,19 +352,15 @@ CK_RV C_GetInfo(CK_INFO_PTR pInfo)
 
 	memset(pInfo, 0, sizeof(CK_INFO));
 	pInfo->cryptokiVersion.major = 2;
-	if (sc_pkcs11_conf.plug_and_play) {
-		pInfo->cryptokiVersion.minor = 20;
-	} else {
-		pInfo->cryptokiVersion.minor = 11;
-	}
+	pInfo->cryptokiVersion.minor = 20;
 	strcpy_bp(pInfo->manufacturerID,
-		  "OpenSC (www.opensc-project.org)",
+		  OPENSC_VS_FF_COMPANY_NAME,
 		  sizeof(pInfo->manufacturerID));
 	strcpy_bp(pInfo->libraryDescription,
-		  "Smart card PKCS#11 API",
+		  OPENSC_VS_FF_PRODUCT_NAME,
 		  sizeof(pInfo->libraryDescription));
-	pInfo->libraryVersion.major = 0;
-	pInfo->libraryVersion.minor = 0; /* FIXME: use 0.116 for 0.11.6 from autoconf */
+	pInfo->libraryVersion.major = OPENSC_VERSION_MAJOR;
+	pInfo->libraryVersion.minor = OPENSC_VERSION_MINOR;
 
 	sc_pkcs11_unlock();
 	return rv;
@@ -377,15 +394,11 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 		return rv;
 
 	sc_log(context, "C_GetSlotList(token=%d, %s)", tokenPresent,
-		 (pSlotList==NULL_PTR && sc_pkcs11_conf.plug_and_play)? "plug-n-play":"refresh");
+			pSlotList==NULL_PTR? "plug-n-play":"refresh");
 
 	/* Slot list can only change in v2.20 */
-	if (pSlotList == NULL_PTR && sc_pkcs11_conf.plug_and_play) {
-		/* Trick NSS into updating the slot list by changing the hotplug slot ID */
-		sc_pkcs11_slot_t *hotplug_slot = list_get_at(&virtual_slots, 0);
-		hotplug_slot->id--;
+	if (pSlotList == NULL_PTR)
 		sc_ctx_detect_readers(context);
-	}
 
 	card_detect_all();
 
@@ -399,28 +412,32 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 	prev_reader = NULL;
 	numMatches = 0;
 	for (i=0; i<list_size(&virtual_slots); i++) {
-	        slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
+		slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
 		/* the list of available slots contains:
 		 * - if present, virtual hotplug slot;
 		 * - any slot with token;
 		 * - without token(s), one empty slot per reader;
+		 * - any slot that has already been seen;
 		 */
-	        if ((!tokenPresent && !slot->reader)
+		if ((!tokenPresent && !slot->reader)
 				|| (!tokenPresent && slot->reader != prev_reader)
-				|| (slot->slot_info.flags & CKF_TOKEN_PRESENT))
+				|| (slot->slot_info.flags & CKF_TOKEN_PRESENT)
+				|| (slot->flags & SC_PKCS11_SLOT_FLAG_SEEN)) {
 			found[numMatches++] = slot->id;
+			slot->flags |= SC_PKCS11_SLOT_FLAG_SEEN;
+		}
 		prev_reader = slot->reader;
 	}
 
 	if (pSlotList == NULL_PTR) {
-		sc_log(context, "was only a size inquiry (%d)\n", numMatches);
+		sc_log(context, "was only a size inquiry (%lu)\n", numMatches);
 		*pulCount = numMatches;
 		rv = CKR_OK;
 		goto out;
 	}
 
 	if (*pulCount < numMatches) {
-		sc_log(context, "buffer was too small (needed %d)\n", numMatches);
+		sc_log(context, "buffer was too small (needed %lu)\n", numMatches);
 		*pulCount = numMatches;
 		rv = CKR_BUFFER_TOO_SMALL;
 		goto out;
@@ -430,7 +447,7 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 	*pulCount = numMatches;
 	rv = CKR_OK;
 
-	sc_log(context, "returned %d slots\n", numMatches);
+	sc_log(context, "returned %lu slots\n", numMatches);
 
 out:
 	if (found != NULL) {
@@ -483,8 +500,17 @@ CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 
 	sc_log(context, "C_GetSlotInfo(0x%lx)", slotID);
 
+	if (sc_pkcs11_conf.init_sloppy) {
+		/* Most likely virtual_slots only contains the hotplug slot and has not
+		 * been initialized because the caller has *not* called C_GetSlotList
+		 * before C_GetSlotInfo, as required by PKCS#11.  Initialize
+		 * virtual_slots to make things work and hope the caller knows what
+		 * it's doing... */
+		card_detect_all();
+	}
+
 	rv = slot_get_slot(slotID, &slot);
-	sc_log(context, "C_GetSlotInfo() get slot rv %i", rv);
+	sc_log(context, "C_GetSlotInfo() get slot rv %lu", rv);
 	if (rv == CKR_OK)   {
 		if (slot->reader == NULL)   {
 			rv = CKR_TOKEN_NOT_PRESENT;
@@ -494,7 +520,7 @@ CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 			if (now >= slot->slot_state_expires || now == 0) {
 				/* Update slot status */
 				rv = card_detect(slot->reader);
-				sc_log(context, "C_GetSlotInfo() card detect rv 0x%X", rv);
+				sc_log(context, "C_GetSlotInfo() card detect rv 0x%lX", rv);
 
 				if (rv == CKR_TOKEN_NOT_RECOGNIZED || rv == CKR_OK)
 					slot->slot_info.flags |= CKF_TOKEN_PRESENT;
@@ -511,7 +537,7 @@ CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 	if (rv == CKR_OK)
 		memcpy(pInfo, &slot->slot_info, sizeof(CK_SLOT_INFO));
 
-	sc_log(context, "C_GetSlotInfo() flags 0x%X", pInfo->flags);
+	sc_log(context, "C_GetSlotInfo() flags 0x%lX", pInfo->flags);
 	sc_log(context, "C_GetSlotInfo(0x%lx) = %s", slotID, lookup_enum( RV_T, rv));
 	sc_pkcs11_unlock();
 	return rv;
@@ -533,7 +559,7 @@ CK_RV C_GetMechanismList(CK_SLOT_ID slotID,
 
 	rv = slot_get_token(slotID, &slot);
 	if (rv == CKR_OK)
-		rv = sc_pkcs11_get_mechanism_list(slot->card, pMechanismList, pulCount);
+		rv = sc_pkcs11_get_mechanism_list(slot->p11card, pMechanismList, pulCount);
 
 	sc_pkcs11_unlock();
 	return rv;
@@ -555,7 +581,7 @@ CK_RV C_GetMechanismInfo(CK_SLOT_ID slotID,
 
 	rv = slot_get_token(slotID, &slot);
 	if (rv == CKR_OK)
-		rv = sc_pkcs11_get_mechanism_info(slot->card, type, pInfo);
+		rv = sc_pkcs11_get_mechanism_info(slot->p11card, type, pInfo);
 
 	sc_pkcs11_unlock();
 	return rv;
@@ -582,7 +608,8 @@ CK_RV C_InitToken(CK_SLOT_ID slotID,
 		goto out;
 	}
 
-	if (slot->card->framework->init_token == NULL) {
+	if (!slot->p11card || !slot->p11card->framework
+		   || !slot->p11card->framework->init_token) {
 		sc_log(context, "C_InitToken() not supported by framework");
 		rv = CKR_FUNCTION_NOT_SUPPORTED;
 		goto out;
@@ -597,7 +624,7 @@ CK_RV C_InitToken(CK_SLOT_ID slotID,
 		}
 	}
 
-	rv = slot->card->framework->init_token(slot,slot->fw_data, pPin, ulPinLen, pLabel);
+	rv = slot->p11card->framework->init_token(slot, slot->fw_data, pPin, ulPinLen, pLabel);
 	if (rv == CKR_OK) {
 		/* Now we should re-bind all tokens so they get the
 		 * corresponding function vector and flags */
@@ -624,20 +651,17 @@ CK_RV C_WaitForSlotEvent(CK_FLAGS flags,   /* blocking/nonblocking flag */
 		return  CKR_ARGUMENTS_BAD;
 
 	sc_log(context, "C_WaitForSlotEvent(block=%d)", !(flags & CKF_DONT_BLOCK));
+#ifndef PCSCLITE_GOOD
 	/* Not all pcsc-lite versions implement consistently used functions as they are */
-	/* FIXME: add proper checking into build to check correct pcsc-lite version for SCardStatusChange/SCardCancel */
 	if (!(flags & CKF_DONT_BLOCK))
 		return CKR_FUNCTION_NOT_SUPPORTED;
+#endif /* PCSCLITE_GOOD */
 	rv = sc_pkcs11_lock();
 	if (rv != CKR_OK)
 		return rv;
 
-	mask = SC_EVENT_CARD_EVENTS;
-
+	mask = SC_EVENT_CARD_EVENTS | SC_EVENT_READER_EVENTS;
 	/* Detect and add new slots for added readers v2.20 */
-	if (sc_pkcs11_conf.plug_and_play) {
-		mask |= SC_EVENT_READER_EVENTS;
-	}
 
 	rv = slot_find_changed(&slot_id, mask);
 	if ((rv == CKR_OK) || (flags & CKF_DONT_BLOCK))
@@ -647,12 +671,7 @@ again:
 	sc_log(context, "C_WaitForSlotEvent() reader_states:%p", reader_states);
 	sc_pkcs11_unlock();
 	r = sc_wait_for_event(context, mask, &found, &events, -1, &reader_states);
-	if (sc_pkcs11_conf.plug_and_play && events & SC_EVENT_READER_ATTACHED) {
-		/* NSS/Firefox Triggers a C_GetSlotList(NULL) only if a slot ID is returned that it does not know yet
-		   Change the first hotplug slot id on every call to make this happen. */
-		sc_pkcs11_slot_t *hotplug_slot = list_get_at(&virtual_slots, 0);
-		*pSlot= hotplug_slot->id -1;
-
+	if (events & SC_EVENT_READER_ATTACHED) {
 		rv = sc_pkcs11_lock();
 		if (rv != CKR_OK)
 			return rv;
