@@ -114,7 +114,7 @@ CK_RV sc_create_object_int(CK_SESSION_HANDLE hSession,	/* the session's handle *
 		goto out;
 	}
 
-	card = session->slot->card;
+	card = session->slot->p11card;
 	if (card->framework->create_object == NULL)
 		rv = CKR_FUNCTION_NOT_SUPPORTED;
 	else
@@ -333,7 +333,7 @@ C_FindObjectsInit(CK_SESSION_HANDLE hSession,	/* the session's handle */
 	if (rv != CKR_OK)
 		goto out;
 
-	sc_log(context, "C_FindObjectsInit(slot = %d)\n", session->slot->id);
+	sc_log(context, "C_FindObjectsInit(slot = %lu)\n", session->slot->id);
 	dump_template(SC_LOG_DEBUG_NORMAL, "C_FindObjectsInit()", pTemplate, ulCount);
 
 	rv = session_start_operation(session, SC_PKCS11_OPERATION_FIND,
@@ -362,8 +362,9 @@ C_FindObjectsInit(CK_SESSION_HANDLE hSession,	/* the session's handle */
 			if (object->ops->get_attribute(session, object, &private_attribute) != CKR_OK)
 			        continue;
 			if (is_private) {
-				sc_log(context, "Object %d/%d: Private object and not logged in.",
-					 slot->id, object->handle);
+				sc_log(context,
+				       "Object %lu/%lu: Private object and not logged in.",
+				       slot->id, object->handle);
 				continue;
 			}
 		}
@@ -373,20 +374,23 @@ C_FindObjectsInit(CK_SESSION_HANDLE hSession,	/* the session's handle */
 		for (j = 0; j < ulCount; j++) {
 			rv = object->ops->cmp_attribute(session, object, &pTemplate[j]);
 			if (rv == 0) {
-				sc_log(context, "Object %d/%d: Attribute 0x%x does NOT match.",
-					 slot->id, object->handle, pTemplate[j].type);
+				sc_log(context,
+				       "Object %lu/%lu: Attribute 0x%lx does NOT match.",
+				       slot->id, object->handle, pTemplate[j].type);
 				match = 0;
 				break;
 			}
 
 			if (context->debug >= 4) {
-				sc_log(context, "Object %d/%d: Attribute 0x%x matches.",
-					 slot->id, object->handle, pTemplate[j].type);
+				sc_log(context,
+				       "Object %lu/%lu: Attribute 0x%lx matches.",
+				       slot->id, object->handle, pTemplate[j].type);
 			}
 		}
 
 		if (match) {
-			sc_log(context, "Object %d/%d matches\n", slot->id, object->handle);
+			sc_log(context, "Object %lu/%lu matches\n", slot->id,
+			       object->handle);
 			/* Realloc handles - remove restriction on only 32 matching objects -dee */
 			if (operation->num_handles >= operation->allocated_handles) {
 				operation->allocated_handles += SC_PKCS11_FIND_INC_HANDLES;
@@ -541,7 +545,8 @@ C_Digest(CK_SESSION_HANDLE hSession,		/* the session's handle */
 	if (rv == CKR_OK)
 		rv = sc_pkcs11_md_final(session, pDigest, pulDigestLen);
 
-out:	sc_log(context, "C_Digest() = %s", lookup_enum ( RV_T, rv ));
+out:
+	sc_log(context, "C_Digest() = %s", lookup_enum ( RV_T, rv ));
 	sc_pkcs11_unlock();
 	return rv;
 }
@@ -685,8 +690,12 @@ C_Sign(CK_SESSION_HANDLE hSession,		/* the session's handle */
 	}
 
 	rv = sc_pkcs11_sign_update(session, pData, ulDataLen);
-	if (rv == CKR_OK)
-		rv = sc_pkcs11_sign_final(session, pSignature, pulSignatureLen);
+	if (rv == CKR_OK) {
+		rv = restore_login_state(session->slot);
+		if (rv == CKR_OK)
+			rv = sc_pkcs11_sign_final(session, pSignature, pulSignatureLen);
+		rv = reset_login_state(session->slot, rv);
+	}
 
 out:
 	sc_log(context, "C_Sign() = %s", lookup_enum ( RV_T, rv ));
@@ -746,7 +755,10 @@ C_SignFinal(CK_SESSION_HANDLE hSession,		/* the session's handle */
 		*pulSignatureLen = length;
 		rv = pSignature ? CKR_BUFFER_TOO_SMALL : CKR_OK;
 	} else {
-		rv = sc_pkcs11_sign_final(session, pSignature, pulSignatureLen);
+		rv = restore_login_state(session->slot);
+		if (rv == CKR_OK)
+			rv = sc_pkcs11_sign_final(session, pSignature, pulSignatureLen);
+		rv = reset_login_state(session->slot, rv);
 	}
 
 out:
@@ -859,7 +871,8 @@ CK_RV C_DecryptInit(CK_SESSION_HANDLE hSession,	/* the session's handle */
 
 	rv = sc_pkcs11_decr_init(session, pMechanism, object, key_type);
 
-out:	sc_log(context, "C_DecryptInit() = %s", lookup_enum ( RV_T, rv ));
+out:
+	sc_log(context, "C_DecryptInit() = %s", lookup_enum ( RV_T, rv ));
 	sc_pkcs11_unlock();
 	return rv;
 }
@@ -878,9 +891,14 @@ CK_RV C_Decrypt(CK_SESSION_HANDLE hSession,	/* the session's handle */
 		return rv;
 
 	rv = get_session(hSession, &session);
-	if (rv == CKR_OK)
-		rv = sc_pkcs11_decr(session, pEncryptedData, ulEncryptedDataLen,
-				pData, pulDataLen);
+	if (rv == CKR_OK) {
+		rv = restore_login_state(session->slot);
+		if (rv == CKR_OK) {
+			rv = sc_pkcs11_decr(session, pEncryptedData,
+					ulEncryptedDataLen, pData, pulDataLen);
+		}
+		rv = reset_login_state(session->slot, rv);
+	}
 
 	sc_log(context, "C_Decrypt() = %s", lookup_enum ( RV_T, rv ));
 	sc_pkcs11_unlock();
@@ -983,13 +1001,17 @@ CK_RV C_GenerateKeyPair(CK_SESSION_HANDLE hSession,	/* the session's handle */
 	}
 
 	slot = session->slot;
-	if (slot->card->framework->gen_keypair == NULL)
+	if (slot->p11card->framework->gen_keypair == NULL)
 		rv = CKR_FUNCTION_NOT_SUPPORTED;
-	else
-		rv = slot->card->framework->gen_keypair(slot, pMechanism,
-				pPublicKeyTemplate, ulPublicKeyAttributeCount,
-				pPrivateKeyTemplate, ulPrivateKeyAttributeCount,
-				phPublicKey, phPrivateKey);
+	else {
+		rv = restore_login_state(slot);
+		if (rv == CKR_OK)
+			rv = slot->p11card->framework->gen_keypair(slot, pMechanism,
+					pPublicKeyTemplate, ulPublicKeyAttributeCount,
+					pPrivateKeyTemplate, ulPrivateKeyAttributeCount,
+					phPublicKey, phPrivateKey);
+		rv = reset_login_state(session->slot, rv);
+	}
 
 out:
 	sc_pkcs11_unlock();
@@ -1084,9 +1106,12 @@ CK_RV C_DeriveKey(CK_SESSION_HANDLE hSession,	/* the session's handle */
 			goto out;
 		}
 
-		rv = sc_pkcs11_deri(session, pMechanism, object, key_type,
-			hSession, *phKey, key_object);
+		rv = restore_login_state(session->slot);
+		if (rv == CKR_OK)
+			rv = sc_pkcs11_deri(session, pMechanism, object, key_type,
+					hSession, *phKey, key_object);
 		/* TODO if (rv != CK_OK) need to destroy the object */
+		rv = reset_login_state(session->slot, rv);
 
 		break;
 	    default:
@@ -1120,10 +1145,10 @@ CK_RV C_GenerateRandom(CK_SESSION_HANDLE hSession,	/* the session's handle */
 	rv = get_session(hSession, &session);
 	if (rv == CKR_OK) {
 		slot = session->slot;
-		if (slot->card->framework->get_random == NULL)
+		if (slot->p11card->framework->get_random == NULL)
 			rv = CKR_RANDOM_NO_RNG;
 		else
-			rv = slot->card->framework->get_random(slot, RandomData, ulRandomLen);
+			rv = slot->p11card->framework->get_random(slot, RandomData, ulRandomLen);
 	}
 
 	sc_pkcs11_unlock();
@@ -1175,7 +1200,8 @@ CK_RV C_VerifyInit(CK_SESSION_HANDLE hSession,	/* the session's handle */
 
 	rv = sc_pkcs11_verif_init(session, pMechanism, object, key_type);
 
-out:	sc_log(context, "C_VerifyInit() = %s", lookup_enum ( RV_T, rv ));
+out:
+	sc_log(context, "C_VerifyInit() = %s", lookup_enum ( RV_T, rv ));
 	sc_pkcs11_unlock();
 	return rv;
 #endif
@@ -1202,10 +1228,15 @@ CK_RV C_Verify(CK_SESSION_HANDLE hSession,	/* the session's handle */
 		goto out;
 
 	rv = sc_pkcs11_verif_update(session, pData, ulDataLen);
-	if (rv == CKR_OK)
-		rv = sc_pkcs11_verif_final(session, pSignature, ulSignatureLen);
+	if (rv == CKR_OK) {
+		rv = restore_login_state(session->slot);
+		if (rv == CKR_OK)
+			rv = sc_pkcs11_verif_final(session, pSignature, ulSignatureLen);
+		rv = reset_login_state(session->slot, rv);
+	}
 
-out:	sc_log(context, "C_Verify() = %s", lookup_enum ( RV_T, rv ));
+out:
+	sc_log(context, "C_Verify() = %s", lookup_enum ( RV_T, rv ));
 	sc_pkcs11_unlock();
 	return rv;
 #endif
@@ -1250,8 +1281,12 @@ CK_RV C_VerifyFinal(CK_SESSION_HANDLE hSession,	/* the session's handle */
 		return rv;
 
 	rv = get_session(hSession, &session);
-	if (rv == CKR_OK)
-		rv = sc_pkcs11_verif_final(session, pSignature, ulSignatureLen);
+	if (rv == CKR_OK) {
+		rv = restore_login_state(session->slot);
+		if (rv == CKR_OK)
+			rv = sc_pkcs11_verif_final(session, pSignature, ulSignatureLen);
+		rv = reset_login_state(session->slot, rv);
+	}
 
 	sc_log(context, "C_VerifyFinal() = %s", lookup_enum ( RV_T, rv ));
 	sc_pkcs11_unlock();
