@@ -49,9 +49,10 @@ extern "C" {
 
 #define SC_PKCS11_SLOT_FOR_PIN_USER	1
 #define SC_PKCS11_SLOT_FOR_PIN_SIGN	2
-#define SC_PKCS11_SLOT_FOR_APPLICATION	4
 #define SC_PKCS11_SLOT_CREATE_ALL	8
 
+#define SC_PKCS11_SLOT_FOR_PINS		(SC_PKCS11_SLOT_FOR_PIN_USER | SC_PKCS11_SLOT_FOR_PIN_SIGN)
+
 extern void *C_LoadModule(const char *name, CK_FUNCTION_LIST_PTR_PTR);
 extern CK_RV C_UnloadModule(void *module);
 
@@ -78,6 +79,8 @@ struct sc_pkcs11_config {
 	unsigned int slots_per_card;
 	unsigned char hide_empty_tokens;
 	unsigned char lock_login;
+	unsigned char atomic;
+	unsigned char init_sloppy;
 	unsigned int pin_unblock_style;
 	unsigned int create_puk_slot;
 	unsigned int zero_ckaid_for_ca_certs;
@@ -150,7 +153,7 @@ struct sc_pkcs11_framework_ops {
 
 	/* Create tokens to virtual slots and
 	 * objects in tokens; called after bind */
-	CK_RV (*create_tokens)(struct sc_pkcs11_card *, struct sc_app_info *, struct sc_pkcs11_slot **);
+	CK_RV (*create_tokens)(struct sc_pkcs11_card *, struct sc_app_info *);
 	CK_RV (*release_token)(struct sc_pkcs11_card *, void *);
 
 	/* Login and logout */
@@ -202,13 +205,19 @@ struct sc_pkcs11_card {
 	unsigned int nmechanisms;
 };
 
+/* If the slot did already show with `C_GetSlotList`, then we need to keep this
+ * slot alive. PKCS#11 2.30 allows allows adding but not removing slots until
+ * the application calls `C_GetSlotList` with `NULL`. This flag tracks the
+ * visibility to the application */
+#define SC_PKCS11_SLOT_FLAG_SEEN 1
+
 struct sc_pkcs11_slot {
 	CK_SLOT_ID id;			/* ID of the slot */
 	int login_user;			/* Currently logged in user */
 	CK_SLOT_INFO slot_info;		/* Slot specific information (information about reader) */
 	CK_TOKEN_INFO token_info;	/* Token specific information (information about card) */
 	sc_reader_t *reader;		/* same as card->reader if there's a card present */
-	struct sc_pkcs11_card *card;	/* The card associated with this slot */
+	struct sc_pkcs11_card *p11card;	/* The card associated with this slot */
 	unsigned int events;		/* Card events SC_EVENT_CARD_{INSERTED,REMOVED} */
 	void *fw_data;			/* Framework specific data */  /* TODO: get know how it used */
 	list_t objects;			/* Objects in this slot */
@@ -217,6 +226,8 @@ struct sc_pkcs11_slot {
 
 	int fw_data_idx;		/* Index of framework data */
 	struct sc_app_info *app_info;	/* Application assosiated to slot */
+	list_t logins;			/* tracks all calls to C_Login if atomic operations are requested */
+	int flags;
 };
 typedef struct sc_pkcs11_slot sc_pkcs11_slot_t;
 
@@ -346,6 +357,15 @@ CK_RV slot_get_token(CK_SLOT_ID id, struct sc_pkcs11_slot **);
 CK_RV slot_token_removed(CK_SLOT_ID id);
 CK_RV slot_allocate(struct sc_pkcs11_slot **, struct sc_pkcs11_card *);
 CK_RV slot_find_changed(CK_SLOT_ID_PTR idp, int mask);
+int slot_get_logged_in_state(struct sc_pkcs11_slot *slot);
+
+/* Login tracking functions */
+CK_RV restore_login_state(struct sc_pkcs11_slot *slot);
+CK_RV reset_login_state(struct sc_pkcs11_slot *slot, CK_RV rv);
+CK_RV push_login_state(struct sc_pkcs11_slot *slot,
+		CK_USER_TYPE userType, CK_CHAR_PTR pPin, CK_ULONG ulPinLen);
+void pop_login_state(struct sc_pkcs11_slot *slot);
+void pop_all_login_states(struct sc_pkcs11_slot *slot);
 
 /* Session manipulation */
 CK_RV get_session(CK_SESSION_HANDLE hSession, struct sc_pkcs11_session ** session);
@@ -371,6 +391,8 @@ CK_RV attr_find(CK_ATTRIBUTE_PTR, CK_ULONG, CK_ULONG, void *, size_t *);
 CK_RV attr_find2(CK_ATTRIBUTE_PTR, CK_ULONG, CK_ATTRIBUTE_PTR, CK_ULONG,
 		CK_ULONG, void *, size_t *);
 CK_RV attr_find_ptr(CK_ATTRIBUTE_PTR, CK_ULONG, CK_ULONG, void **, size_t *);
+CK_RV attr_find_ptr2(CK_ATTRIBUTE_PTR pTemp1, CK_ULONG ulCount1,
+		CK_ATTRIBUTE_PTR pTemp2, CK_ULONG ulCount2, CK_ULONG type, void **ptr, size_t * sizep);
 CK_RV attr_find_and_allocate_ptr(CK_ATTRIBUTE_PTR, CK_ULONG, CK_ULONG, void **, size_t *);
 CK_RV attr_find_var(CK_ATTRIBUTE_PTR, CK_ULONG, CK_ULONG, void *, size_t *);
 CK_RV attr_extract(CK_ATTRIBUTE_PTR, void *, size_t *);
