@@ -50,13 +50,20 @@ static struct sc_pkcs11_slot * reader_get_slot(sc_reader_t *reader)
 	return NULL;
 }
 
-static void init_slot_info(CK_SLOT_INFO_PTR pInfo)
+static void init_slot_info(CK_SLOT_INFO_PTR pInfo, sc_reader_t *reader)
 {
-	strcpy_bp(pInfo->slotDescription, "Virtual hotplug slot", 64);
-	strcpy_bp(pInfo->manufacturerID, "OpenSC (www.opensc-project.org)", 32);
+	if (reader) {
+		strcpy_bp(pInfo->slotDescription, reader->name, 64);
+		strcpy_bp(pInfo->manufacturerID, reader->vendor, 32);
+		pInfo->hardwareVersion.major = reader->version_major;
+		pInfo->hardwareVersion.minor = reader->version_minor;
+	} else {
+		strcpy_bp(pInfo->slotDescription, "Virtual hotplug slot", 64);
+		strcpy_bp(pInfo->manufacturerID, OPENSC_VS_FF_COMPANY_NAME, 32);
+		pInfo->hardwareVersion.major = OPENSC_VERSION_MAJOR;
+		pInfo->hardwareVersion.minor = OPENSC_VERSION_MINOR;
+	}
 	pInfo->flags = CKF_REMOVABLE_DEVICE | CKF_HW_SLOT;
-	pInfo->hardwareVersion.major = 0;
-	pInfo->hardwareVersion.minor = 0;
 	pInfo->firmwareVersion.major = 0;
 	pInfo->firmwareVersion.minor = 0;
 }
@@ -75,38 +82,69 @@ static int object_list_seeker(const void *el, const void *key)
 
 CK_RV create_slot(sc_reader_t *reader)
 {
-	struct sc_pkcs11_slot *slot;
+	/* find unused virtual hotplug slots */
+	struct sc_pkcs11_slot *slot = reader_get_slot(NULL);
 
-	if (list_size(&virtual_slots) >= sc_pkcs11_conf.max_virtual_slots)
-		return CKR_FUNCTION_FAILED;
+	/* create a new slot if no empty slot is available */
+	if (!slot) {
+		if (list_size(&virtual_slots) >= sc_pkcs11_conf.max_virtual_slots)
+			return CKR_FUNCTION_FAILED;
 
-	slot = (struct sc_pkcs11_slot *)calloc(1, sizeof(struct sc_pkcs11_slot));
-	if (!slot)
-		return CKR_HOST_MEMORY;
+		slot = (struct sc_pkcs11_slot *)calloc(1, sizeof(struct sc_pkcs11_slot));
+		if (!slot)
+			return CKR_HOST_MEMORY;
+
+		list_append(&virtual_slots, slot);
+		if (0 != list_init(&slot->objects)) {
+			return CKR_HOST_MEMORY;
+		}
+		list_attributes_seeker(&slot->objects, object_list_seeker);
+
+		if (0 != list_init(&slot->logins)) {
+			return CKR_HOST_MEMORY;
+		}
+	} else {
+		/* reuse the old list of logins/objects since they should be empty */
+		list_t logins = slot->logins;
+		list_t objects = slot->objects;
+
+		memset(slot, 0, sizeof *slot);
+
+		slot->logins = logins;
+		slot->objects = objects;
+	}
 
-	list_append(&virtual_slots, slot);
 	slot->login_user = -1;
 	slot->id = (CK_SLOT_ID) list_locate(&virtual_slots, slot);
-	sc_log(context, "Creating slot with id 0x%lx", slot->id);
+	init_slot_info(&slot->slot_info, reader);
+	sc_log(context, "Initializing slot with id 0x%lx", slot->id);
 
-	list_init(&slot->objects);
-	list_attributes_seeker(&slot->objects, object_list_seeker);
-
-	init_slot_info(&slot->slot_info);
 	if (reader != NULL) {
 		slot->reader = reader;
+		strcpy_bp(slot->slot_info.manufacturerID, reader->vendor, 32);
 		strcpy_bp(slot->slot_info.slotDescription, reader->name, 64);
+		slot->slot_info.hardwareVersion.major = reader->version_major;
+		slot->slot_info.hardwareVersion.minor = reader->version_minor;
 	}
 
 	return CKR_OK;
 }
 
-void delete_slot(struct sc_pkcs11_slot *slot)
+void empty_slot(struct sc_pkcs11_slot *slot)
 {
 	if (slot) {
-		list_destroy(&slot->objects);
-		list_delete(&virtual_slots, slot);
-		free(slot);
+		if (slot->flags & SC_PKCS11_SLOT_FLAG_SEEN) {
+			/* Keep the slot visible to the application. The slot's state has
+			 * already been reset by `slot_token_removed()`, lists have been
+			 * emptied. We replace the reader with a virtual hotplug slot. */
+			slot->reader = NULL;
+			init_slot_info(&slot->slot_info, NULL);
+		} else {
+			list_destroy(&slot->objects);
+			list_destroy(&slot->logins);
+			list_delete(&virtual_slots, slot);
+			free(slot);
+		}
 	}
 }
 
@@ -152,7 +190,7 @@ CK_RV initialize_reader(sc_reader_t *reader)
 CK_RV card_removed(sc_reader_t * reader)
 {
 	unsigned int i;
-	struct sc_pkcs11_card *card = NULL;
+	struct sc_pkcs11_card *p11card = NULL;
 	/* Mark all slots as "token not present" */
 	sc_log(context, "%s: card removed", reader->name);
 
@@ -161,23 +199,23 @@ CK_RV card_removed(sc_reader_t * reader)
 		sc_pkcs11_slot_t *slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
 		if (slot->reader == reader) {
 			/* Save the "card" object */
-			if (slot->card)
-				card = slot->card;
+			if (slot->p11card)
+				p11card = slot->p11card;
 			slot_token_removed(slot->id);
 		}
 	}
 
-	if (card) {
-		card->framework->unbind(card);
-		sc_disconnect_card(card->card);
-		for (i=0; i < card->nmechanisms; ++i) {
-			if (card->mechanisms[i]->free_mech_data) {
-				card->mechanisms[i]->free_mech_data(card->mechanisms[i]->mech_data);
+	if (p11card) {
+		p11card->framework->unbind(p11card);
+		sc_disconnect_card(p11card->card);
+		for (i=0; i < p11card->nmechanisms; ++i) {
+			if (p11card->mechanisms[i]->free_mech_data) {
+				p11card->mechanisms[i]->free_mech_data(p11card->mechanisms[i]->mech_data);
 			}
-			free(card->mechanisms[i]);
+			free(p11card->mechanisms[i]);
 		}
-		free(card->mechanisms);
-		free(card);
+		free(p11card->mechanisms);
+		free(p11card);
 	}
 
 	return CKR_OK;
@@ -224,7 +262,7 @@ again:
 	for (i=0; i<list_size(&virtual_slots); i++) {
 		sc_pkcs11_slot_t *slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
 		if (slot->reader == reader) {
-			p11card = slot->card;
+			p11card = slot->p11card;
 			break;
 		}
 	}
@@ -246,13 +284,24 @@ again:
 			return sc_to_cryptoki_error(rc, NULL);
 		}
 
+		/* escape commands are only guaranteed to be working with a card
+		 * inserted. That's why by now, after sc_connect_card() the reader's
+		 * metadata may have changed. We re-initialize the metadata for every
+		 * slot of this reader here. */
+		if (reader->flags & SC_READER_ENABLE_ESCAPE) {
+			for (i = 0; i<list_size(&virtual_slots); i++) {
+				sc_pkcs11_slot_t *slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
+				if (slot->reader == reader)
+					init_slot_info(&slot->slot_info, reader);
+			}
+		}
+
 		sc_log(context, "%s: Connected SC card %p", reader->name, p11card->card);
 	}
 
 	/* Detect the framework */
 	if (p11card->framework == NULL) {
 		struct sc_app_info *app_generic = sc_pkcs15_get_application_by_type(p11card->card, "generic");
-		struct sc_pkcs11_slot *first_slot = NULL;
 
 		sc_log(context, "%s: Detecting Framework. %i on-card applications", reader->name, p11card->card->app_count);
 		sc_log(context, "%s: generic application %s", reader->name, app_generic ? app_generic->label : "<none>");
@@ -284,14 +333,18 @@ again:
 				rv = CKR_OK;
 			}
 			if (rv != CKR_OK)   {
-				sc_log(context, "%s: cannot bind 'generic' token: rv 0x%X", reader->name, rv);
+				sc_log(context,
+				       "%s: cannot bind 'generic' token: rv 0x%lX",
+				       reader->name, rv);
 				return rv;
 			}
 
 			sc_log(context, "%s: Creating 'generic' token.", reader->name);
-			rv = frameworks[i]->create_tokens(p11card, app_generic, &first_slot);
+			rv = frameworks[i]->create_tokens(p11card, app_generic);
 			if (rv != CKR_OK)   {
-				sc_log(context, "%s: create 'generic' token error 0x%X", reader->name, rv);
+				sc_log(context,
+				       "%s: create 'generic' token error 0x%lX",
+				       reader->name, rv);
 				return rv;
 			}
 		}
@@ -307,14 +360,17 @@ again:
 			sc_log(context, "%s: Binding %s token.", reader->name, app_name);
 			rv = frameworks[i]->bind(p11card, app_info);
 			if (rv != CKR_OK)   {
-				sc_log(context, "%s: bind %s token error Ox%X", reader->name, app_name, rv);
+				sc_log(context, "%s: bind %s token error Ox%lX",
+				       reader->name, app_name, rv);
 				continue;
 			}
 
 			sc_log(context, "%s: Creating %s token.", reader->name, app_name);
-			rv = frameworks[i]->create_tokens(p11card, app_info, &first_slot);
+			rv = frameworks[i]->create_tokens(p11card, app_info);
 			if (rv != CKR_OK)   {
-				sc_log(context, "%s: create %s token error 0x%X", reader->name, app_name, rv);
+				sc_log(context,
+				       "%s: create %s token error 0x%lX",
+				       reader->name, app_name, rv);
 				return rv;
 			}
 		}
@@ -338,14 +394,15 @@ card_detect_all(void)
 			struct sc_pkcs11_slot *slot;
 			card_removed(reader);
 			while ((slot = reader_get_slot(reader))) {
-				delete_slot(slot);
+				empty_slot(slot);
 			}
 			_sc_delete_reader(context, reader);
 			i--;
 		} else {
 			if (!reader_get_slot(reader))
 				initialize_reader(reader);
-			card_detect(sc_ctx_get_reader(context, i));
+			else
+				card_detect(sc_ctx_get_reader(context, i));
 		}
 	}
 	sc_log(context, "All cards detected");
@@ -353,7 +410,7 @@ card_detect_all(void)
 }
 
 /* Allocates an existing slot to a card */
-CK_RV slot_allocate(struct sc_pkcs11_slot ** slot, struct sc_pkcs11_card * card)
+CK_RV slot_allocate(struct sc_pkcs11_slot ** slot, struct sc_pkcs11_card * p11card)
 {
 	unsigned int i;
 	struct sc_pkcs11_slot *tmp_slot = NULL;
@@ -361,13 +418,13 @@ CK_RV slot_allocate(struct sc_pkcs11_slot ** slot, struct sc_pkcs11_card * card)
 	/* Locate a free slot for this reader */
 	for (i=0; i< list_size(&virtual_slots); i++) {
 		tmp_slot = (struct sc_pkcs11_slot *)list_get_at(&virtual_slots, i);
-		if (tmp_slot->reader == card->reader && tmp_slot->card == NULL)
+		if (tmp_slot->reader == p11card->reader && tmp_slot->p11card == NULL)
 			break;
 	}
 	if (!tmp_slot || (i == list_size(&virtual_slots)))
 		return CKR_FUNCTION_FAILED;
-	sc_log(context, "Allocated slot 0x%lx for card in reader %s", tmp_slot->id, card->reader->name);
-	tmp_slot->card = card;
+	sc_log(context, "Allocated slot 0x%lx for card in reader %s", tmp_slot->id, p11card->reader->name);
+	tmp_slot->p11card = p11card;
 	tmp_slot->events = SC_EVENT_CARD_INSERTED;
 	*slot = tmp_slot;
 	return CKR_OK;
@@ -432,20 +489,25 @@ CK_RV slot_token_removed(CK_SLOT_ID id)
 	}
 
 	/* Release framework stuff */
-	if (slot->card != NULL) {
+	if (slot->p11card != NULL) {
 		if (slot->fw_data != NULL &&
-				slot->card->framework != NULL && slot->card->framework->release_token != NULL)
-			slot->card->framework->release_token(slot->card, slot->fw_data);
+				slot->p11card->framework != NULL && slot->p11card->framework->release_token != NULL) {
+			slot->p11card->framework->release_token(slot->p11card, slot->fw_data);
+			slot->fw_data = NULL;
+		}
+		slot->p11card = NULL;
 	}
 
 	/* Reset relevant slot properties */
 	slot->slot_info.flags &= ~CKF_TOKEN_PRESENT;
 	slot->login_user = -1;
-	slot->card = NULL;
+	pop_all_login_states(slot);
 
 	if (token_was_present)
 		slot->events = SC_EVENT_CARD_REMOVED;
 
+	memset(&slot->token_info, 0, sizeof slot->token_info);
+
 	return CKR_OK;
 }
 
@@ -458,7 +520,9 @@ CK_RV slot_find_changed(CK_SLOT_ID_PTR idp, int mask)
 	card_detect_all();
 	for (i=0; i<list_size(&virtual_slots); i++) {
 		sc_pkcs11_slot_t *slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
-		sc_log(context, "slot 0x%lx token: %d events: 0x%02X",slot->id, (slot->slot_info.flags & CKF_TOKEN_PRESENT), slot->events);
+		sc_log(context, "slot 0x%lx token: %lu events: 0x%02X",
+		       slot->id, (slot->slot_info.flags & CKF_TOKEN_PRESENT),
+		       slot->events);
 		if ((slot->events & SC_EVENT_CARD_INSERTED)
 				&& !(slot->slot_info.flags & CKF_TOKEN_PRESENT)) {
 			/* If a token has not been initialized, clear the inserted event */
