@@ -254,8 +254,9 @@ authentic_pkcs15_new_file(struct sc_profile *profile, struct sc_card *card,
 		file->path.count = -1;
 	}
 
-	sc_log(ctx, "file(size:%i,type:%i/%i,id:%04X), path(type:%X,'%s')",  file->size, file->type, file->ef_structure, file->id,
-			file->path.type, sc_print_path(&file->path));
+	sc_log(ctx, "file(size:%"SC_FORMAT_LEN_SIZE_T"u,type:%i/%i,id:%04X), path(type:%X,'%s')",
+	       file->size, file->type, file->ef_structure, file->id,
+	       file->path.type, sc_print_path(&file->path));
 	if (out)
 		*out = file;
 	else
@@ -350,8 +351,10 @@ authentic_sdo_allocate_prvkey(struct sc_profile *profile, struct sc_card *card,
 	LOG_TEST_RET(ctx, rv, "Cannot instantiate new PRKEY-RSA file");
 
 	sdo = calloc(1, sizeof(struct sc_authentic_sdo));
-	if (!sdo)
+	if (!sdo) {
+		sc_file_free(file);
 		LOG_TEST_RET(ctx, SC_ERROR_OUT_OF_MEMORY, "Cannot allocate 'sc_authentic_sdo'");
+	}
 	*out = sdo;
 
 	sdo->magic = AUTHENTIC_SDO_MAGIC;
@@ -500,8 +503,7 @@ authentic_free_sdo_data(struct sc_authentic_sdo *sdo)
 	if (!sdo)
 		return;
 
-	if (sdo->file)
-		sc_file_free(sdo->file);
+	sc_file_free(sdo->file);
 
 	for (ii=0; ii<rsa_mechs_num; ii++)
 		if (sdo->docp.mech == authentic_v3_rsa_mechs[ii])
@@ -524,8 +526,10 @@ authentic_pkcs15_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p
 	int	 rv;
 
 	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "create private key(keybits:%i,usage:%X,access:%X,ref:%X)", keybits,
-			key_info->usage, key_info->access_flags, key_info->key_reference);
+	sc_log(ctx,
+	       "create private key(keybits:%"SC_FORMAT_LEN_SIZE_T"u,usage:%X,access:%X,ref:%X)",
+	       keybits, key_info->usage, key_info->access_flags,
+	       key_info->key_reference);
 	if (keybits < 1024 || keybits > 2048 || (keybits % 256))
 		LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Invalid RSA key size");
 
@@ -546,8 +550,8 @@ authentic_pkcs15_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p
 		| SC_PKCS15_PRKEY_ACCESS_ALWAYSSENSITIVE
 		| SC_PKCS15_PRKEY_ACCESS_SENSITIVE;
 
-        rv = authentic_sdo_allocate_prvkey(profile, card, key_info, &sdo);
-        LOG_TEST_RET(ctx, rv, "IasEcc: init SDO private key failed");
+	rv = authentic_sdo_allocate_prvkey(profile, card, key_info, &sdo);
+	LOG_TEST_RET(ctx, rv, "IasEcc: init SDO private key failed");
 
 	rv = sc_card_ctl(card, SC_CARDCTL_AUTHENTIC_SDO_CREATE, sdo);
 	if (rv == SC_ERROR_FILE_ALREADY_EXISTS)   {
@@ -603,8 +607,10 @@ authentic_pkcs15_generate_key(struct sc_profile *profile, sc_pkcs15_card_t *p15c
 	int rv;
 
 	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "generate key(bits:%i,path:%s,AuthID:%s\n", keybits,
-			sc_print_path(&key_info->path), sc_pkcs15_print_id(&object->auth_id));
+	sc_log(ctx,
+	       "generate key(bits:%"SC_FORMAT_LEN_SIZE_T"u,path:%s,AuthID:%s\n",
+	       keybits, sc_print_path(&key_info->path),
+	       sc_pkcs15_print_id(&object->auth_id));
 
 	if (!object->content.value || object->content.len != sizeof(struct sc_authentic_sdo))
 		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "Invalid PrKey SDO data");
@@ -668,8 +674,10 @@ authentic_pkcs15_store_key(struct sc_profile *profile, struct sc_pkcs15_card *p1
 	int rv;
 
 	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "Store IAS/ECC key(keybits:%i,AuthID:%s,path:%s)",
-			keybits, sc_pkcs15_print_id(&object->auth_id), sc_print_path(&key_info->path));
+	sc_log(ctx,
+	       "Store IAS/ECC key(keybits:%"SC_FORMAT_LEN_SIZE_T"u,AuthID:%s,path:%s)",
+	       keybits, sc_pkcs15_print_id(&object->auth_id),
+	       sc_print_path(&key_info->path));
 
 	if (!object->content.value || object->content.len != sizeof(struct sc_authentic_sdo))
 		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "Invalid PrKey SDO data");
@@ -715,7 +723,8 @@ authentic_pkcs15_delete_rsa_sdo (struct sc_profile *profile, struct sc_pkcs15_ca
 	int rv;
 
 	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "delete SDO RSA key (ref:%i,size:%i)", key_info->key_reference, key_info->modulus_length);
+	sc_log(ctx, "delete SDO RSA key (ref:%i,size:%"SC_FORMAT_LEN_SIZE_T"u)",
+	       key_info->key_reference, key_info->modulus_length);
 
 	rv = authentic_pkcs15_new_file(profile, p15card->card, SC_PKCS15_TYPE_PRKEY_RSA, key_info->key_reference, &file);
 	LOG_TEST_GOTO_ERR(ctx, rv, "PRKEY_RSA instantiation file error");
@@ -735,8 +744,7 @@ authentic_pkcs15_delete_rsa_sdo (struct sc_profile *profile, struct sc_pkcs15_ca
 	LOG_TEST_GOTO_ERR(ctx, rv, "SC_CARDCTL_AUTHENTIC_SDO_DELETE failed for private key");
 
 err:
-	if (file)
-		sc_file_free(file);
+	sc_file_free(file);
 	LOG_FUNC_RETURN(ctx, rv);
 }
 
@@ -853,15 +861,39 @@ authentic_emu_update_tokeninfo(struct sc_profile *profile, struct sc_pkcs15_card
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
 
+static int
+authentic_pkcs15_init_card(struct sc_profile *profile, struct sc_pkcs15_card *p15card)
+{
+	LOG_FUNC_RETURN(p15card->card->ctx, SC_ERROR_NOT_SUPPORTED);
+}
+
+
+static int
+authentic_pkcs15_create_dir(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
+		struct sc_file *df)
+{
+	LOG_FUNC_RETURN(p15card->card->ctx, SC_ERROR_NOT_SUPPORTED);
+}
+
+
+static int
+authentic_pkcs15_create_pin(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
+		struct sc_file *df, struct sc_pkcs15_object *pin_obj,
+		const unsigned char *pin, size_t pin_len,
+		const unsigned char *puk, size_t puk_len)
+{
+	LOG_FUNC_RETURN(p15card->card->ctx, SC_ERROR_NOT_SUPPORTED);
+}
+
 
 static struct sc_pkcs15init_operations
 sc_pkcs15init_authentic_operations = {
 	authentic_pkcs15_erase_card,
-	NULL,					/* init_card  */
-	NULL,					/* create_dir */
+	authentic_pkcs15_init_card,
+	authentic_pkcs15_create_dir,
 	NULL,					/* create_domain */
 	NULL,					/* select_pin_reference */
-	NULL,					/* create_pin */
+	authentic_pkcs15_create_pin,
 	authentic_pkcs15_select_key_reference,
 	authentic_pkcs15_create_key,
 	authentic_pkcs15_store_key,
