@@ -2,7 +2,7 @@
  * CardOS specific operation for PKCS15 initialization
  *
  * Copyright (C) 2005 Nils Larsch <nils@larsch.net>
- * Copyright (C) 2002 Olaf Kirch <okir@lst.de>
+ * Copyright (C) 2002 Olaf Kirch <okir@suse.de>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -409,7 +409,7 @@ static int
 cardos_delete_object(sc_profile_t *profile, struct sc_pkcs15_card *p15card,
 		struct sc_pkcs15_object *obj, const struct sc_path *path)
 {
-	int r, stored_in_ef = 0, algorithm = 0;
+	int r = SC_SUCCESS, stored_in_ef = 0, algorithm = 0;
 	size_t keybits;
 	sc_file_t *file = NULL;
 	struct sc_pkcs15_prkey_info *key_info;
@@ -481,8 +481,9 @@ cardos_store_pin(sc_profile_t *profile, sc_card_t *card,
 	 * "no padding required". */
 	maxlen = MIN(profile->pin_maxlen, sizeof(pinpadded));
 	if (pin_len > maxlen) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "invalid pin length: %u (max %u)\n",
-		         pin_len, maxlen);
+		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+			 "invalid pin length: %"SC_FORMAT_LEN_SIZE_T"u (max %u)\n",
+			 pin_len, maxlen);
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
 	memcpy(pinpadded, pin, pin_len);
