@@ -30,10 +30,13 @@
 #include "libopensc/log.h"
 #include "pkcs15-init.h"
 #include "profile.h"
+#include "libopensc/asn1.h"
+#include "pkcs11/pkcs11.h"
 
 #undef KEEP_AC_NONE_FOR_INIT_APPLET
 
 #define MYEID_MAX_PINS   14
+#define MYEID_MAX_RSA_KEY_LEN 2048
 
 unsigned char MYEID_DEFAULT_PUBKEY[] = {0x01, 0x00, 0x01};
 #define MYEID_DEFAULT_PUBKEY_LEN       sizeof(MYEID_DEFAULT_PUBKEY)
@@ -183,8 +186,7 @@ myeid_init_card(sc_profile_t *profile,
         sc_format_path("3F00", &path);
 	r = sc_select_file(p15card->card, &path, &file);
 
-	if (file)
-		sc_file_free(file);
+	sc_file_free(file);
 
 	LOG_FUNC_RETURN(p15card->card->ctx, r);
 }
@@ -200,6 +202,7 @@ myeid_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df
 	static const char *create_dfs[] = {
 		"PKCS15-PrKDF",
 		"PKCS15-PuKDF",
+		"PKCS15-SKDF",
 		"PKCS15-CDF",
 		"PKCS15-CDF-TRUSTED",
 		"PKCS15-DODF",
@@ -209,6 +212,7 @@ myeid_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df
 	static const int create_dfs_val[] = {
 		SC_PKCS15_PRKDF,
 		SC_PKCS15_PUKDF,
+		SC_PKCS15_SKDF,
 		SC_PKCS15_CDF,
 		SC_PKCS15_CDF_TRUSTED,
 		SC_PKCS15_DODF
@@ -228,7 +232,9 @@ myeid_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df
 		for (ii = 0; create_dfs[ii]; ii++) {
 			sc_log(ctx, "Create '%s'", create_dfs[ii]);
 
-			if (sc_profile_get_file(profile, create_dfs[ii], &file)) {
+			r = sc_profile_get_file(profile, create_dfs[ii], &file);
+			sc_file_free(file);
+			if (r) {
 				sc_log(ctx, "Inconsistent profile: cannot find %s", create_dfs[ii]);
 				LOG_FUNC_RETURN(ctx, SC_ERROR_INCONSISTENT_PROFILE);
 			}
@@ -287,8 +293,10 @@ myeid_create_pin(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 	int r;
 
 	LOG_FUNC_CALLED(ctx);
-	sc_log(ctx, "PIN('%s',ref:%i,flags:0x%X,pin_len:%d,puk_len:%d)\n",
-			pin_obj->label, auth_info->attrs.pin.reference, auth_info->attrs.pin.flags, pin_len, puk_len);
+	sc_log(ctx,
+	       "PIN('%s',ref:%i,flags:0x%X,pin_len:%"SC_FORMAT_LEN_SIZE_T"u,puk_len:%"SC_FORMAT_LEN_SIZE_T"u)\n",
+	       pin_obj->label, auth_info->attrs.pin.reference,
+	       auth_info->attrs.pin.flags, pin_len, puk_len);
 
 	if (auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN)
 		return SC_ERROR_OBJECT_NOT_VALID;
@@ -344,19 +352,33 @@ myeid_new_file(sc_profile_t *profile, sc_card_t *card,
 	sc_file_t *file;
 	sc_path_t *p;
 	char name[64];
-	const char *tag;
+	const char *tag = NULL;
 	int r;
 
 	LOG_FUNC_CALLED(card->ctx);
-	if (type == SC_PKCS15_TYPE_PRKEY_RSA || type == SC_PKCS15_TYPE_PRKEY_EC)
+	switch (type) {
+	case SC_PKCS15_TYPE_PRKEY_RSA:
+	case SC_PKCS15_TYPE_PRKEY_EC:
 		tag = "private-key";
-	else if (type == SC_PKCS15_TYPE_PUBKEY_RSA || type == SC_PKCS15_TYPE_PUBKEY_EC)
+		break;
+	case SC_PKCS15_TYPE_PUBKEY_RSA:
+	case SC_PKCS15_TYPE_PUBKEY_EC:
 		tag = "public-key";
-	else if ((type & SC_PKCS15_TYPE_CLASS_MASK) == SC_PKCS15_TYPE_CERT)
-		tag = "certificate";
-	else if ((type & SC_PKCS15_TYPE_CLASS_MASK) == SC_PKCS15_TYPE_DATA_OBJECT)
-		tag = "data";
-	else {
+		break;
+	case SC_PKCS15_TYPE_SKEY_GENERIC:
+	case SC_PKCS15_TYPE_SKEY_DES:
+	case SC_PKCS15_TYPE_SKEY_3DES:
+		tag = "secret-key";
+		break;
+	default:
+		if ((type & SC_PKCS15_TYPE_CLASS_MASK) == SC_PKCS15_TYPE_CERT)
+			tag = "certificate";
+		else if ((type & SC_PKCS15_TYPE_CLASS_MASK) == SC_PKCS15_TYPE_DATA_OBJECT)
+			tag = "data";
+		break;
+	}
+
+	if (!tag) {
 		sc_log(card->ctx, "Unsupported file type");
 		return SC_ERROR_INVALID_ARGUMENTS;
 	}
@@ -404,67 +426,181 @@ myeid_encode_public_key(sc_profile_t *profile, sc_card_t *card,
 	LOG_FUNC_RETURN(card->ctx, 0);
 }
 
+/*
+ */
+static void
+_add_supported_algo(struct sc_profile *profile, struct sc_pkcs15_card *p15card, struct sc_pkcs15_object *object,
+		    unsigned operations, unsigned mechanism, const struct sc_object_id *oid)
+{
+	struct sc_supported_algo_info *algo;
+	algo = sc_pkcs15_get_supported_algo(p15card, operations, mechanism);
+	if (!algo) {
+		unsigned ref = 1, ii;
+
+		for (ii=0;ii<SC_MAX_SUPPORTED_ALGORITHMS && p15card->tokeninfo->supported_algos[ii].reference; ii++)
+			if (p15card->tokeninfo->supported_algos[ii].reference >= ref)
+				ref = p15card->tokeninfo->supported_algos[ii].reference + 1;
+		if (ii < SC_MAX_SUPPORTED_ALGORITHMS) {
+			algo = &p15card->tokeninfo->supported_algos[ii];
+			algo->reference = ref;
+			algo->mechanism = mechanism;
+			algo->operations = operations;
+			algo->algo_id = *oid;
+			profile->dirty = 1;
+			profile->pkcs15.do_last_update = 1;
+		}
+
+	}
+	sc_pkcs15_add_supported_algo_ref(object, algo);
+}
+
+static void
+myeid_fixup_supported_algos(struct sc_profile *profile, struct sc_pkcs15_card *p15card, struct sc_pkcs15_object *object)
+{
+	static const struct sc_object_id id_aes128_ecb = { { 2, 16, 840, 1, 101, 3, 4, 1, 1, -1 } };
+	static const struct sc_object_id id_aes128_cbc = { { 2, 16, 840, 1, 101, 3, 4, 1, 2, -1 } };
+	static const struct sc_object_id id_aes256_ecb = { { 2, 16, 840, 1, 101, 3, 4, 1, 41, -1 } };
+	static const struct sc_object_id id_aes256_cbc = { { 2, 16, 840, 1, 101, 3, 4, 1, 42, -1 } };
+	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_pkcs15_skey_info *skey_info = (struct sc_pkcs15_skey_info *) object->data;
+
+	LOG_FUNC_CALLED(ctx);
+	switch (object->type) {
+	case SC_PKCS15_TYPE_SKEY_GENERIC:
+		switch (skey_info->key_type | (skey_info->value_len << 16)) {
+		case CKM_AES_ECB | (128 << 16):
+			_add_supported_algo(profile, p15card, object, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_ECB, &id_aes128_ecb);
+			_add_supported_algo(profile, p15card, object, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_CBC, &id_aes128_cbc);
+			break;
+		case CKM_AES_ECB | (256 << 16):
+			_add_supported_algo(profile, p15card, object, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_ECB, &id_aes256_ecb);
+			_add_supported_algo(profile, p15card, object, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_CBC, &id_aes256_cbc);
+			break;
+		}
+		break;
+	}
+}
+
 
 /*
- * Store a private key
+ * Create a private key file
  */
 static int
 myeid_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		struct sc_pkcs15_object *object) {
 	struct sc_context *ctx = p15card->card->ctx;
 	struct sc_card *card = p15card->card;
-	struct sc_pkcs15_prkey_info *key_info = (struct sc_pkcs15_prkey_info *) object->data;
+	struct sc_pkcs15_prkey_info *prkey_info = (struct sc_pkcs15_prkey_info *) object->data;
+	struct sc_pkcs15_skey_info *skey_info = (struct sc_pkcs15_skey_info *) object->data;
+	struct sc_pkcs15_id *id;
+	struct sc_path *path;
+	int *key_reference;
 	struct sc_file *file = NULL;
-	int keybits = key_info->modulus_length, r;
+	struct sc_pkcs15_object *pin_object = NULL;
+	struct sc_pkcs15_auth_info *pkcs15_auth_info = NULL;
+	unsigned char sec_attrs[] = {0xFF, 0xFF, 0xFF};
+	int r, ef_structure = 0, keybits = 0, pin_reference = -1;
 
 	LOG_FUNC_CALLED(card->ctx);
 
-	/* Check that the card supports the requested modulus length */
 	switch (object->type) {
 		case SC_PKCS15_TYPE_PRKEY_RSA:
-			if (sc_card_find_rsa_alg(p15card->card, keybits) == NULL)
-				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS,
-					"Unsupported RSA key size");
+			ef_structure = SC_CARDCTL_MYEID_KEY_RSA;
+			keybits = prkey_info->modulus_length;
 			break;
 		case SC_PKCS15_TYPE_PRKEY_EC:
-			/* Here the information about curve is not available, that's why algorithm is checked
-			   without curve OID. */
-			if (sc_card_find_ec_alg(p15card->card, keybits, NULL) == NULL)
-				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported EC key size");
+			ef_structure = SC_CARDCTL_MYEID_KEY_EC;
+			keybits = prkey_info->field_length;
 			break;
-		default:
-			LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS,
-					"Unsupported key type");
+		case SC_PKCS15_TYPE_SKEY_DES:
+		case SC_PKCS15_TYPE_SKEY_3DES:
+			ef_structure = SC_CARDCTL_MYEID_KEY_DES;
+			keybits = skey_info->value_len;
+			break;
+		case SC_PKCS15_TYPE_SKEY_GENERIC:
+			keybits = skey_info->value_len;
+			switch (skey_info->key_type) {
+			case CKM_AES_ECB:
+				ef_structure = SC_CARDCTL_MYEID_KEY_AES;
+				break;
+			case CKM_DES_ECB:
+				ef_structure = SC_CARDCTL_MYEID_KEY_DES;
+				break;
+			}
 			break;
 	}
+	if (!ef_structure) {
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS,
+				"Unsupported key type");
+	}
 
-	sc_log(ctx, "create MyEID private key ID:%s", sc_pkcs15_print_id(&key_info->id));
+	myeid_fixup_supported_algos(profile, p15card, object);
+
+	if ((object->type & SC_PKCS15_TYPE_CLASS_MASK) == SC_PKCS15_TYPE_PRKEY) {
+		id = &prkey_info->id;
+		path = &prkey_info->path;
+		key_reference = &prkey_info->key_reference;
+	} else {
+		id = &skey_info->id;
+		path = &skey_info->path;
+		key_reference = &skey_info->key_reference;
+	}
+
+	sc_log(ctx, "create MyEID key ID:%s", sc_pkcs15_print_id(id));
 
 	/* Get the private key file */
-	r = myeid_new_file(profile, card, object->type, key_info->key_reference, &file);
-	LOG_TEST_RET(ctx, r, "Cannot get new MyEID private key file");
+	r = myeid_new_file(profile, card, object->type, *key_reference, &file);
+	LOG_TEST_RET(ctx, r, "Cannot get new MyEID key file");
 
 	if (!file || !file->path.len)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Cannot determine private key file");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Cannot determine key file");
 
 	sc_log(ctx, "Key file size %d", keybits);
 	file->size = keybits;
+	file->ef_structure = ef_structure;
 
-	if (object->type == SC_PKCS15_TYPE_PRKEY_RSA)
-		file->ef_structure = SC_CARDCTL_MYEID_KEY_RSA;
-	else if (object->type == SC_PKCS15_TYPE_PRKEY_EC)
-		file->ef_structure = SC_CARDCTL_MYEID_KEY_EC;
+	memcpy(path->value, &file->path.value, file->path.len);
+	*key_reference = file->path.value[file->path.len - 1] & 0xFF;
 
-	memcpy(&key_info->path.value, &file->path.value, file->path.len);
-	key_info->key_reference = file->path.value[file->path.len - 1] & 0xFF;
-
-	sc_log(ctx, "Path of MyEID private key file to create %s",
+	sc_log(ctx, "Path of MyEID key file to create %s",
 			sc_print_path(&file->path));
 
+	if (object->auth_id.len >= 1) {
+		r = sc_pkcs15_find_pin_by_auth_id(p15card, &object->auth_id, &pin_object);
+
+		if (r != SC_SUCCESS)
+			sc_file_free(file);
+		LOG_TEST_RET(ctx, r, "Failed to get pin object by auth_id");
+
+		if (pin_object->type != SC_PKCS15_TYPE_AUTH_PIN) {
+			sc_file_free(file);
+			LOG_TEST_RET(ctx, SC_ERROR_OBJECT_NOT_VALID, "Invalid object returned when locating pin object.");
+		}
+
+		pkcs15_auth_info =  (struct sc_pkcs15_auth_info*) pin_object->data;
+
+		if (pkcs15_auth_info == NULL || pkcs15_auth_info->auth_type != SC_PKCS15_PIN_AUTH_TYPE_PIN) {
+			sc_file_free(file);
+			LOG_TEST_RET(ctx, SC_ERROR_OBJECT_NOT_VALID, "NULL or invalid sc_pkcs15_auth_info in pin object");
+		}
+
+		pin_reference = pkcs15_auth_info->attrs.pin.reference;
+
+		if (pin_reference >= 1 && pin_reference < MYEID_MAX_PINS) {
+			sec_attrs[0] = (pin_reference << 4 | (pin_reference & 0x0F));
+			sec_attrs[1] = (pin_reference << 4 | (pin_reference & 0x0F));
+			sc_file_set_sec_attr(file, sec_attrs, sizeof(sec_attrs));
+		}
+	}
+	else {
+		sc_file_free(file);
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Invalid AuthID value for a private key.");
+	}
+
 	/* Now create the key file */
 	r = sc_pkcs15init_create_file(profile, p15card, file);
 	sc_file_free(file);
-	LOG_TEST_RET(ctx, r, "Cannot create MyEID private key file");
+	LOG_TEST_RET(ctx, r, "Cannot create MyEID key file");
 
 	LOG_FUNC_RETURN(ctx, r);
 }
@@ -478,52 +614,42 @@ myeid_store_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		struct sc_pkcs15_prkey *prkey) {
 	struct sc_context *ctx = p15card->card->ctx;
 	struct sc_card *card = p15card->card;
-	struct sc_pkcs15_prkey_info *key_info = (struct sc_pkcs15_prkey_info *) object->data;
 	struct sc_cardctl_myeid_gen_store_key_info args;
 	struct sc_file *file = NULL;
-	int r, keybits = key_info->modulus_length;
+	struct sc_pkcs15_id *id;
+	struct sc_path *path;
+	int r;
 
 	LOG_FUNC_CALLED(ctx);
 
-	switch (object->type) {
-		case SC_PKCS15_TYPE_PRKEY_RSA:
-			if (sc_card_find_rsa_alg(p15card->card, keybits) == NULL)
-				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported RSA key size");
-			break;
-		case SC_PKCS15_TYPE_PRKEY_EC:
-			if (!sc_valid_oid(&prkey->u.ec.params.id))
-                                if (sc_pkcs15_fix_ec_parameters(ctx, &prkey->u.ec.params))
-                                        LOG_FUNC_RETURN(ctx, SC_ERROR_OBJECT_NOT_VALID);
-			if (sc_card_find_ec_alg(p15card->card, keybits, &prkey->u.ec.params.id) == NULL)
-				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported algorithm or key size");
-			if(key_info->field_length != 0)
-				keybits = key_info->field_length;
-			else
-				key_info->field_length = keybits;
-			break;
-		default:
-			LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Store key failed: Unsupported key type");
-			break;
+	if ((object->type & SC_PKCS15_TYPE_CLASS_MASK) == SC_PKCS15_TYPE_PRKEY) {
+		struct sc_pkcs15_prkey_info *prkey_info = (struct sc_pkcs15_prkey_info *) object->data;
+		id = &prkey_info->id;
+		path = &prkey_info->path;
+	} else {
+		struct sc_pkcs15_skey_info *skey_info = (struct sc_pkcs15_skey_info *) object->data;
+		id = &skey_info->id;
+		path = &skey_info->path;
 	}
 
 	sc_log(ctx, "store MyEID key with ID:%s and path:%s",
-			sc_pkcs15_print_id(&key_info->id), sc_print_path(&key_info->path));
+			sc_pkcs15_print_id(id), sc_print_path(path));
 
-	r = sc_select_file(card, &key_info->path, &file);
+	r = sc_select_file(card, path, &file);
 	LOG_TEST_RET(ctx, r, "Cannot store MyEID key: select key file failed");
 
 	r = sc_pkcs15init_authenticate(profile, p15card, file, SC_AC_OP_UPDATE);
 	LOG_TEST_RET(ctx, r, "No authorisation to store MyEID private key");
 
-	if (file)
-		sc_file_free(file);
+	sc_file_free(file);
 
 	/* Fill in data structure */
 	memset(&args, 0, sizeof (args));
 
 	args.op_type = OP_TYPE_STORE;
-	if(object->type == SC_PKCS15_TYPE_PRKEY_RSA) {
-		//args.key_len_bits = keybits;
+
+	switch (object->type) {
+	case SC_PKCS15_TYPE_PRKEY_RSA:
 		args.key_type = SC_CARDCTL_MYEID_KEY_RSA;
 		args.pubexp_len = prkey->u.rsa.exponent.len;
 		args.pubexp = prkey->u.rsa.exponent.data;
@@ -539,16 +665,33 @@ myeid_store_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		args.invq_len = prkey->u.rsa.iqmp.len;
 		args.invq = prkey->u.rsa.iqmp.data;
 
+		//args.key_len_bits = keybits;
 		args.key_len_bits = prkey->u.rsa.modulus.len;
 		args.mod = prkey->u.rsa.modulus.data;
-	}
-	else {
+		break;
+	case SC_PKCS15_TYPE_PRKEY_EC:
 		args.key_type = SC_CARDCTL_MYEID_KEY_EC;
 		args.d = prkey->u.ec.privateD.data;
 		args.d_len = prkey->u.ec.privateD.len;
 		args.ecpublic_point = prkey->u.ec.ecpointQ.value;
 		args.ecpublic_point_len = prkey->u.ec.ecpointQ.len;
 		args.key_len_bits = prkey->u.ec.params.field_length;
+		break;
+	case SC_PKCS15_TYPE_SKEY_GENERIC:
+	case SC_PKCS15_TYPE_SKEY_DES:
+	case SC_PKCS15_TYPE_SKEY_2DES:
+	case SC_PKCS15_TYPE_SKEY_3DES:
+		switch (prkey->algorithm) {
+		case SC_ALGORITHM_AES:
+			args.key_type = SC_CARDCTL_MYEID_KEY_AES;
+			break;
+		case SC_ALGORITHM_DES:
+			args.key_type = SC_CARDCTL_MYEID_KEY_DES;
+			break;
+		}
+		args.d = prkey->u.secret.data;
+		args.d_len = prkey->u.secret.data_len;
+		break;
 	}
 	/* Store RSA key  */
 	r = sc_card_ctl(card, SC_CARDCTL_MYEID_GENERATE_STORE_KEY, &args);
@@ -567,8 +710,11 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 	struct sc_cardctl_myeid_gen_store_key_info args;
 	struct sc_file *file = NULL;
 	int r;
+	unsigned int cla,tag;
+	size_t taglen;
 	size_t keybits = key_info->modulus_length;
-	unsigned char raw_pubkey[256];
+	u8 raw_pubkey[MYEID_MAX_RSA_KEY_LEN / 8];
+	u8* dataptr;
 
 	LOG_FUNC_CALLED(ctx);
 	if (object->type != SC_PKCS15_TYPE_PRKEY_RSA && object->type != SC_PKCS15_TYPE_PRKEY_EC)
@@ -583,13 +729,16 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		case SC_PKCS15_TYPE_PRKEY_EC:
 			/* EC is supported in MyEID v > 3.5. TODO: set correct return value if older MyEID version. */
 			/* Here the information about curve is not available, that's why supported algorithm is checked
-			   without curve OID. */
-			if (sc_card_find_ec_alg(p15card->card, keybits, NULL) == NULL)
-				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported EC key size");
+			   without curve OID. */						
+                    
 			if(key_info->field_length != 0)
 				keybits = key_info->field_length;
 			else
 				key_info->field_length = keybits;
+			
+			if (sc_card_find_ec_alg(p15card->card, keybits, NULL) == NULL)
+				LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported EC key size");
+			
 			break;
 		default:
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported key type");
@@ -617,7 +766,7 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		args.key_type = SC_CARDCTL_MYEID_KEY_EC;
 	}
 
-	/* Generate RSA key  */
+	/* Generate the key  */
 	r = sc_card_ctl(card, SC_CARDCTL_MYEID_GENERATE_STORE_KEY, &args);
 	LOG_TEST_RET(ctx, r, "Card control 'MYEID_GENERATE_STORE_KEY' failed");
 
@@ -653,7 +802,10 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		else if (object->type == SC_PKCS15_TYPE_PRKEY_EC) {
 			struct sc_ec_parameters *ecparams = (struct sc_ec_parameters *)key_info->params.data;
 
-			sc_log(ctx, "curve '%s', len %i, oid '%s'", ecparams->named_curve, ecparams->field_length, sc_dump_oid(&(ecparams->id)));
+			sc_log(ctx,
+			       "curve '%s', len %"SC_FORMAT_LEN_SIZE_T"u, oid '%s'",
+			       ecparams->named_curve, ecparams->field_length,
+			       sc_dump_oid(&(ecparams->id)));
 			pubkey->algorithm = SC_ALGORITHM_EC;
 
 			r = sc_select_file(card, &file->path, NULL);
@@ -667,13 +819,23 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 			r = sc_card_ctl(card, SC_CARDCTL_MYEID_GETDATA, &data_obj);
 			LOG_TEST_RET(ctx, r, "Cannot get EC public key: 'MYEID_GETDATA' failed");
 
+			dataptr = data_obj.Data;
+			r = sc_asn1_read_tag((const u8 **)&dataptr, data_obj.DataLen, &cla, &tag, &taglen);
+			LOG_TEST_RET(ctx, r, "Invalid EC public key data. Cannot parse DER structure.");
+
+			if (taglen == 0)
+			    LOG_FUNC_RETURN(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
+
 			if (pubkey->u.ec.ecpointQ.value)
 				free(pubkey->u.ec.ecpointQ.value);
-			pubkey->u.ec.ecpointQ.value = malloc(data_obj.DataLen - 2);
-                        if (pubkey->u.ec.ecpointQ.value == NULL)
+
+			pubkey->u.ec.ecpointQ.value = malloc(taglen);
+
+			if (pubkey->u.ec.ecpointQ.value == NULL)
 				LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
-			memcpy(pubkey->u.ec.ecpointQ.value, data_obj.Data + 2, data_obj.DataLen - 2);
-			pubkey->u.ec.ecpointQ.len = data_obj.DataLen - 2;
+
+			memcpy(pubkey->u.ec.ecpointQ.value, dataptr, taglen);
+			pubkey->u.ec.ecpointQ.len = taglen;
 
 			if (pubkey->u.ec.params.named_curve)
 				free(pubkey->u.ec.params.named_curve);
@@ -691,8 +853,7 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		}
 	}
 
-	if (file)
-		sc_file_free(file);
+	sc_file_free(file);
 
 	LOG_FUNC_RETURN(ctx, r);
 }
