@@ -1,7 +1,7 @@
 /*
  * Initialize Cards according to PKCS#15
  *
- * Copyright (C) 2002 Olaf Kirch <okir@lst.de>
+ * Copyright (C) 2002 Olaf Kirch <okir@suse.de>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -327,12 +327,10 @@ sc_profile_load(struct sc_profile *profile, const char *filename)
 	scconf_context	*conf;
 	const char *profile_dir = NULL;
 	char path[PATH_MAX];
-	int             res = 0, i;
+	int res = 0, i;
 #ifdef _WIN32
 	char temp_path[PATH_MAX];
-	DWORD temp_len;
-	long rc;
-	HKEY hKey;
+	size_t temp_len;
 #endif
 
 	LOG_FUNC_CALLED(ctx);
@@ -341,26 +339,15 @@ sc_profile_load(struct sc_profile *profile, const char *filename)
 		if (profile_dir)
 			break;
 	}
+
 	if (!profile_dir) {
 #ifdef _WIN32
-		rc = RegOpenKeyEx(HKEY_CURRENT_USER, "Software\\OpenSC Project\\OpenSC", 0, KEY_QUERY_VALUE, &hKey);
-		if (rc == ERROR_SUCCESS) {
-			temp_len = PATH_MAX;
-			rc = RegQueryValueEx(hKey, "ProfileDir", NULL, NULL, (LPBYTE) temp_path, &temp_len);
-			if ((rc == ERROR_SUCCESS) && (temp_len < PATH_MAX))
-				profile_dir = temp_path;
-			RegCloseKey(hKey);
-		}
-		if (!profile_dir) {
-			rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, "Software\\OpenSC Project\\OpenSC", 0, KEY_QUERY_VALUE, &hKey);
-			if (rc == ERROR_SUCCESS) {
-				temp_len = PATH_MAX;
-				rc = RegQueryValueEx(hKey, "ProfileDir", NULL, NULL, (LPBYTE) temp_path, &temp_len);
-				if ((rc == ERROR_SUCCESS) && (temp_len < PATH_MAX))
-					profile_dir = temp_path;
-				RegCloseKey(hKey);
-			}
-		}
+		temp_len = PATH_MAX;
+		res = sc_ctx_win32_get_config_value(NULL, "ProfileDir", "Software\\OpenSC Project\\OpenSC",
+				temp_path, &temp_len);
+		if (res)
+			LOG_FUNC_RETURN(ctx, res);
+		profile_dir = temp_path;
 #else
 		profile_dir = SC_PKCS15_PROFILE_DIRECTORY;
 #endif
@@ -595,6 +582,8 @@ sc_profile_get_file_instance(struct sc_profile *profile, const char *name,
 	file->id += index;
         if(file->type == SC_FILE_TYPE_BSO) {
 		r = sc_profile_add_file(profile, name, file);
+		if (r < 0)
+			sc_file_free(file);
 		LOG_TEST_RET(ctx, r, "Profile error: cannot add BSO file");
 	}
 	else if (file->path.len)   {
@@ -602,11 +591,15 @@ sc_profile_get_file_instance(struct sc_profile *profile, const char *name,
 		file->path.value[file->path.len - 1] = file->id & 0xFF;
 
 		r = sc_profile_add_file(profile, name, file);
+		if (r < 0)
+			sc_file_free(file);
 		LOG_TEST_RET(ctx, r, "Profile error: cannot add file");
 	}
 
 	if (ret)
 		*ret = file;
+	else
+		sc_file_free(file);
 
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
@@ -2380,13 +2373,17 @@ expr_eval(struct num_exp_ctx *ctx, unsigned int *vp, unsigned int pri)
 		case '*':
 		case '/':
 			new_pri++;
+			/* fall through */
 		case '+':
 		case '-':
 			new_pri++;
+			/* fall through */
 		case '&':
 			new_pri++;
+			/* fall through */
 		case '|':
 			new_pri++;
+			/* fall through */
 		case ')':
 			break;
 		default:
