@@ -396,8 +396,8 @@ static int cardos_sm4h(const unsigned char *in, size_t inlen, unsigned char
 	unsigned int i,j;
 
 	if (keylen != 16) {
-		printf("key has wrong size, need 16 bytes, got %zd. aborting.\n",
-			keylen);
+		printf("key has wrong size, need 16 bytes, got %"SC_FORMAT_LEN_SIZE_T"d. aborting.\n",
+		       keylen);
 		return 0;
 	}
 
@@ -593,7 +593,8 @@ static int cardos_format(const char *opt_startkey)
 	if (check_apdu(&apdu))
 		return 1;
 	if (apdu.resplen < 0x04) {
-		printf("expected 4-6 bytes form GET DATA for startkey data, but got only %zu\n", apdu.resplen);
+		printf("expected 4-6 bytes form GET DATA for startkey data, but got only %"SC_FORMAT_LEN_SIZE_T"u\n",
+		       apdu.resplen);
 		printf("aborting\n");
 		return 1;
 	}
@@ -914,7 +915,8 @@ static int cardos_change_startkey(const char *change_startkey_apdu)
 	if (check_apdu(&apdu))
 		return 1;
 	if (apdu.resplen < 0x04) {
-		printf("expected 4-6 bytes form GET DATA for startkey data, but got only %zu\n", apdu.resplen);
+		printf("expected 4-6 bytes form GET DATA for startkey data, but got only %"SC_FORMAT_LEN_SIZE_T"u\n",
+		       apdu.resplen);
 		printf("aborting\n");
 		return 1;
 	}
@@ -999,7 +1001,8 @@ change_startkey:
 	if (check_apdu(&apdu))
 		return 1;
 	if (apdu.resplen < 0x04) {
-		printf("expected 4-6 bytes form GET DATA for startkey data, but got only %zu\n", apdu.resplen);
+		printf("expected 4-6 bytes form GET DATA for startkey data, but got only %"SC_FORMAT_LEN_SIZE_T"u\n",
+		       apdu.resplen);
 		printf("aborting\n");
 		return 1;
 	}
