@@ -85,6 +85,7 @@ static const char *option_help[] = {
 static int do_echo(int argc, char **argv);
 static int do_ls(int argc, char **argv);
 static int do_find(int argc, char **argv);
+static int do_find_tags(int argc, char **argv);
 static int do_cd(int argc, char **argv);
 static int do_cat(int argc, char **argv);
 static int do_info(int argc, char **argv);
@@ -127,6 +128,9 @@ static struct command	cmds[] = {
 	{ do_find,
 		"find",	"[<start id> [<end id>]]",
 		"find all files in the current DF"	},
+	{ do_find_tags,
+		"find_tags",	"[<start tag> [<end tag>]]",
+		"find all tags of data objects in the current context"	},
 	{ do_cd,
 		"cd",	"{.. | <file id> | aid:<DF name>}",
 		"change to another DF"			},
@@ -256,8 +260,7 @@ static int usage(int (*func)(int, char **))
 
 static void die(int ret)
 {
-	if (current_file != NULL)
-		sc_file_free(current_file);
+	sc_file_free(current_file);
 	if (card) {
 		sc_unlock(card);
 		sc_disconnect_card(card);
@@ -379,7 +382,10 @@ arg_to_path(const char *arg, sc_path_t *path, int is_id)
 static void print_file(const sc_file_t *file)
 {
 	const char *format = " %02X%02X ";
-	const char *st = "???";
+	const char *st = NULL;
+
+	if(!file->type_attr_len)
+		st = "???";
 
 	switch (file->type) {
 	case SC_FILE_TYPE_WORKING_EF:
@@ -394,7 +400,10 @@ static void print_file(const sc_file_t *file)
 		break;
 	}
 	printf(format, file->id >> 8, file->id & 0xFF);
-	printf("\t%4s", st);
+	if (st == NULL)
+		printf("\t0x%02X", *file->type_attr);
+	else
+		printf("\t%4s", st);
 	printf(" %5lu", (unsigned long)file->size);
 	if (file->namelen) {
 		printf("\tName: ");
@@ -579,6 +588,73 @@ static int do_find(int argc, char **argv)
 	return 0;
 }
 
+static int do_find_tags(int argc, char **argv)
+{
+	u8 start[2], end[2], rbuf[256];
+	int r;
+	unsigned int tag, tag_end;
+
+	start[0] = 0x00;
+	start[1] = 0x00;
+	end[0] = 0xFF;
+	end[1] = 0xFF;
+	switch (argc) {
+	case 2:
+		if (arg_to_fid(argv[1], end) != 0)
+			return usage(do_find_tags);
+		/* fall through */
+	case 1:
+		if (arg_to_fid(argv[0], start) != 0)
+			return usage(do_find_tags);
+		/* fall through */
+	case 0:
+		break;
+	default:
+		return usage(do_find_tags);
+	}
+	tag = (start[0] << 8) | start[1];
+	tag_end = (end[0] << 8) | end[1];
+
+	printf("Tag\tType\n");
+	while (1) {
+		printf("(%04X)\r", tag);
+		fflush(stdout);
+
+		r = sc_get_data(card, tag, rbuf, sizeof rbuf);
+		if (r >= 0) {
+			printf(" %04X ", tag);
+			if (tag == 0)
+				printf("\tdump file");
+			if ((0x0001 <= tag && tag <= 0x00FE)
+					|| (0x1F1F <= tag && tag <= 0xFFFF))
+				printf("\tBER-TLV");
+			if (tag == 0x00FF || tag == 0x02FF)
+				printf("\tspecial function");
+			if (0x0100 <= tag && tag <= 0x01FF)
+				printf("\tproprietary");
+			if (tag == 0x0200)
+				printf("\tRFU");
+			if (0x0201 <= tag && tag <= 0x02FE)
+				printf("\tSIMPLE-TLV");
+			printf("\n");
+			if (r > 0)
+				util_hex_dump_asc(stdout, rbuf, r, -1);
+		} else {
+			switch (r) {
+				case SC_ERROR_NOT_ALLOWED:
+				case SC_ERROR_SECURITY_STATUS_NOT_SATISFIED:
+					printf("(%04X)\t%s\n", tag, sc_strerror(r));
+					break;
+			}
+		}
+
+		if (tag >= tag_end)
+			break;
+		tag++;
+	}
+	return 0;
+}
+
 static int do_cd(int argc, char **argv)
 {
 	sc_path_t path;
@@ -607,8 +683,7 @@ static int do_cd(int argc, char **argv)
 			printf("unable to go up: %s\n", sc_strerror(r));
 			return -1;
 		}
-		if (current_file)
-			sc_file_free(current_file);
+		sc_file_free(current_file);
 		current_file = file;
 		current_path = path;
 		return 0;
@@ -628,8 +703,7 @@ static int do_cd(int argc, char **argv)
 		return -1;
 	}
 	current_path = path;
-	if (current_file)
-		sc_file_free(current_file);
+	sc_file_free(current_file);
 	current_file = file;
 
 	return 0;
@@ -741,8 +815,7 @@ static int do_cat(int argc, char **argv)
 	err = 0;
 err:
 	if (not_current) {
-		if (file != NULL)
-			sc_file_free(file);
+		sc_file_free(file);
 		select_current_path_or_die();
 	}
 
@@ -754,7 +827,7 @@ static int do_info(int argc, char **argv)
 	sc_file_t *file;
 	sc_path_t path;
 	size_t i;
-	const char *st;
+	const char *st = NULL;
 	int r, not_current = 1;
 	const id2str_t *ac_ops = NULL;
 
@@ -774,6 +847,9 @@ static int do_info(int argc, char **argv)
 	} else
 		return usage(do_info);
 
+	if(!file->type_attr_len)
+		st = "Unknown File";
+
 	switch (file->type) {
 	case SC_FILE_TYPE_WORKING_EF:
 	case SC_FILE_TYPE_INTERNAL_EF:
@@ -782,11 +858,11 @@ static int do_info(int argc, char **argv)
 	case SC_FILE_TYPE_DF:
 		st = "Dedicated File";
 		break;
-	default:
-		st = "Unknown File";
-		break;
 	}
-	printf("\n%s  ID %04X", st, file->id);
+	if (st == NULL)
+		printf("\nFile type [%02x] ID %04X", *file->type_attr, file->id);
+	else
+		printf("\n%s  ID %04X", st, file->id);
 	if (file->sid)
 		printf(", SFI %02X", file->sid);
 	printf("\n\n%-15s%s\n", "File path:", path_to_filename(&path, '/'));
@@ -1018,7 +1094,7 @@ static int do_verify(int argc, char **argv)
 			printf("Please enter PIN: ");
 			r = util_getpass(&pin, &len, stdin);
 			if (r < 0) {
-				printf("No PIN entered - aborting VERIFY.\n");
+				printf("util_getpass error.\n");
 				return -1;
 			}
 
@@ -1238,8 +1314,7 @@ static int do_get(int argc, char **argv)
 
 	err = 0;
 err:
-	if (file)
-		sc_file_free(file);
+	sc_file_free(file);
 	if (outf != NULL && outf != stdout)
 		fclose(outf);
 	select_current_path_or_die();
@@ -1412,8 +1487,7 @@ static int do_put(int argc, char **argv)
 
 	err = 0;
 err:
-	if (file)
-		sc_file_free(file);
+	sc_file_free(file);
 	if (outf)
 		fclose(outf);
 	select_current_path_or_die();
@@ -1649,20 +1723,24 @@ static int do_asn1(int argc, char **argv)
 		goto err;
 	}
 	if ((size_t)r != len) {
-		printf("expecting %lu, got only %d bytes.\n", (unsigned long) len, r);
-		goto err;
+		printf("WARNING: expecting %lu, got %d bytes.\n", (unsigned long) len, r);
+		/* some cards return a bogus value for file length. As
+		 * long as the actual length is not higher than the expected
+		 * length, continue */
+		if(r > (signed)len) {
+			goto err;
+		}
 	}
 
 	/* asn1 dump */
-	sc_asn1_print_tags(buf, len);
+	sc_asn1_print_tags(buf, r);
 
 	err = 0;
 err:
 	if (buf)
 		free(buf);
 	if (not_current) {
-		if (file)
-			sc_file_free(file);
+		sc_file_free(file);
 		select_current_path_or_die();
 	}
 	return -err;
@@ -1836,7 +1914,7 @@ int main(int argc, char * const argv[])
 		return 1;
 	}
 
-	ctx->enable_default_driver = 1;
+	ctx->flags |= SC_CTX_FLAG_ENABLE_DEFAULT_DRIVER;
 
 	if (verbose > 1) {
 		ctx->debug = verbose;
