@@ -58,9 +58,11 @@ enum {
 	OPT_MD5,
 	OPT_PKCS1,
 	OPT_BIND_TO_AID,
+	OPT_VERSION,
 };
 
 static const struct option options[] = {
+	{ "version",		0, NULL,		OPT_VERSION },
 	{ "sign",		0, NULL,		's' },
 	{ "decipher",		0, NULL,		'c' },
 	{ "key",		1, NULL,		'k' },
@@ -84,12 +86,13 @@ static const struct option options[] = {
 };
 
 static const char *option_help[] = {
+	"Print OpenSC package version",
 	"Performs digital signature operation",
 	"Decipher operation",
 	"Selects the private key ID to use",
 	"Uses reader number <arg>",
-	"Selects the input file to use",
-	"Outputs to file <arg>",
+	"Selects the input file to use (defaults to stdin)",
+	"Outputs to file <arg> (defaults to stdout)",
 	"Format for ECDSA signature <arg>: 'rs' (default), 'sequence', 'openssl'",
 	"Outputs raw 8 bit data",
 	"Input file is a SHA-1 hash",
@@ -140,7 +143,7 @@ static char * get_pin(struct sc_pkcs15_object *obj)
 			return strdup(opt_pincode);
 	}
 
-	sprintf(buf, "Enter PIN [%s]: ", obj->label);
+	sprintf(buf, "Enter PIN [%.*s]: ", (int) sizeof obj->label, obj->label);
 	while (1) {
 		pincode = getpass(buf);
 		if (strlen(pincode) == 0)
@@ -157,13 +160,19 @@ static int read_input(u8 *buf, int buflen)
 	FILE *inf;
 	int c;
 
-	inf = fopen(opt_input, "rb");
-	if (inf == NULL) {
-		fprintf(stderr, "Unable to open '%s' for reading.\n", opt_input);
-		return -1;
+	if (opt_input==NULL) {
+		inf = stdin;
+	} else {
+		inf = fopen(opt_input, "rb");
+		if (inf == NULL) {
+			fprintf(stderr, "Unable to open '%s' for reading.\n", opt_input);
+			return -1;
+		}
 	}
 	c = fread(buf, 1, buflen, inf);
-	fclose(inf);
+	if (inf!=stdin) {
+		fclose(inf);
+	}
 	if (c < 0) {
 		perror("read");
 		return -1;
@@ -203,8 +212,7 @@ static int sign(struct sc_pkcs15_object *obj)
 	int r, c, len;
 
 	if (opt_input == NULL) {
-		fprintf(stderr, "No input file specified.\n");
-		return 2;
+		fprintf(stderr, "No input file specified. Reading from stdin\n");
 	}
 
 	c = read_input(buf, sizeof(buf));
@@ -258,8 +266,7 @@ static int decipher(struct sc_pkcs15_object *obj)
 	int r, c, len;
 
 	if (opt_input == NULL) {
-		fprintf(stderr, "No input file specified.\n");
-		return 2;
+		fprintf(stderr, "No input file specified. Reading from stdin\n");
 	}
 	c = read_input(buf, sizeof(buf));
 	if (c < 0)
@@ -350,8 +357,9 @@ int main(int argc, char * const argv[])
 	int err = 0, r, c, long_optind = 0;
 	int do_decipher = 0;
 	int do_sign = 0;
+	int do_print_version = 0;
 	int action_count = 0;
-        struct sc_pkcs15_object *key;
+	struct sc_pkcs15_object *key;
 	sc_context_param_t ctx_param;
 
 	while (1) {
@@ -361,6 +369,10 @@ int main(int argc, char * const argv[])
 		if (c == '?')
 			util_print_usage_and_die(app_name, options, option_help, NULL);
 		switch (c) {
+		case OPT_VERSION:
+			do_print_version = 1;
+			action_count++;
+			break;
 		case 's':
 			do_sign++;
 			action_count++;
@@ -426,6 +438,11 @@ int main(int argc, char * const argv[])
 	if (action_count == 0)
 		util_print_usage_and_die(app_name, options, option_help, NULL);
 
+	if (do_print_version)   {
+		printf("%s\n", OPENSC_SCM_REVISION);
+		action_count--;
+	}
+
 	if (!(opt_crypt_flags & SC_ALGORITHM_RSA_HASHES))
 		opt_crypt_flags |= SC_ALGORITHM_RSA_HASH_NONE;
 
@@ -473,7 +490,7 @@ int main(int argc, char * const argv[])
 		fprintf(stderr, "Found %s!\n", p15card->tokeninfo->label);
 
 	if (do_decipher) {
-		if ((err = get_key(SC_PKCS15_PRKEY_USAGE_DECRYPT, &key))
+		if ((err = get_key(SC_PKCS15_PRKEY_USAGE_DECRYPT|SC_PKCS15_PRKEY_USAGE_UNWRAP, &key))
 		 || (err = decipher(key)))
 			goto end;
 		action_count--;
