@@ -78,14 +78,20 @@
 #define HAVE_IO_H
 #define ENABLE_PCSC
 #define HAVE_WINSCARD_H
-#define DEFAULT_PCSC_PROVIDER "winscard.dll"
+#ifndef DEFAULT_PCSC_PROVIDER
+#define DEFAULT_PCSC_PROVIDER "libpcsclite.so.1"
+#endif
 
 #define SC_PKCS15_PROFILE_DIRECTORY "C:\\Program Files\\OpenSC Project\\OpenSC\\profiles"
 
 #define PATH_MAX FILENAME_MAX
 
 #ifndef PACKAGE_VERSION
-#define PACKAGE_VERSION "0.16.0"
+#define PACKAGE_VERSION "0.17.0"
+#endif
+
+#ifndef VERSION
+#define VERSION PACKAGE_VERSION
 #endif
 
 #ifndef PACKAGE_NAME
@@ -97,27 +103,31 @@
 #endif
 
 #ifndef OPENSC_VERSION_MAJOR
-#define OPENSC_VERSION_MAJOR 0
+#define OPENSC_VERSION_MAJOR "0"
 #endif
 
 #ifndef OPENSC_VERSION_MINOR
-#define OPENSC_VERSION_MINOR 15
+#define OPENSC_VERSION_MINOR "17"
 #endif
 
 #ifndef OPENSC_VS_FF_COMPANY_NAME
-#define OPENSC_VS_FF_COMPANY_NAME "OpenSC Project"
+#define OPENSC_VS_FF_COMPANY_NAME  "OpenSC Project"
 #endif
 
 #ifndef OPENSC_VS_FF_PRODUCT_NAME
-#define OPENSC_VS_FF_PRODUCT_NAME "OpenSC smartcard framework"
+#define OPENSC_VS_FF_PRODUCT_NAME  "OpenSC smartcard framework"
+#endif
+
+#ifndef CVCDIR
+#define CVCDIR "%PROGRAMFILES%\\OpenSC Project\\OpenSC\\cvc"
 #endif
 
 #ifndef DEFAULT_PKCS11_PROVIDER
-#define DEFAULT_PKCS11_PROVIDER "opensc-pkcs11.dll"
+#define DEFAULT_PKCS11_PROVIDER "opensc-pkcs11.so"
 #endif
 
 #ifndef OPENSC_SCM_REVISION
-#define OPENSC_SCM_REVISION "OpenSC-0.16.0, rev: 7eeba1f, commit-time: 2016-06-03 11:19:51 +0200"
+#define OPENSC_SCM_REVISION "OpenSC-0.16.0-307-gfb74935c, rev: fb74935c, commit-time: 2017-07-19 09:29:18 +0200"
 #endif
 
 #endif
